/*
  Project:      POVColor
  File:         app.cpp
  Purpose:      Application implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/app.h>
#include <wx/cmdline.h>
#include <wx/utils.h>
#include <wx/log.h>
#include <wx/intl.h>
#include <wx/filefn.h>
#include <wx/filename.h>
#include <wx/stdpaths.h>
#include <wx/fileconf.h>

#include <wx/image.h>
#include <wx/imagpng.h>

#include "app.h"
#include "art.h"
#include "options.h"
#include "mainwnd.h"


// PovColor
//-----------------------------------------------------------------------------

// Implement application
IMPLEMENT_APP(PovColor);

// Ctor
PovColor::PovColor() :
    m_config(NULL),
    m_options(NULL)
{
}

// Dtor
PovColor::~PovColor()
{
}

// Init
bool PovColor::OnInit()
{
    // Set app and vendor name
    SetAppName(wxT("povcolor"));
    SetAppDisplayName(wxT("POVColor"));

    SetVendorName(wxT("bavii"));
    SetVendorDisplayName(wxT("BAVII"));

    // Init base and also handle command line
    if(wxApp::OnInit() == false)
    {
        return false;
    }

    // Create directories
    CreateUserDirs();

    // Prepare the configuration object
    m_config = new wxFileConfig(GetAppName(), GetVendorName(), 
                                GetConfigFile(), wxEmptyString,
                                wxCONFIG_USE_LOCAL_FILE);
    wxConfigBase::Set(m_config);
  
    // Load options object, uses default configuration object
    m_options = new Options();
    m_options->Load(m_config);

    // Initialize image support
    wxImage::AddHandler(new wxPNGHandler);

    // Art provider
    Art* art = new Art;
    art->SetDirectory(GetPixmapsDir());
    art->AddSize(1, 23, wxT("16x16"));
    art->AddSize(24, 39, wxT("32x32"));
    art->AddSize(40, 55, wxT("48x48"));
    art->AddSize(54, 71, wxT("64x64"));
    art->AddExt(wxT("png"));
    art->AddExt(wxT("xpm"));

    wxArtProvider::Push(art);

    // Create and show main window
    MainWindow* wnd = new MainWindow();
    SetTopWindow(wnd);
    wnd->Show(true);

    // Open files
    for(size_t pos = 0, count = m_files.GetCount(); pos < count; pos++)
    {
        wnd->OpenFile(m_files[pos]);
    }

    // Success
    return true;
}

// Exit
int PovColor::OnExit()
{
    // Save options
    m_options->Save(m_config);
    delete m_options;
    m_options = NULL;

    // Save configuration
    wxConfigBase::DontCreateOnDemand();
    wxConfigBase::Set(NULL);
    m_config->Flush();
    delete m_config;
    m_config = NULL;
  
    // Return
    return wxApp::OnExit();
}

// Init command line
void PovColor::OnInitCmdLine(wxCmdLineParser& parser)
{
    wxApp::OnInitCmdLine(parser);

    // Paths options
    parser.AddOption(wxT(""), wxT("appdatadir"), wxT("location of application data directory"), wxCMD_LINE_VAL_STRING);
    parser.AddOption(wxT(""), wxT("userdatadir"), wxT("location of the user data directory"), wxCMD_LINE_VAL_STRING);

    // Files
    parser.AddParam(wxT("file(s) to open"), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL | wxCMD_LINE_PARAM_MULTIPLE);
}

// Handle command line
bool PovColor::OnCmdLineParsed(wxCmdLineParser& parser)
{
    if(wxApp::OnCmdLineParsed(parser) == false)
    {
        parser.Usage();
        return false;
    }

    // Remember paths
    wxString path;
    wxFileName fn;

    if(parser.Found(wxT("appdatadir"), &path))
    {
        fn.AssignDir(path);
        fn.Normalize();
        m_appDataDir = fn.GetPath();
    }
    else
    {
        m_appDataDir = wxStandardPaths::Get().GetDataDir();
    }

    if(parser.Found(wxT("userdatadir"), &path))
    {
        fn.AssignDir(path);
        fn.Normalize();
        m_userDataDir = fn.GetPath();
    }
    else
    {
        m_userDataDir = wxStandardPaths::Get().GetUserDataDir();
    }

    // Remember files to open
    for(size_t pos = 0, count = parser.GetParamCount(); pos < count; pos++)
    {
        // Make any files absolute with respect to current directory
        wxFileName fn(parser.GetParam(pos));
        fn.Normalize();

        m_files.Add(fn.GetFullPath());
    }

    return true;
}

// App data directory
wxString PovColor::GetAppDataDir()
{
    return m_appDataDir;
}

// User data directory
wxString PovColor::GetUserDataDir()
{
    return m_userDataDir;
}

// Pixmaps directory
wxString PovColor::GetPixmapsDir()
{
    wxFileName fn;

    fn.AssignDir(GetAppDataDir());
    fn.AppendDir(wxT("pixmaps"));
    
    return fn.GetPath();
}

// Location of help file
wxString PovColor::GetHelpFile()
{
    // Get path
    wxFileName fn;

    fn.AssignDir(GetAppDataDir());
    fn.AppendDir(wxT("doc"));
    fn.SetFullName(wxT("index.html"));

    return fn.GetFullPath();
}

// Location of the standard doc file like AUTHORS,COPYING
wxString PovColor::GetStandardDocFile(const wxString& file)
{
    wxFileName fn;

    fn.AssignDir(GetAppDataDir());
    fn.AppendDir(wxT("doc"));
    fn.SetFullName(file);

    return fn.GetFullPath();
}

// Location of the config file for wxFileConfig
wxString PovColor::GetConfigFile()
{
    wxFileName fn;
    fn.AssignDir(GetUserDataDir());
    fn.SetName(GetAppName());
    fn.SetExt(wxT("conf"));

    return fn.GetFullPath();
}

// Create app directories
void PovColor::CreateUserDirs()
{
    if(wxFileName::Mkdir(GetUserDataDir(), 0777, wxPATH_MKDIR_FULL) == false)
    {
        ::wxLogError(_("Unable to create user data directory"));
    }
}


