/*
  File:         art.cpp
  Purpose:      Art provider implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would
     be appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not
     be misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include <wx/defs.h>
#include <wx/artprov.h>
#include <wx/filename.h>
#include <wx/log.h>
#include <wx/image.h>

#include "art.h"

#include <wx/arrimpl.cpp>


// Art
//-----------------------------------------------------------------------------

// Complete array
WX_DEFINE_OBJARRAY(ArrayArtSize);

// Since wxWidgets < 3 doesn't have wxArtProvider::GetIconBundle
// Art::GetIconBundle remembers the last set provider to call
#if !wxCHECK_VERSION(3, 0, 0)
    static Art* gs_art = 0;
#endif

// ctor
Art::Art()
{
    #if !wxCHECK_VERSION(3, 0, 0)
        gs_art = this;
    #endif
}

// dtor
Art::~Art()
{
    #if !wxCHECK_VERSION(3, 0, 0)
        if(gs_art == this)
        {
            gs_art = 0;
        }
    #endif
}

// Get Icon
wxIcon Art::GetIcon(const wxArtID& id, const wxArtClient& client, const wxSize& size)
{
    return wxArtProvider::GetIcon(id, client, size);
}

// Get bitmap
wxBitmap Art::GetBitmap(const wxArtID& id, const wxArtClient& client, const wxSize& size)
{
    return wxArtProvider::GetBitmap(id, client, size);
}

// Get icon bundle
wxIconBundle Art::GetIconBundle(const wxArtID& id, const wxArtClient& client)
{
    #if wxCHECK_VERSION(3, 0, 0)
        // CreateIconBundle is virtual and called from art provider
        return wxArtProvider::GetIconBundle(id, client);
    #else
        // CreatIconBundle does not exist yet in wxWidgets < 3
        if(gs_art)
        {
            return gs_art->CreateIconBundle(id, client);
        }
        else
        {
            return wxIconBundle();
        }
    #endif
}

// create bitmap
wxBitmap Art::CreateBitmap(const wxArtID& id, const wxArtClient& WXUNUSED(client), const wxSize& size)
{
    // Determine size-specific directory
    wxString sizeDir;
    if(size != wxDefaultSize)
    {
        for(size_t pos = 0, count = m_sizes.GetCount(); pos < count; pos++)
        {
            if(size.x >= m_sizes[pos].min && size.x <= m_sizes[pos].max)
            {
                sizeDir = m_sizes[pos].dir;
                break;
            }
        }
    }

    // Get the bitmap
    wxBitmap bmp;

    if(sizeDir.IsEmpty() == false)
    {
        bmp = FetchBitmap(id, sizeDir);
    }

    if(bmp.IsOk() == false)
    {
        bmp = FetchBitmap(id);
    }

    return bmp;
}

// get an icon bundle
wxIconBundle Art::CreateIconBundle(const wxArtID& id, const wxArtClient& WXUNUSED(client))
{
    wxIconBundle bundle;
    wxIcon icon;

    // First add top icon if it exists
    icon = FetchIcon(id);
    if(icon.IsOk())
    {
        bundle.AddIcon(icon);
    }

    // Now add icon from each size directory, if one mathces top directory
    // size it will replace it
    for(size_t pos = 0, count = m_sizes.GetCount(); pos < count; pos++)
    {
        icon = FetchIcon(id, m_sizes[pos].dir);
        if(icon.IsOk())
        {
            bundle.AddIcon(icon);
        }
    }

    return bundle;
}

// fetch a bitmap
wxBitmap Art::FetchBitmap(const wxString& name, const wxString& sizeDir)
{
    wxLogNull nolog;

    // Try extensions
    for(size_t pos = 0, count = m_extensions.GetCount(); pos < count; pos++)
    {
        wxFileName fn;

        fn.Clear();
        if(sizeDir.IsEmpty() == false)
        {
            fn.AssignDir(sizeDir);
        }
        fn.SetName(name);
        fn.SetExt(m_extensions[pos]);
        fn.Normalize(wxPATH_NORM_ALL, m_directory);
    
        wxImage img;
        if(img.LoadFile(fn.GetFullPath(), wxBITMAP_TYPE_ANY) && img.IsOk())
        {
            return wxBitmap(img);
        }
    }

    return wxNullBitmap;
}

// fetch an icon
wxIcon Art::FetchIcon(const wxString& name, const wxString& sizeDir)
{
    wxBitmap bmp = FetchBitmap(name, sizeDir);
    if(bmp.IsOk())
    {
        wxIcon icon;
        icon.CopyFromBitmap(bmp);
        if(icon.IsOk())
        {
            return icon;
        }
    }

    return wxNullIcon;
}



