/*
  Project:      POVColor
  File:         coldlg.cpp
  Purpose:      Color dialog implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/stattext.h>
#include <wx/dcclient.h>
#include <wx/dcbuffer.h>
#include <wx/intl.h>
#include <wx/sizer.h>
#include <wx/utils.h>
#include <wx/settings.h>
#include <wx/config.h>
#include <wx/tokenzr.h>

#include "app.h"
#include "coldlg.h"
#include "options.h"
#include "colorcell.h"
#include "colorconv.h"


// Defines
#define CELLSIZE 18 // Size of basic/custom color cell
#define CELLSPACE 6 // Spacing between basic/custom color cells

#define AREASPACE_X 15 // Space between left (basic/custom/preview) and right
#define AREASPACE_Y 15 // Space between basic/custom and custon/preview

#define BAR_X 64  // Size of a bar
#define BAR_Y 256
#define BARSPACE_X 15 // Space between the bars

#define PREVIEW_Y 65 // Height of preview, width is equal to left stuff 


// ColorBarWidget - A widget to draw a color bar with one component gradient
//-----------------------------------------------------------------------------

// Events
const wxEventType EVENT_COLORBAR_CHANGING = wxNewEventType();
const wxEventType EVENT_COLORBAR_CHANGED = wxNewEventType();

ColorBarEvent::ColorBarEvent(wxEventType type, wxWindowID id, const Color& color) :
        wxCommandEvent(type, id), m_color(color)
{
}

ColorBarEvent::ColorBarEvent(const ColorBarEvent& copy) :
        wxCommandEvent(copy), m_color(copy.m_color)
{
}

ColorBarEvent::~ColorBarEvent()
{
}

// Color bar widget

#define SLIDER_X 12
#define SLIDER_Y 14

// Constructor
ColorBarWidget::ColorBarWidget(wxWindow* parent,
                               wxWindowID id,
                               const wxPoint& pos,
                               const wxSize& size,
                               ColorConvertor* convertor,
                               size_t index)
{
    wxControl::Create(parent, id, pos, size, wxBORDER_SUNKEN);
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    SetMinSize(size);

    // Build the parts
    m_convertor = convertor;
    m_index = index;

    m_components = new double[m_convertor->GetComponentCount()];
    m_convertor->FromColor(m_color, m_components);
}

// Destructor
ColorBarWidget::~ColorBarWidget()
{
    delete [] m_components;
}

// Set color
void ColorBarWidget::SetColor(const Color& color)
{
    // Save the color and convert the components
    m_color = color;
    m_convertor->FromColor(m_color, m_components);

    // Draw
    Refresh();
    Update();
}

// Set components
void ColorBarWidget::SetComponents(const double* components)
{
    size_t cur, count = m_convertor->GetComponentCount();
    for(cur = 0; cur < count; cur++)
        m_components[cur] = components[cur];

    m_convertor->ToColor(m_color, m_components);

    // Draw
    Refresh();
    Update();
}

// Left button pressed
void ColorBarWidget::OnLeftDown(wxMouseEvent& evt)
{
    // Capture the mouse
    CaptureMouse();

    // Determine position
    wxRect rect(GetClientRect());
    double pos = (double)(rect.y + rect.height - (int)evt.GetY()) / (double)(rect.height);

    pos = clip(pos, 0.0, 1.0);

    // Set component and update color
    m_components[m_index] = pos;
    m_convertor->ToColor(m_color, m_components);

    Refresh();
    Update();

    // Emit event
    ColorBarEvent cevt(EVENT_COLORBAR_CHANGING, GetId(), m_color);
    cevt.SetEventObject(this);

    GetParent()->ProcessEvent(cevt);
}

// Left button up
void ColorBarWidget::OnLeftUp(wxMouseEvent& WXUNUSED(evt))
{
    // Release
    if(HasCapture())
        ReleaseMouse();
    
    // Emit event
    ColorBarEvent cevt(EVENT_COLORBAR_CHANGED, GetId(), m_color);
    cevt.SetEventObject(this);

    GetParent()->ProcessEvent(cevt);
}

// Motion
void ColorBarWidget::OnMotion(wxMouseEvent& evt)
{
    if(!HasCapture())
    {
        return;
    }

    // Determine position
    wxRect rect(GetClientRect());
    double pos = (double)(rect.y + rect.height - (int)evt.GetY()) / (double)(rect.height);

    pos = clip(pos, 0.0, 1.0);

    // Set component and update color
    m_components[m_index] = pos;
    m_convertor->ToColor(m_color, m_components);

    Refresh();
    Update();

    // Emit event
    ColorBarEvent cevt(EVENT_COLORBAR_CHANGING, GetId(), m_color);
    cevt.SetEventObject(this);

    GetParent()->ProcessEvent(cevt);
}

void ColorBarWidget::OnCaptureLost(wxMouseCaptureLostEvent& WXUNUSED(evt))
{
    // Emit event
    ColorBarEvent cevt(EVENT_COLORBAR_CHANGED, GetId(), m_color);
    cevt.SetEventObject(this);

    GetParent()->ProcessEvent(cevt);
}

void ColorBarWidget::OnPaint(wxPaintEvent& WXUNUSED(evt))
{
    wxAutoBufferedPaintDC dc(this);
    wxRect rect(GetClientRect());

    dc.SetClippingRegion(rect);

    // Block size
    int block;
    if(rect.width < rect.height)
    {
        block = rect.width / 2;
    }
    else
    {
        block = rect.height / 2;
    }

    // Some ponts
    int xEnd = rect.x + rect.width;
    int yEnd = rect.y + rect.height;
    int x1 = rect.x;
    int x2 = rect.x + block;

    // Background colors
    Color bg1(0.0, 0.0, 0.0);
    Color bg2(1.0, 1.0, 1.0);

    // The component will be adjusted
    double delta = 1.0 / (double)(rect.height - 1);
    double original = m_components[m_index];

    // Draw vertically
    m_components[m_index] = 1.0;
    bool swap = false;
    int swapPos = 0;
    int y;

    for(y = rect.y; y < yEnd; y++, swapPos++)
    {
        // Time to swap?
        if(swapPos == block)
        {
            swapPos = 0;
            swap = !swap;
        }

        // Color to draw
        Color c;
        m_convertor->ToColor(c, m_components);

        wxPen pn1(c.Blend(bg1));
        wxPen pn2(c.Blend(bg2));

        if(swap)
        {
            dc.SetPen(pn1);
            dc.DrawLine(x1, y, x2, y);
            dc.SetPen(pn2);
            dc.DrawLine(x2, y, xEnd, y);
        }
        else
        {
            dc.SetPen(pn2);
            dc.DrawLine(x1, y, x2, y);
            dc.SetPen(pn1);
            dc.DrawLine(x2, y, xEnd, y);
        }

        // Next color
        m_components[m_index] -= delta;
    }

    // Change it back so everything works right
    m_components[m_index] = original;

    // Position item
    y = rect.y + (int)((1.0 - original) * (double)rect.height);
  
    wxPoint slider1[4];
    wxPoint slider2[4];

    slider1[0].x = rect.x;
    slider1[0].y = y - (SLIDER_Y / 2);
  
    slider1[1].x = rect.x;
    slider1[1].y = y + (SLIDER_Y / 2);
  
    slider1[2].x = rect.x + SLIDER_X;
    slider1[2].y = y;
  
    slider1[3].x = slider1[0].x;
    slider1[3].y = slider1[0].y;
  
    slider2[0].x = rect.x + rect.width - 1;
    slider2[0].y = y - (SLIDER_Y / 2);
  
    slider2[1].x = rect.x + rect.width - 1;
    slider2[1].y = y + (SLIDER_Y / 2);
  
    slider2[2].x = rect.x + rect.width - SLIDER_X;
    slider2[2].y = y;
  
    slider2[3].x = slider2[0].x;
    slider2[3].y = slider2[0].y;

    // Draw sliders
    dc.SetPen(*wxBLACK_PEN);
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.DrawPolygon(4, slider1);
  
    dc.SetPen(*wxWHITE_PEN);
    dc.SetBrush(*wxBLACK_BRUSH);
    dc.DrawPolygon(4, slider2);
  
    // Done with our drawing
    dc.SetPen(wxNullPen);
    dc.SetBrush(wxNullBrush);

    dc.DestroyClippingRegion();
}


BEGIN_EVENT_TABLE(ColorBarWidget, wxControl)
    EVT_LEFT_DOWN(ColorBarWidget::OnLeftDown)
    EVT_LEFT_UP(ColorBarWidget::OnLeftUp)
    EVT_MOTION(ColorBarWidget::OnMotion)
    EVT_MOUSE_CAPTURE_LOST(ColorBarWidget::OnCaptureLost)
    EVT_PAINT(ColorBarWidget::OnPaint)
END_EVENT_TABLE()


// ColorPanel
//-----------------------------------------------------------------------------

// Constructor
ColorPanel::ColorPanel(wxWindow* parent, ColorConvertor* convertor) :
        wxPanel(parent)
{
    wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
    size_t count = convertor->GetComponentCount();

    size_t pos;
    for(pos = 0; pos < count; pos++)
    {
        // Color bar
        ColorBarWidget* widget = new ColorBarWidget(this,
                                                    wxID_ANY,
                                                    wxDefaultPosition,
                                                    wxSize(BAR_X, BAR_Y),
                                                    convertor,
                                                    pos);
        // Add to the main sizer
        sizer->Add(widget, wxSizerFlags(1).Expand().Border());

        // Add the entry
        m_entries.Add(widget);
    }

    SetSizerAndFit(sizer);
}

// Set the color 
void ColorPanel::SetColor(const Color& color)
{
    size_t pos, count;
    for(pos = 0, count = m_entries.GetCount(); pos < count; pos++)
    {
        m_entries[pos]->SetColor(color);
    }
}

// One of the bars has changed
void ColorPanel::OnBarChanged(ColorBarEvent& evt)
{
    ColorBarWidget* sender = (ColorBarWidget*)evt.GetEventObject();
    const double* components = sender->GetComponents();

    size_t pos, count;
    for(pos = 0, count = m_entries.GetCount(); pos < count; pos++)
    {
        if(m_entries[pos] != sender)
        {
            m_entries[pos]->SetComponents(components);
        }
    }

    // Allow event to go higher
    evt.Skip();
}

// Event table
BEGIN_EVENT_TABLE(ColorPanel, wxPanel)
    EVT_COLORBAR_CHANGING(wxID_ANY, ColorPanel::OnBarChanged)
    EVT_COLORBAR_CHANGED(wxID_ANY, ColorPanel::OnBarChanged)
END_EVENT_TABLE()


// ColorNotebook
//-----------------------------------------------------------------------------

// Constructor
ColorNotebook::ColorNotebook(wxWindow* parent) : wxNotebook(parent, wxID_ANY)
{
    ColorConvertor* conv;
    size_t pos = 0;

    conv = ColorConvertor::GetConvertor(pos++);
    while(conv)
    {
        // New item
        ColorPanel* panel = new ColorPanel(this, conv);
        AddPage(panel, conv->GetName());

        // Next convertor
        conv = ColorConvertor::GetConvertor(pos++);
    }
}

// Set a color
void ColorNotebook::SetColor(const Color& color)
{
    size_t pos, count;
    for(pos = 0, count = GetPageCount(); pos < count; pos++)
    {
        ColorPanel* panel = (ColorPanel*)GetPage(pos);

        panel->SetColor(color);
    }
}

// A bar changed in one of the pages (the active page)
void ColorNotebook::OnBarChanged(ColorBarEvent& evt)
{
    // Set other pages to the same color
    size_t pos, count;
    int cur = GetSelection();

    for(pos = 0, count = GetPageCount(); pos < count; pos++)
    {
        if(pos != (size_t)cur)
        {
            ColorPanel* panel = (ColorPanel*)GetPage(pos);
            panel->SetColor(evt.GetColor());
        }
    }
    // Allow event to propogate to the parent window so it can handle
    // updating preview color and copying code
    evt.Skip();
}

// Event table
BEGIN_EVENT_TABLE(ColorNotebook, wxNotebook)
    EVT_COLORBAR_CHANGING(wxID_ANY, ColorNotebook::OnBarChanged)
    EVT_COLORBAR_CHANGED(wxID_ANY, ColorNotebook::OnBarChanged)
END_EVENT_TABLE()


// ColorDialog
//-----------------------------------------------------------------------------


ColorDialog::ColorDialog(wxWindow* parent, const ColorItem* item)
{
    // Get color
    if(item)
        m_item = *item;
    
    // Create dialog
    if(!wxDialog::Create(parent, -1, _("Color Editor"), wxDefaultPosition,
        wxSize(900, 900), wxDEFAULT_DIALOG_STYLE))
    {
        return;
    }
    
    // Initialize stuff
    CreateWidgets();
    CenterOnParent(wxBOTH); 
    
    InitBasicColors();
    LoadCustomColors();

    // Set color
    m_previewColor->SetColor(m_item);
    m_notebook->SetColor(m_item);

    CopyCode();
}
  
// Destructor
ColorDialog::~ColorDialog()
{
}

// Get color item
const ColorItem& ColorDialog::GetColorItem() const
{
    return m_item;
}

// User pressed OK
void ColorDialog::OnOK(wxCommandEvent& WXUNUSED(evt))
{
    SaveCustomColors();
    EndModal(wxID_OK);
}
  
// Color cell selected
void ColorDialog::OnColorCellSelected(ColorCellEvent& evt)
{
    if(evt.GetEventObject() != m_previewColor)
    {
        m_item = evt.GetColor();
        m_previewColor->SetColor(m_item);
        m_notebook->SetColor(m_item);
        CopyCode();
    }
}

// Color cell alt-selected
void ColorDialog::OnColorCellAltSelected(ColorCellEvent& evt)
{
    int i;
    for(i = 0; i < 16; i++)
    {
        if(m_customColors[i] == evt.GetEventObject())
        {
            // It was a custom color
            m_customColors[i]->SetColor(m_item);
        }
    }
}

// Color bar changing
void ColorDialog::OnColorBarChanging(ColorBarEvent& evt)
{
    m_item = evt.GetColor();
    m_previewColor->SetColor(m_item);
}

// Color bar changed
void ColorDialog::OnColorBarChanged(ColorBarEvent& evt)
{
    // Do 'changing' stuff and copy code
    OnColorBarChanging(evt);
    CopyCode();
}
  
// Create the dialog items
void ColorDialog::CreateWidgets()
{
    wxBusyCursor busy;
    int i;
  
    // Color stuff on left
    wxBoxSizer* colorSizer = new wxBoxSizer(wxVERTICAL);

    // Basic colors
    colorSizer->Add(new wxStaticText(this, -1, _("Basic Colors")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));

    wxGridSizer* basicSizer = new wxGridSizer(8, CELLSPACE, CELLSPACE);

    for(i = 0; i < 48; i++)
    {
        m_basicColors[i] = new ColorCellWidget(this, wxID_ANY, wxDefaultPosition, wxSize(CELLSIZE, CELLSIZE));
        basicSizer->Add(m_basicColors[i], 0);
    }

    colorSizer->Add(basicSizer, wxSizerFlags(0).Center().Border(wxBOTTOM));
      
    // Custom colors
    colorSizer->Add(new wxStaticText(this, -1, _("Custom Colors")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));

    wxGridSizer* customSizer = new wxGridSizer(8, CELLSPACE, CELLSPACE);

    for(i = 0; i < 16; i++)
    {
        m_customColors[i] = new ColorCellWidget(this, wxID_ANY, wxDefaultPosition, wxSize(CELLSIZE, CELLSIZE));
        customSizer->Add(m_customColors[i], 0);
    }

    colorSizer->Add(customSizer, wxSizerFlags(0).Center().Border(wxBOTTOM));
    
    // Preview
    m_previewColor = new ColorCellWidget(this, wxID_ANY, wxDefaultPosition, wxSize(8 * CELLSIZE + 7 * CELLSPACE, PREVIEW_Y));
    colorSizer->Add(m_previewColor, wxSizerFlags(1).Center());

    // Color notebook
    m_notebook = new ColorNotebook(this);
  
    // Top sizer
    wxBoxSizer* topSizer = new wxBoxSizer(wxHORIZONTAL);
  
    topSizer->Add(colorSizer, wxSizerFlags(0).Expand().Border(wxRIGHT, AREASPACE_X));
    topSizer->Add(m_notebook, wxSizerFlags(1).Expand());
  
    // Main sizer
    wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  
    mainSizer->Add(topSizer, wxSizerFlags(1).Expand().DoubleBorder());
 
    // Buttons
    wxSizer* buttons = CreateSeparatedButtonSizer(wxOK | wxCANCEL);
    if(buttons)
    {
        mainSizer->Add(buttons, wxSizerFlags(0).Expand().DoubleBorder(wxLEFT | wxRIGHT | wxBOTTOM));
    }
  
    // Apply sizers and fit size
    SetSizerAndFit(mainSizer);
} 
  
// Initialize basic colors
void ColorDialog::InitBasicColors()
{
    int x = 0;
  
    #define ADDCOLOR(r, g, b) m_basicColors[x++]->SetColor(Color(r, g, b))
  
    // Basics
    ADDCOLOR(1.0, 0.0, 0.0); // Red
    ADDCOLOR(0.0, 1.0, 0.0); // Green
    ADDCOLOR(0.0, 0.0, 1.0); // Blue
    ADDCOLOR(1.0, 1.0, 0.0); // Yellow
    ADDCOLOR(0.0, 1.0, 1.0); // Cyan
    ADDCOLOR(1.0, 0.0, 1.0); // Magenta
    ADDCOLOR(0.0, 0.0, 0.0); // Black
    ADDCOLOR(1.0, 1.0, 1.0); // White
    
    // Grays
    ADDCOLOR(0.11, 0.11, 0.11);
    ADDCOLOR(0.22, 0.22, 0.22);
    ADDCOLOR(0.33, 0.33, 0.33);
    ADDCOLOR(0.44, 0.44, 0.44);
    ADDCOLOR(0.55, 0.55, 0.55);
    ADDCOLOR(0.66, 0.66, 0.66);
    ADDCOLOR(0.77, 0.77, 0.77);
    ADDCOLOR(0.88, 0.88, 0.88);
    
    // Others (From colors.inc)
    
    // Reds
    ADDCOLOR(1.0, 0.498039, 0.0);
    ADDCOLOR(0.556863, 0.137255, 0.137255);
    ADDCOLOR(0.55, 0.09, 0.09);
    ADDCOLOR(0.309804, 0.184314, 0.309804);
    ADDCOLOR(0.8, 0.196078, 0.6);
    ADDCOLOR(0.737255, 0.560784, 0.560784);
    ADDCOLOR(0.917647, 0.678431, 0.917647);
    ADDCOLOR(0.8, 0.498039, 0.196078);
    
    // Greens
    ADDCOLOR(0.184314, 0.309804, 0.184314);
    ADDCOLOR(0.309804, 0.309804, 0.184314);
    ADDCOLOR(0.6, 0.196078, 0.8);
    ADDCOLOR(0.137255, 0.556863, 0.137255);
    ADDCOLOR(0.13, 0.37, 0.31);
    ADDCOLOR(0.258824, 0.435294, 0.258824);
    ADDCOLOR(0.0, 1.0, 0.498039);
    ADDCOLOR(0.32, 0.49, 0.46);

    // Blues
    ADDCOLOR(0.439216, 0.858824, 0.576471);
    ADDCOLOR(0.372549, 0.623529, 0.623529);
    ADDCOLOR(0.00, 0.00, 0.61);
    ADDCOLOR(0.0, 0.498039, 1.0);
    ADDCOLOR(0.196078, 0.6, 0.8);
    ADDCOLOR(0.137255, 0.419608, 0.556863);
    ADDCOLOR(0.22, 0.69, 0.87);
    ADDCOLOR(0.35, 0.35, 0.67);

    // Browns
    ADDCOLOR(0.647059, 0.164706, 0.164706);
    ADDCOLOR(0.96, 0.80, 0.69);
    ADDCOLOR(0.92, 0.78, 0.62);
    ADDCOLOR(0.35, 0.16, 0.14);
    ADDCOLOR(0.36, 0.25, 0.20);
    ADDCOLOR(0.623529, 0.623529, 0.372549);
    ADDCOLOR(0.42, 0.26, 0.15);
    ADDCOLOR(0.36, 0.20, 0.09);
    
    #undef ADDCOLOR
} 

// Load custom colors
void ColorDialog::LoadCustomColors()
{
    // Init to white
    int pos;
  
    for(pos = 0; pos < 16; pos++)
        m_customColors[pos]->SetColor(Color(1.0, 1.0, 1.0));
  
    // Configuration object
    wxConfigBase* config = wxGetApp().GetConfig();
    if(!config)
    {
        return;
    }
    
    // Set path
    wxConfigPathChanger pathChanger(config, wxT("/CustomColors/"));
  
    // Read colors
    for(pos = 0; pos < 16; pos++)
    {
        wxString key, value;
    
        // Read value
        key.Printf(wxT("Color%d"), pos + 1);
        if(config->Read(key, &value))
        {
            // Value found
            wxStringTokenizer tokens(value, wxT(","));
            double r, g, b, f, t;
      
            // Decode colors
            r = g = b = f = t = 0.0;
      
            tokens.GetNextToken().ToDouble(&r);
            tokens.GetNextToken().ToDouble(&g);
            tokens.GetNextToken().ToDouble(&b);
            tokens.GetNextToken().ToDouble(&f);
            tokens.GetNextToken().ToDouble(&t);
      
            // Set custom color
            m_customColors[pos]->SetColor(Color(r, g, b, f, t));
        }
    }
}
  
// Save custom colors
void ColorDialog::SaveCustomColors()
{
    int pos;
  
    // Configuration object
    wxConfigBase* config = wxGetApp().GetConfig();
    if(!config)
    {
        return;
    }
    
    // Set path
    wxConfigPathChanger pathChanger(config, wxT("/CustomColors/"));
  
    // Write colors
    for(pos = 0; pos < 16; pos++)
    {
        wxString key, value;
    
        // Key name
        key.Printf(wxT("Color%d"), pos + 1);
    
        // Values
        Color c = m_customColors[pos]->GetColor();
        value.Printf(wxT("%f,%f,%f,%f,%f"), c.GetRed(), c.GetGreen(), c.GetBlue(),
            c.GetFilter(), c.GetTransmit());
        
        config->Write(key, value);
    }
} 

// Copy code to clipboard
void ColorDialog::CopyCode()
{
    Options& opt = wxGetApp().GetOptions();
  
    if(opt.GetAutoCopyCode())
    {
        m_item.CopyCode();
    }
} 
  
// Event table
BEGIN_EVENT_TABLE(ColorDialog, wxDialog)
    EVT_BUTTON(wxID_OK, ColorDialog::OnOK)
    EVT_COLORCELL_SELECTED(wxID_ANY, ColorDialog::OnColorCellSelected)
    EVT_COLORCELL_ALTSELECTED(wxID_ANY, ColorDialog::OnColorCellAltSelected)
    EVT_COLORBAR_CHANGING(wxID_ANY, ColorDialog::OnColorBarChanging)
    EVT_COLORBAR_CHANGED(wxID_ANY, ColorDialog::OnColorBarChanged)
END_EVENT_TABLE() 
  
