/*
  Project:      POVColor
  File:         coldlg.h
  Purpose:      Color dialog
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __POVCOLOR_COLDLG_H__
#define __POVCOLOR_COLDLG_H__


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/dialog.h>
#include <wx/panel.h>
#include <wx/notebook.h>
#include <wx/control.h>
#include <wx/dynarray.h>

#include "color.h"
#include "colorcell.h"

// Forward declarations
class WXDLLEXPORT wxWindow;
class WXDLLEXPORT wxStaticText;

class ColorConvertor;


// ColorBarWidget - A widget to draw a color bar with one component gradient
//-----------------------------------------------------------------------------

// Events for a color bar
extern const wxEventType EVENT_COLORBAR_CHANGING;
extern const wxEventType EVENT_COLORBAR_CHANGED;

class ColorBarEvent : public wxCommandEvent
{
public:
    ColorBarEvent(wxEventType type, wxWindowID id, const Color& color);
    ColorBarEvent(const ColorBarEvent& copy);

    ~ColorBarEvent();

    wxEvent* Clone() const { return new ColorBarEvent(*this); }

    const Color& GetColor() const { return m_color; }

private:
    // No assign
    ColorBarEvent& operator=(const ColorBarEvent& rhs);

    Color m_color;
};

typedef void (wxEvtHandler::*ColorBarEventFunction)(ColorBarEvent&);

#define ColorBarEventHandler(func) \
    (wxObjectEventFunction)(wxEventFunction)(wxCommandEventFunction) \
    wxStaticCastEvent(ColorBarEventFunction, &func)

#define EVT_COLORBAR_CHANGING(id, fn) \
    DECLARE_EVENT_TABLE_ENTRY(EVENT_COLORBAR_CHANGING, id, wxID_ANY, \
    ColorBarEventHandler(fn), (wxObject*)NULL),

#define EVT_COLORBAR_CHANGED(id, fn) \
    DECLARE_EVENT_TABLE_ENTRY(EVENT_COLORBAR_CHANGED, id, wxID_ANY, \
    ColorBarEventHandler(fn), (wxObject*)NULL),

// Color bar widget
class ColorBarWidget : public wxControl
{
public:
    // Constructor
    ColorBarWidget(wxWindow* parent,
                   wxWindowID id,
                   const wxPoint& pos,
                   const wxSize& size,
                   ColorConvertor* convertor,
                   size_t index);
    // Destructor
    ~ColorBarWidget();

    // Set the color
    void SetColor(const Color& color);

    // Get color
    const Color& GetColor() const { return m_color; }
    
    // Set/Get the components
    void SetComponents(const double* components);
    const double* GetComponents() const { return m_components; }

    // Events
    void OnLeftDown(wxMouseEvent& evt);
    void OnLeftUp(wxMouseEvent& evt);
    void OnMotion(wxMouseEvent& evt);
    void OnCaptureLost(wxMouseCaptureLostEvent& evt);
    void OnPaint(wxPaintEvent& evt);

private:
    // No copy/assign
    ColorBarWidget(const ColorBarWidget& copy);
    ColorBarWidget& operator=(const ColorBarWidget& rhs);

    Color m_color;
    ColorConvertor* m_convertor;
    size_t m_index;
    double* m_components;


DECLARE_EVENT_TABLE();
};


// ColorPanel for a given color type
//-----------------------------------------------------------------------------
WX_DEFINE_ARRAY_PTR(ColorBarWidget*, ColorPanelEntryList);

class ColorPanel : public wxPanel
{
public:
    ColorPanel(wxWindow* parent, ColorConvertor* convertor);

    void SetColor(const Color& color);
    void OnBarChanged(ColorBarEvent& evt);

private:
    // No copy/assign
    ColorPanel(const ColorPanel& copy);
    ColorPanel& operator=(const ColorPanel& rhs);

    // List of color bar entries
    ColorPanelEntryList m_entries;

DECLARE_EVENT_TABLE()
};


// ColorNotebook
//-----------------------------------------------------------------------------
class ColorNotebook : public wxNotebook
{
public:
    ColorNotebook(wxWindow* parent);

    void SetColor(const Color& color);
    void OnBarChanged(ColorBarEvent& evt);

private:
    // No copy/assign
    ColorNotebook(const ColorNotebook& copy);
    ColorNotebook& operator=(const ColorNotebook& rhs);

DECLARE_EVENT_TABLE()
};


// ColorDialog - Editor for ColorItem
//-----------------------------------------------------------------------------
class ColorDialog : public wxDialog
{
public:
    // ctor/dtor (no default ctor)
    ColorDialog(wxWindow* parent, const ColorItem* item = NULL);
    ~ColorDialog();
    
    // Return selected color
    const ColorItem& GetColorItem() const;
    
    // Button events
    void OnOK(wxCommandEvent& evt);
    
    // Paint
    void OnPaint(wxPaintEvent& evt);
    
    // Color cell selected
    void OnColorCellSelected(ColorCellEvent& evt);
    void OnColorCellAltSelected(ColorCellEvent& evt);

    // Color bar changed
    void OnColorBarChanging(ColorBarEvent& evt);
    void OnColorBarChanged(ColorBarEvent& evt);

      
private:
    // No copy/assign
    ColorDialog(const ColorDialog& copy);
    ColorDialog& operator=(const ColorDialog& copy);

    // Create dialog controls
    void CreateWidgets();
    
    // Initialize colors
    void InitBasicColors();
    void LoadCustomColors();
    void SaveCustomColors();
    
    // Copy item code
    void CopyCode();
    
    // Current color
    ColorItem m_item;
    
    // Color cells
    ColorCellWidget* m_basicColors[48];
    ColorCellWidget* m_customColors[16];
    ColorCellWidget* m_previewColor;

    // Color notebook
    ColorNotebook* m_notebook;
    
DECLARE_EVENT_TABLE();    
};

#endif // Include once wrapper

