/*
  Project:      POVColor
  File:         color.cpp
  Purpose:      Color implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/math.h>
#include <wx/dc.h>
#include <wx/icon.h>
#include <wx/dataobj.h>
#include <wx/module.h>
#include <wx/clipbrd.h>

#include "app.h"
#include "art.h"
#include "color.h"
#include "options.h"
#include "util.h"
#include "coldlg.h"
#include "input.h"
#include "output.h"


// Color
//-----------------------------------------------------------------------------

// Constructors
Color::Color() : m_r(0.0), m_g(0.0), m_b(0.0), m_f(0.0), m_t(0.0)
{
}

Color::Color(const Color& copy) :
    m_r(copy.m_r), m_g(copy.m_g), m_b(copy.m_b), m_f(copy.m_f), m_t(copy.m_t)
{
}

Color::Color(double r, double g, double b, double f, double t) :
    m_r(r), m_g(g), m_b(b), m_f(f), m_t(t)
{
}

// Destructor
Color::~Color()
{
}

// Assignment
Color& Color::operator=(const Color& copy)
{
    if(this != &copy)
    {
        m_r = copy.m_r;
        m_g = copy.m_g;
        m_b = copy.m_b;
        m_f = copy.m_f;
        m_t = copy.m_t;
    }

    return *this;
}

// Equality
bool Color::operator ==(const Color& rhs)
{
    // TODO: use some small value for epsilon and do a difference test
    return (m_r == rhs.m_r) && (m_g == rhs.m_g) && (m_b == rhs.m_b) &&
        (m_f == rhs.m_f) && (m_t == rhs.m_t);
}

// Blend a color
wxColour Color::Blend(const Color& back) const
{
    double fr, fg, fb;
    
    // The formula to use to blend appears to be quite well
    #define BLEND(back, front, f, t) ((back * front * f) + (back * t) + (front * (1.0 - f - t)))
    
    fr = clip(BLEND(back.m_r, m_r, m_f, m_t), 0.0, 1.0);
    fg = clip(BLEND(back.m_g, m_g, m_f, m_t), 0.0, 1.0);
    fb = clip(BLEND(back.m_b, m_b, m_f, m_t), 0.0, 1.0);
    
    return GammaCorrect((int)(fr * 255.0), (int)(fg * 255.0), (int)(fb * 255.0));
}
  
// Prepare color string
wxString Color::PrepareColorString() const
{
    wxString clr;
    wxString result;
    
    // Spelling
    Options& opt = wxGetApp().GetOptions();
    
    clr = opt.GetColorKeyword();
    
    if(m_f == 0.0 && m_t == 0.0)
    {
        // rgb
        result.Printf(wxT("%s rgb <%.4f, %.4f, %.4f>"),
            clr.c_str(), m_r, m_g, m_b);
    }
    else if(m_f == 0.0)
    {
        // rgbt
        result.Printf(wxT("%s rgbt <%.4f, %.4f, %.4f, %.4f>"),
            clr.c_str(), m_r, m_g, m_b, m_t);
    }
    else if(m_t == 0.0)
    {
        // rgbf
        result.Printf(wxT("%s rgbf <%.4f, %.4f, %.4f, %.4f>"),
            clr.c_str(), m_r, m_g, m_b, m_f);
    }
    else
    {
        // rgbft
        result.Printf(wxT("%s rgbft <%.4f, %.4f, %.4f, %.4f, %.4f>"),
            clr.c_str(), m_r, m_g, m_b, m_f, m_t);
    }
        
    return result;
} 

// Gamma correction variables
double Color::sm_gammaValue = -0.1;
int Color::sm_gammaTable[256];

// Do gamma correction
wxColour Color::GammaCorrect(int r, int g, int b)
{
    // Get options
    Options& opt = wxGetApp().GetOptions();
    double gamma;
    
    // Gamma correction enabled
    if(opt.GetGammaCorrectionEnabled() == false)
        return wxColour(r, g, b);
    
    // Get dispaly gamma
    gamma = opt.GetDisplayGamma();
        
    // Rebuild gamma tables?
    if(gamma != sm_gammaValue)
    {
        int pos;
        for(pos = 0; pos < 256; pos++)
        {
            sm_gammaTable[pos] = (int)(pow((double)pos / 255.0, 1.0 / gamma) * 255.0);
        }
        
        sm_gammaValue = gamma;
    }
        
    // Return color (transmit/filter ignored)
    return wxColour(sm_gammaTable[r], sm_gammaTable[g], sm_gammaTable[b]);
}
  

// ColorItem
//-----------------------------------------------------------------------------

// Constructor
ColorItem::ColorItem()
{
}
  
ColorItem::ColorItem(const ColorItem& copy) :
    PovItem(copy),
    Color(copy)
{
}
  
ColorItem::ColorItem(const Color& copy) : Color(copy)
{
}

// Destructor
ColorItem::~ColorItem()
{
}

// Assignment
ColorItem& ColorItem::operator =(const ColorItem& copy)
{
    if(this != &copy)
    {
        PovItem::operator=(copy);
        Color::operator=(copy);
    }
    
    return *this;
}

ColorItem& ColorItem::operator =(const Color& copy)
{
    if(this != &copy)
    {
        Color::operator=(copy);
    }
    
    return *this;
}
  
// Get code
void ColorItem::CopyCode() const
{
    if(wxTheClipboard->Open())
    {
        wxTextDataObject* text = new wxTextDataObject(PrepareColorString());
    
        wxTheClipboard->SetData(text);
        wxTheClipboard->Close();
    }
}

// Clone
PovItem* ColorItem::Clone() const
{
    return new ColorItem(*this);
}

// Load color item
bool ColorItem::Load(PovInput& in)
{
    Token tok;
    double vector[5];
  
    // Get first token
    tok = in.GetNextToken();
  
    // Skip 'color/colour' token
    if(tok.GetType() == Token::TOKEN_KEYWORD_COLOR)
        tok = in.GetNextToken();
    
    if(tok.GetType() == Token::TOKEN_KEYWORD_RGB)
    {
        // RGB color value
        if(in.ReadVector(3, vector))
        {
            SetColor(vector[0], vector[1], vector[2]);
            return true;
        }
        else
        {
            return false;
        }
    }
    else if(tok.GetType() == Token::TOKEN_KEYWORD_RGBF)
    {
        // RGBF color value
        if(in.ReadVector(4, vector))
        {
            SetColor(vector[0], vector[1], vector[2], vector[3]);
            return true;
        }
        else
        {
            return false;
        }   
    }
    else if(tok.GetType() == Token::TOKEN_KEYWORD_RGBT)
    {
        // RGBT color value
        if(in.ReadVector(4, vector))
        {
            SetColor(vector[0], vector[1], vector[2], 0.0, vector[3]);
            return true;
        }
        else
        {
            return false;
        }
    }
    else if(tok.GetType() == Token::TOKEN_KEYWORD_RGBFT)
    {
        // RGBFT color value
        if(in.ReadVector(5, vector))
        {
            SetColor(vector[0], vector[1], vector[2], vector[3], vector[4]);
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        // Literal values
        int index = 0;
        bool found;
    
    
        // Keep looking until we find an invalid keyword
        vector[0] = vector[1] = vector[2] = vector[3] = vector[4] = 0.0;
        found = false;
    
        while(true)
        {
            if(tok.GetType() == Token::TOKEN_KEYWORD_RED)
            {
                index = 0;
            }
            else if(tok.GetType() == Token::TOKEN_KEYWORD_GREEN)
            {
                index = 1;
            }
            else if(tok.GetType() == Token::TOKEN_KEYWORD_BLUE)
            {
                index = 2;
            }
            else if(tok.GetType() == Token::TOKEN_KEYWORD_FILTER)
            {
                index = 3;
            }
            else if(tok.GetType() == Token::TOKEN_KEYWORD_TRANSMIT)
            {
                index = 4;
            }
            else
            {
                // Invalid token
                in.UngetToken();
        
                if(found)
                {
                    SetColor(vector[0], vector[1], vector[2], vector[3], vector[4]);
                    return true;
                }
                else
                {
                  return false;
                }
            }

            // Next item should be a number
            tok = in.GetNextToken();
            if(tok.GetType() != Token::TOKEN_NUMBER)
            {
                in.UngetToken();
                return false;
            }
          
            vector[index] = tok.GetNumber();
            found = true;
        
            // Next token
            tok = in.GetNextToken();
        }
      
        // Should never get here
        return false;
    }
}
  
// Save color item
bool ColorItem::Save(PovOutput& out, long indent) const
{
    wxString result, name;
  
    // Name
    name = GetName();
    
    // Indent
    out.WriteIndent(indent);
  
    // Format and write the string
    result.Printf(wxT("#declare %s = %s;\n"),
        name.c_str(), PrepareColorString().c_str());
    out.WriteString(result);
      
    return (out.GetLastError() != wxSTREAM_WRITE_ERROR);    
}
  
// Edit color item
bool ColorItem::Edit(wxWindow* parent)
{
    ColorDialog dlg(parent, this);
  
    if(dlg.ShowModal() == wxID_OK)
    {
        *this = dlg.GetColorItem();
        return true;
    }
    else
    {
        return false;
    }
}
  
// Draw color item
void ColorItem::Draw(wxDC& dc, const wxRect& rect) const
{
    int block;
    int x1, x2, y1, y2;
    int xEnd, yEnd;
    bool xSwap, ySwap;
  
    // Determine block size
    if(rect.width < rect.height)
        block = rect.width / 2;
    else
        block = rect.height / 2;
        
    // Ending
    xEnd = rect.x + rect.width - 1;
    yEnd = rect.y + rect.height - 1;
    
    // Drawing brushes
    wxBrush br1(Blend(Color(0.0, 0.0, 0.0)));
    wxBrush br2(Blend(Color(1.0, 1.0, 1.0)));
    
    // No pen
    dc.SetPen(*wxTRANSPARENT_PEN);
    
    // Horizontal pass
    xSwap = false;
    for(x1 = rect.x; x1 <= xEnd; x1 += block, xSwap = !xSwap)
    {
        // Ending X point for this block
        x2 = x1 + block - 1;
        if(x2 > xEnd)
            x2 = xEnd;
        
        // Vertical pass
        ySwap = false;
        for(y1 = rect.y; y1 < yEnd; y1 += block, ySwap = !ySwap)
        {
            // Ending Y point for this block
            y2 = y1 + block - 1;
            if(y2 > yEnd)
                y2 = yEnd;
        
            // Which color to draw
            if(xSwap == ySwap)
                dc.SetBrush(br1);
            else
                dc.SetBrush(br2);
        
            // Draw the box
            dc.DrawRectangle(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
        }
    }
  
    // Done
    dc.SetBrush(wxNullBrush);
    dc.SetPen(wxNullPen);
}

// Draw the icon
void ColorItem::DrawIcon(wxDC& dc, const wxRect& rect) const
{
    wxBitmap bmp;

    if(IsEnabled())
    {
        bmp = Art::GetBitmap(wxT("colore"));
    }
    else
    {
        bmp = Art::GetBitmap(wxT("colord"));
    }

    if(bmp.IsOk())
    {
        dc.DrawBitmap(bmp, rect.x, rect.y, true);
    }
}

  
  
