/*
  Project:      POVColor
  File:         color.h
  Purpose:      Color item
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __POVCOLOR_COLOR_H__
#define __POVCOLOR_COLOR_H__


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/colour.h>

#include "item.h"
#include "util.h"


// Color - A general color object
//-----------------------------------------------------------------------------
class Color
{
public:
    // Constructors
    Color();
    Color(const Color& copy);
    Color(double r, double g, double b, double f = 0.0, double t = 0.0); // RGB

    // Destructor
    ~Color();

    // Assignment
    Color& operator =(const Color&rhs);
    
    // Red
    void SetRed(double r) { m_r = clip(r, 0.0, 1.0); }
    double GetRed() const { return m_r; }
            
    // Green
    void SetGreen(double g) { m_g = clip(g, 0.0, 1.0); }
    double GetGreen() const { return m_g; }
        
    // Blue
    void SetBlue(double b) { m_b = clip(b, 0.0, 1.0); }
    double GetBlue() const { return m_b; }
    
    // Filter
    void SetFilter(double f) { m_f = clip(f, 0.0, 1.0); }
    double GetFilter() const { return m_f; }
    
    // Transmit
    void SetTransmit(double t) { m_t = clip(t, 0.0, 1.0);  }
    double GetTransmit() const { return m_t; }
        
    // All
    void SetColor(double r, double g, double b, double f = 0.0, double t = 0.0)
    {
        m_r = clip(r, 0.0, 1.0);
        m_g = clip(g, 0.0, 1.0);
        m_b = clip(b, 0.0, 1.0);
        m_f = clip(f, 0.0, 1.0);
        m_t = clip(t, 0.0, 1.0);
    }
    void GetColor(double& r, double& g, double& b, double& f, double& t) const
    {
        r = m_r;
        g = m_g;
        b = m_b;
        f = m_f;
        t = m_t;    
    }
    
    // Equality
    bool operator==(const Color& rhs);
    
    // Blend with a certain background (background filter/transmit is ignored)
    wxColour Blend(const Color& back) const;
    
    // Prepare a POVRay color string
    wxString PrepareColorString() const;
  
private:
    // Color information
    double m_r, m_g, m_b, m_f, m_t;
    
    // Return gamma corrected color
    static wxColour GammaCorrect(int r, int g, int b);  

    // Gamma table
    static double sm_gammaValue;
    static int sm_gammaTable[256];
  
};
  

// ColorItem - An item that is a color
//-----------------------------------------------------------------------------
class ColorItem : public PovItem, public Color
{
public:
    // Constructors
    ColorItem();
    ColorItem(const ColorItem& copy);
    ColorItem(const Color& copy);
    
    // Destructor
    ~ColorItem();
    
    // Assignment
    ColorItem& operator =(const ColorItem& rhs);
    ColorItem& operator =(const Color& rhs);    
    
    
    void CopyCode() const;
    PovItem* Clone() const;
    bool Load(PovInput& in);
    bool Save(PovOutput& out, long indent) const;
    bool Edit(wxWindow* parent);
    void Draw(wxDC& dc, const wxRect& rect) const;
    void DrawIcon(wxDC& dc, const wxRect& rect) const;
};


#endif // Include once wrapper
