/*
  Project:      POVColor
  File:         colorconv.cpp
  Purpose:      Color convertor implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <math.h>

#include "colorconv.h"
#include "color.h"
#include "util.h"


// ColorConverter
//-----------------------------------------------------------------------------
ColorConvertor::ColorConvertor()
{
}

ColorConvertor::~ColorConvertor()
{
}


// RGBColorConvertor
//-----------------------------------------------------------------------------
class RGBColorConvertor : public ColorConvertor
{
public:
    wxString GetName() const
    {
        return wxT("RGB");
    }

    size_t GetComponentCount() const
    {
        return 5;
    }

    void FromColor(const Color& color, double* components) const
    {
        color.GetColor(components[0], components[1], components[2], components[3], components[4]);
    }

    void ToColor(Color& color, double* components) const
    {
        color.SetColor(components[0], components[1], components[2], components[3], components[4]);
    }
};

// HSLColorConvertor
//-----------------------------------------------------------------------------
class HSLColorConvertor : public ColorConvertor
{
public:
    wxString GetName() const
    {
        return wxT("HSL");
    }

    size_t GetComponentCount() const
    {
        return 5;
    }

    void FromColor(const Color& color, double* components) const
    {
        // Get color
        double r, g, b, f, t;
        color.GetColor(r, g, b, f, t);

        // Set filter, transmit
        components[3] = f;
        components[4] = t;

        // Convert
        double delta, lo, hi;
        double h, s, l;

        hi = max(r, max(g, b));
        lo = min(r, min(g, b));
        delta = hi - lo;

        l = 0.5 * (lo + hi);

        if(delta == 0.0)
        {
            h = 0.0;
            s = 0.0;
        }
        else
        {
            s = delta / ( (l <= 0.5) ? (lo + hi) : (2.0 - hi - lo) );

            if(r == hi)
            {
                h = ( g == lo ? 5.0 + (hi - b) / delta : 1.0 - (hi - g) / delta );
            }
            else if(g == hi)
            {
                h = ( b == lo ? 1.0 + (hi - r) / delta : 3.0 - (hi - b) / delta );
            }
            else
            {
                h = ( r == lo ? 3.0 + (hi - g) / delta : 5.0 - (hi - r) / delta );
            }

            h /= 6.0;
        }

        components[0] = h;
        components[1] = s;
        components[2] = l;

    }

    void ToColor(Color& color, double* components) const
    {
        // Get components
        double h, s, l, f, t;
        h = components[0];
        s = components[1];
        l = components[2];
        f = components[3];
        t = components[4];

        // Convert
        double frac, v, x, y, z;
        double r, g, b;

        v = (l <= 0.5) ? (l * (1.0 + s)) : (l + s - (l * s));

        if(s == 0.0 || h < 0.0)
        {
            color.SetColor(v, v, v, f, t);
            return;
        }

        frac = (6.0 * h - floor(6.0 * h));
        y = 2.0 * l - v;
        x = y + (v - y) * frac;
        z = v - (v - y) * frac;

        switch((int)(6.0 * h))
        {
            case 0:
            default:
            {
                r = v;
                g = x;
                b = y;
                break;
            }

            case 1:
            {
                r = z;
                g = v;
                b = y;
                break;
            }

            case 2:
            {
                r = y;
                g = v;
                b = x;
                break;
            }

            case 3:
            {
                r = y;
                g = z;
                b = v;
                break;
            }

            case 4:
            {
                r = x;
                g = y;
                b = v;
                break;
            }

            case 5:
            {
                r = v;
                g = y;
                b = z;
                break;
            }
        }

        color.SetColor(r, g, b, f, t);
    }
};

// HSVColorConvertor
//-----------------------------------------------------------------------------
class HSVColorConvertor : public ColorConvertor
{
public:
    wxString GetName() const
    {
        return wxT("HSV");
    }

    size_t GetComponentCount() const
    {
        return 5;
    }

    void FromColor(const Color& color, double* components) const
    {
        // Get color
        double r, g, b, f, t;
        color.GetColor(r, g, b, f, t);

        // Set filter, transmit
        components[3] = f;
        components[4] = t;

        // Convert
        double delta, lo, hi;
        double h, s, v;

        hi = max(r, max(g, b));
        lo = min(r, min(g, b));
        delta = hi - lo;

        v = hi;

        if(hi == 0.0)
        {
            h = 0.0;
            s = 0.0;
        }
        else
        {
            s = delta / hi;

            if(r == hi)
            {
                h = ( g == lo ? 5.0 + (hi - b) / delta : 1.0 - (hi - g) / delta );
            }
            else if(g == hi)
            {
                h = ( b == lo ? 1.0 + (hi - r) / delta : 3.0 - (hi - b) / delta );
            }
            else
            {
                h = ( r == lo ? 3.0 + (hi - g) / delta : 5.0 - (hi - r) / delta );
            }

            h /= 6.0;
        }

        components[0] = h;
        components[1] = s;
        components[2] = v;

    }

    void ToColor(Color& color, double* components) const
    {
        // Get components
        double h, s, v, f, t;
        h = components[0];
        s = components[1];
        v = components[2];
        f = components[3];
        t = components[4];

        // Convert
        double frac, m, n, k;
        double r, g, b;

        if(s == 0.0 || h < 0.0)
        {
            color.SetColor(v, v, v, f, t);
            return;
        }

        frac = (6.0 * h - floor(6.0 * h));
        m = v * (1.0 - s);
        n = v * (1.0 - s * frac);
        k = v * (1.0 - s * (1.0 - frac));

        switch((int)(6.0 * h))
        {
            case 0:
            default:
            {
                r = v;
                g = k;
                b = m;
                break;
            }

            case 1:
            {
                r = n;
                g = v;
                b = m;
                break;
            }

            case 2:
            {
                r = m;
                g = v;
                b = k;
                break;
            }

            case 3:
            {
                r = m;
                g = n;
                b = v;
                break;
            }

            case 4:
            {
                r = k;
                g = m;
                b = v;
                break;
            }

            case 5:
            {
                r = v;
                g = m;
                b = n;
                break;
            }
        }

        color.SetColor(r, g, b, f, t);
    }
};

// CMYKColorConvertor
//-----------------------------------------------------------------------------
class CMYKColorConvertor : public ColorConvertor
{
public:
    wxString GetName() const
    {
        return wxT("CMYK");
    }

    size_t GetComponentCount() const
    {
        return 6;
    }

    void FromColor(const Color& color, double* components) const
    {
        // Get color
        double r, g, b, f, t;
        color.GetColor(r, g, b, f, t);

        // Set filter, transmit
        components[4] = f;
        components[5] = t;

        // Convert
        double c, m, y, k;

        c = 1.0 - r;
        m = 1.0 - g;
        y = 1.0 - b;
        k = 1.0;

        if(c < k)
        {
            k = c;
        }
        
        if(m < k)
        {
            k = m;
        }

        if(y < k)
        {
            k = y;
        }

        if(k == 1.0)
        {
            c = 0.0;
            m = 0.0;
            y = 0.0;
        }
        else
        {
            c = (c - k) / (1.0 - k);
            m = (m - k) / (1.0 - k);
            y = (y - k) / (1.0 - k);
        }
        
        components[0] = c;
        components[1] = m;
        components[2] = y;
        components[3] = k;
    }

    void ToColor(Color& color, double* components) const
    {
        // Get components
        double c, m, y, k, f, t;
        c = components[0];
        m = components[1];
        y = components[2];
        k = components[3];
        f = components[4];
        t = components[5];

        // Convert
        double r, g, b;

        r = 1.0 - (c + k);
        g = 1.0 - (m + k);
        b = 1.0 - (y + k);

        color.SetColor(r, g, b, f, t);
    }
};


// Get color convertors
//-----------------------------------------------------------------------------
ColorConvertor* ColorConvertor::GetConvertor(size_t index)
{
    static RGBColorConvertor s_rgb;
    static HSLColorConvertor s_hsl;
    static HSVColorConvertor s_hsv;
    static CMYKColorConvertor s_cmyk;

    switch(index)
    {
        case 0:
            return &s_rgb;
            
        case 1:
            return &s_hsl;

        case 2:
            return &s_hsv;

        case 3:
            return &s_cmyk;

        default:
            return NULL;
    }
}

