/*
  Project:      POVColor
  File:         file.cpp
  Purpose:      Implementation of file object
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/string.h>
#include <wx/msgdlg.h>
#include <wx/filedlg.h>
#include <wx/wfstream.h>

#include "file.h"
#include "input.h"
#include "output.h"
#include "app.h"


// File
//-----------------------------------------------------------------------------

// Constructors
File::File() :
    m_modified(false)
{
}

// Copy file data, not name
File::File(const File& copy) :
    m_modified(false)
{
    size_t pos, count;
  
    count = copy.GetCount();
  
    if(count > 0)
    {
        for(pos = 0; pos < count; pos++)
        {
            m_items.Add(copy[pos].Clone());
        }
        m_modified = true;
    }
    else
    {
        // No data
        m_modified = false;
    }
}

// Destructor
File::~File()
{
    Clear();
}
  
// Get state
bool File::IsModified() const
{
    return m_modified;
}
  
// Set state
void File::SetModified(bool modified)
{
    m_modified = modified;
}
  
// Count
size_t File::GetCount() const
{
    return m_items.GetCount();
} 

// Get an item by index
PovItem& File::operator[](size_t k)
{
    return *m_items[k];
} 

const PovItem& File::operator[](size_t k) const
{
    return *m_items[k];
}   
  
  
// Add an item to the file
void File::AddItem(PovItem* item)
{
    // Disable it if another item is enabled
    if(IsEnabled(item->GetName()))
    {
        item->Enable(false);
    }

    m_items.Add(item);
    m_modified = true;
}
  
// Remove an item
void File::Remove(size_t item)
{
    // Delete and set modified
    delete m_items[item];
    m_items.RemoveAt(item, 1);
    m_modified = true;
} 
  
// Is an item enabled
bool File::IsEnabled(const wxString& name) const
{
    size_t count, pos;
  
    count = m_items.GetCount();
  
    for(pos = 0; pos < count; pos++)
    {
        if(m_items[pos]->GetName() == name && m_items[pos]->IsEnabled())
        {
            return true;
        }
    }
    
    return false;
} 
  
// Enable/disable an item and disable all other items with the
// same name if needed
void File::Enable(size_t item, bool enabled)
{
    size_t count = GetCount();

    if(item >= count)
    {
        return;
    }

    PovItem& theItem = *m_items[item];

    // If state is the same, do nothing
    if(theItem.IsEnabled() == enabled)
    {
        return;
    }

    // Set the state
    theItem.Enable(enabled);

    // If enabling, disable other items as needed
    if(enabled)
    {
        for(size_t pos = 0; pos < count; pos++)
        {
            if(pos != item && m_items[pos]->GetName() == theItem.GetName())
            {
                m_items[pos]->Enable(false);
            }
        }
    }

    // File has been modified
    SetModified(false);
}

// Sort the items
void File::Sort(bool ascending)
{
    size_t count = GetCount();
    size_t i, j;
    bool changed = false;

    // A basic sort will be fine
    for(i = 0; i < count - 1; i++)
    {
        for(j = i + 1; j < count; j++)
        {
            if(ascending && m_items[j]->GetName() < m_items[i]->GetName())
            {
                PovItem* tmp = m_items[i];
                m_items[i] = m_items[j];
                m_items[j] = tmp;
                changed = true;
            }
            else if(!ascending && m_items[i]->GetName() < m_items[j]->GetName())
            {
                PovItem* tmp = m_items[i];
                m_items[i] = m_items[j];
                m_items[j] = tmp;
                changed = true;
            }
        }
    }

    if(changed)
    {
        SetModified(true);
    }
}

// Move the item up or down
void File::Move(size_t dest, size_t src)
{
    size_t count = m_items.GetCount();

    if(src >= count)
    {
        return;
    }

    if(dest >= count)
    {
        dest = count - 1;
    }

    if(dest == src)
    {
        return;
    }

    // mark as modified
    m_modified = true;

    if(dest < src)
    {
        // Moving up
        PovItem* tmp = m_items[src];
        size_t pos;

        for(pos = src; pos > dest; pos--)
        {
            m_items[pos] = m_items[pos - 1];
        }

        m_items[dest] = tmp;
    }
    else if(dest > src)
    {
        // Moving down
        PovItem* tmp = m_items[src];
        size_t pos;

        for(pos = src; pos < dest; pos++)
        {
            m_items[pos] = m_items[pos + 1];
        }

        m_items[dest] = tmp;
    }
}

// Clear item and state information
void File::Clear()
{
    size_t count, pos;
  
    // Clear items
    count = m_items.GetCount();
    for(pos = 0; pos < count; pos++)
    {
        delete m_items[pos];
    }
    
    m_items.Clear();
    
    // Clear state
    m_modified = false;
    m_fname.Clear();
}
  
// Get the file name
const wxString& File::GetFileName() const
{
    return m_fname;
} 
  
  
// Load a file
bool File::Load(wxString fname)
{
    if(LoadFromFile(fname))
    {
        // Successfully loaded file
        m_modified = false;
        m_fname = fname;
    
        return true;
    }
    else
    {
        // Unsuccessfully loaded file
        Clear();
        return false;
    }
}
  
// Load a copy of a file
bool File::LoadCopy(wxString fname)
{
    if(LoadFromFile(fname))
    {
        // Successfully loaded file, don't remember name
        m_modified = true;
        return true;
    }
    else
    {
        // Unsuccessful load
        Clear();
        return false;
    }
}
  
// Save a file
bool File::Save(wxString fname)
{
    if(SaveToFile(fname))
    {
        m_fname = fname;
        m_modified = false;
        return true;
    }
    else
    {
        // Unsuccessfully saved, forget name
        m_fname.Empty();
        m_modified = true;
        return false;
    }
}
  
// Save a copy of a file
bool File::SaveCopy(wxString fname)
{
    // Do not change name or state
    return SaveToFile(fname);
}
  
// Really load
bool File::LoadFromFile(const wxString& fname)
{
    // Clear current file if any
    Clear();
  
    // Input file stream
    wxFFileInputStream stream(fname);
    if(stream.IsOk())
    {
        PovInput input(stream);
  
        return input.Load(*this);
    }
    else
    {
        return false;
    }
}
  
// Really save
bool File::SaveToFile(const wxString& fname)
{
    // Output stream
    wxFFileOutputStream stream(fname);
  
    if(stream.IsOk())
    {
        PovOutput output(stream, fname);
  
        return output.SaveFile(*this);  
    }
    else
    {
        return false;
    }
}


