/*
  Project:      POVColor
  File:         gammadlg.cpp
  Purpose:      Gamma dialog implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/intl.h>
#include <wx/stattext.h>
#include <wx/checkbox.h>
#include <wx/button.h>
#include <wx/slider.h>
#include <wx/sizer.h>
#include <wx/dcclient.h>
#include <wx/math.h>
#include <wx/utils.h>
#include <wx/dcbuffer.h>

#include "app.h"
#include "gammadlg.h"
#include "options.h"
#include "util.h"

// Defines for calibrate dialog
#define OUTERBAR_X 8
#define OUTERBAR_Y 8
#define INNERBAR_X 4 
#define INNERBAR_Y 4
#define BAR_X 500
#define BAR_Y 40
#define BOX_COUNT 15
#define BARSPACE_Y 32

// Defines for gamma dialog
#define BOX_X 150 
#define BOX_Y 75
#define BOXSPACE_X 32
#define BOXSPACE_Y 32


// Calibrate dialog - Aid in the calibration of a monitor
//-----------------------------------------------------------------------------
class CalibrateDialog : public wxDialog
{
public:
    // ctor/dtor (no default ctor)
    CalibrateDialog(wxWindow* parent);
    ~CalibrateDialog();
 
    // Paint events
    void OnPaint(wxPaintEvent& evt);
  
private:
    // No copy/assign
    CalibrateDialog(const CalibrateDialog& copy);
    CalibrateDialog& operator=(const CalibrateDialog& copy);

    // Create dialog items
    void CreateWidgets();
  
    // Areas
    wxSizerItem* m_prevArea;
  
DECLARE_EVENT_TABLE();
};

// Constructors
CalibrateDialog::CalibrateDialog(wxWindow* parent)
{
    // Create dialog
    if(!wxDialog::Create(parent, -1, _T("Calibrate Display"), wxDefaultPosition,
        wxSize(900, 900), wxDEFAULT_DIALOG_STYLE))
    {
        return;
    }
    
    // Create items
    CreateWidgets();
    CenterOnParent(wxBOTH);
}

// Destructor
CalibrateDialog::~CalibrateDialog()
{
}

// Paint the window
void CalibrateDialog::OnPaint(wxPaintEvent& WXUNUSED(evt))
{
    wxPaintDC dc(this);
    
    // Area to draw
    wxRect rc = m_prevArea->GetRect();
    
    // No pen
    dc.SetPen(*wxTRANSPARENT_PEN);

    // Brush used
    wxBrush br;
    
    // Outer area
    br.SetColour(wxColour(0, 0, 0));
    
    dc.SetBrush(br);
    dc.DrawRectangle(rc.x, rc.y, rc.width, rc.height);
    dc.SetBrush(wxNullBrush);
    
    // Inner border
    br.SetColour(wxColour(127, 127, 127));
    
    dc.SetBrush(br);
    dc.DrawRectangle(rc.x + OUTERBAR_X, rc.y + OUTERBAR_Y,
        rc.width - 2 * OUTERBAR_X, rc.height - 2 * OUTERBAR_Y);
    dc.SetBrush(wxNullBrush);
    
    // Individual items
    rc.x += (OUTERBAR_X + INNERBAR_X);
    rc.y += (OUTERBAR_Y + INNERBAR_Y);
    rc.width -= (2 * (OUTERBAR_X + INNERBAR_X));
    rc.height -= (2 * (OUTERBAR_Y + INNERBAR_Y));
    
    int pos;
    int width = rc.width / BOX_COUNT;
    
    // Center remaining items
    rc.x += (rc.width - (width * BOX_COUNT)) / 2;
    
    for(pos = 0; pos < BOX_COUNT; pos++)
    {
        // Brush color
        // TODO: make sure this is correct
        int c = pos * 255 / (BOX_COUNT - 1);
        br.SetColour(wxColour(c, c, c));
        
        // Draw item
        dc.SetBrush(br);
        dc.DrawRectangle(rc.x + pos * width, rc.y, width, rc.height);
        dc.SetBrush(wxNullBrush);
    }
    
    dc.SetPen(wxNullPen);
}

// Create dialog items
void CalibrateDialog::CreateWidgets()
{
    wxBusyCursor busy;
    
    // Top sizer
    wxBoxSizer* topSizer = new wxBoxSizer(wxVERTICAL);
    
    // Action text
    topSizer->Add(new wxStaticText(this, -1,
        _("1. Adjust contrast so that the last box is as bright as"
        " possible, but still different from next to last box.")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(new wxStaticText(this, -1,
        _("2. Adjust brightness so that the first box is as dark as"
        " possible, but still different from the next box.")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(new wxStaticText(this, -1,
        _("3. Adjust tint so that all boxes have no color, appearing only"
        " to be grey.")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM, BARSPACE_Y));      
        
    // Preview area
    m_prevArea = topSizer->Add(new wxSizerItem(2 * (OUTERBAR_X + INNERBAR_X) + BAR_X,
        2 * (OUTERBAR_Y + INNERBAR_Y) + BAR_Y, wxSizerFlags(1).Expand()));
    
    // Main sizer
    wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
    
    mainSizer->Add(topSizer, wxSizerFlags(1).Expand().DoubleBorder());
    
    // Buttons
    wxSizer* buttons = CreateSeparatedButtonSizer(wxOK);
    if(buttons)
    {
        mainSizer->Add(buttons, wxSizerFlags(0).Expand().DoubleBorder(wxLEFT | wxRIGHT | wxBOTTOM));
    }

    // Apply sizers
    SetSizerAndFit(mainSizer);
}

// Event table for calibration dialog
BEGIN_EVENT_TABLE(CalibrateDialog, wxDialog)
    EVT_PAINT(CalibrateDialog::OnPaint)
END_EVENT_TABLE()


// GammaDialog
//-----------------------------------------------------------------------------

enum
{
    ID_GAMMASLIDER = wxID_HIGHEST + 1,
    ID_CALIBRATEDISPLAY
};

// Constructors
GammaDialog::GammaDialog(wxWindow* parent)
{
    // Create dialog
    if(!wxDialog::Create(parent, -1, _T("Display Gamma"), wxDefaultPosition,
        wxSize(900, 900), wxDEFAULT_DIALOG_STYLE))
    {
        return;
    }
    
    // Create items
    CreateWidgets();
    CenterOnParent(wxBOTH);
}
  
// Destructors
GammaDialog::~GammaDialog()
{
} 

// Set values
void GammaDialog::OnInitDialog(wxInitDialogEvent& WXUNUSED(evt))
{
    // Set control state
    Options& opt = wxGetApp().GetOptions();
    
    // Checkbox
    m_gammaCheck->SetValue(opt.GetGammaCorrectionEnabled());
        
    // Value
    double gamma = opt.GetDisplayGamma();
        
    m_gammaSlider->SetValue((int)(gamma * 10.0));
        
    // Text
    m_gammaText->SetLabel(wxString::Format(_("Display Gamma: %.1f"), gamma));
}
    
// User pressed OK
void GammaDialog::OnOK(wxCommandEvent& WXUNUSED(evt))
{
    // Get new gamma and save it to configuration
    Options& opt = wxGetApp().GetOptions();
    
    opt.SetGammaCorrectionEnabled(m_gammaCheck->GetValue());
    opt.SetDisplayGamma((double)m_gammaSlider->GetValue() / 10.0);
        
    // Done
    EndModal(wxID_OK);
}
  
// User pressed 'Calibrate Display' button
void GammaDialog::OnCalibrate(wxCommandEvent& WXUNUSED(evt))
{
    CalibrateDialog dlg(this);
    dlg.ShowModal();
}
    
// Gamma value changed    
void GammaDialog::OnGammaChange(wxScrollEvent& evt)
{
    // Update text
    double gamma = (double)evt.GetPosition() / 10.0;
    m_gammaText->SetLabel(wxString::Format(_("Display Gamma: %.1f"), gamma));
  
    // Redraw gamma items
    wxClientDC dc(this);
  
    DrawAreas(dc);
}
    
// Paint the window
void GammaDialog::OnPaint(wxPaintEvent& WXUNUSED(evt))
{
    // Paint DC
    wxPaintDC dc(this);
  
    DrawAreas(dc);
}
  
    
// Create dialog items
void GammaDialog::CreateWidgets()
{
    wxBusyCursor busy;
  
    // Top sizer
    wxBoxSizer* topSizer = new wxBoxSizer(wxVERTICAL);
  
    // Action text
    topSizer->Add(new wxStaticText(this, -1, _("1. Squint your eyes or look from a distance")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(new wxStaticText(this, -1, _("2. Adjust slider until colors appear to blend together")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM, BOXSPACE_Y));
      
    // Gray area
    wxBoxSizer* graySizer = new wxBoxSizer(wxHORIZONTAL);
    
    graySizer->Add(new wxSizerItem(BOX_X, BOX_Y, wxSizerFlags(1).Expand()));
    graySizer->Add(BOXSPACE_X, 0);
    m_grayArea = graySizer->Add(new wxSizerItem(BOX_X, BOX_Y, wxSizerFlags(1).Expand()));
    graySizer->Add(BOXSPACE_X, 0);
    graySizer->Add(new wxSizerItem(BOX_X, BOX_Y, wxSizerFlags(1).Expand()));
    
    topSizer->Add(graySizer, wxSizerFlags(0).Expand().Border(wxBOTTOM, BOXSPACE_Y));
    
    // RGB area
    wxBoxSizer* rgbSizer = new wxBoxSizer(wxHORIZONTAL);
    
    m_redArea = rgbSizer->Add(new wxSizerItem(BOX_X, BOX_Y, wxSizerFlags(1).Expand()));
    rgbSizer->Add(BOXSPACE_X, 0);
    m_greenArea = rgbSizer->Add(new wxSizerItem(BOX_X, BOX_Y, wxSizerFlags(1).Expand()));
    rgbSizer->Add(BOXSPACE_X, 0);
    m_blueArea = rgbSizer->Add(new wxSizerItem(BOX_X, BOX_Y, wxSizerFlags(1).Expand()));
    
    topSizer->Add(rgbSizer, wxSizerFlags(0).Expand().Border(wxBOTTOM, BOXSPACE_Y));
  
    // Enable/disable area
    m_gammaCheck = new wxCheckBox(this, -1, _("Enable Gamma Correction."));
    
    topSizer->Add(m_gammaCheck, wxSizerFlags(0).Expand().Border(wxBOTTOM));
    
    // Text area
    m_gammaText = new wxStaticText(this, -1, wxEmptyString, wxDefaultPosition,
        wxDefaultSize, wxST_NO_AUTORESIZE);
    
    topSizer->Add(m_gammaText, wxSizerFlags(0).Expand().Border(wxBOTTOM));
  
    // Slider
    m_gammaSlider = new wxSlider(this, ID_GAMMASLIDER, 1, 1, 99);
    topSizer->Add(m_gammaSlider, wxSizerFlags(0).Expand().Border(wxBOTTOM));

    // Calibrate button
    topSizer->Add(new wxButton(this, ID_CALIBRATEDISPLAY, _("Calibrate Display")),
            wxSizerFlags(0).Right());
  
    // Main sizer
    wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  
    mainSizer->Add(topSizer, wxSizerFlags(1).Expand().DoubleBorder());

    // Buttons
    wxSizer* buttons = CreateSeparatedButtonSizer(wxOK | wxCANCEL);
    if(buttons)
    {
        mainSizer->Add(buttons, wxSizerFlags(0).Expand().DoubleBorder(wxLEFT | wxRIGHT | wxBOTTOM));
    }
  
    // Apply sizers and fit size
    SetSizerAndFit(mainSizer);
}
  
// Draw all areas
void GammaDialog::DrawAreas(wxDC& dc)
{
    // Get gamma
    double gamma = (double)m_gammaSlider->GetValue() / 10.0;
    
    // Calculate color needed
    int c = (int)(pow(0.5, 1.0 / gamma) * 255.0);
    
    // Gray area
    DrawArea(dc, m_grayArea->GetRect(), wxColour(255, 255, 255), wxColour(c, c, c));
    
    // Red area
    DrawArea(dc, m_redArea->GetRect(), wxColour(255, 0, 0), wxColour(c, 0, 0));
    
    // Green area
    DrawArea(dc, m_greenArea->GetRect(), wxColour(0, 255, 0), wxColour(0, c, 0));
    
    // Blue area
    DrawArea(dc, m_blueArea->GetRect(), wxColour(0, 0, 255), wxColour(0, 0, c));  
}
  
// Draw an area
void GammaDialog::DrawArea(wxDC& dc, const wxRect& rc, const wxColour& outer, const wxColour& inner)
{
    // Outer pens
    wxPen pblack(wxColour(0, 0, 0));
    wxPen pouter(outer);
    
    // Inner pen/brush
    wxPen pinner(inner);
    wxBrush binner(inner);
    
    // Draw outer area
    int y;
    bool state = true;
    for(y = rc.y; y < rc.y + rc.height; y++, state = !state)
    {
        if(state)
        {
            dc.SetPen(pouter);
        }
        else
        {
            dc.SetPen(pblack);
        }
        
        dc.DrawLine(rc.x, y, rc.x + rc.width, y);
    }
        
    // Draw inner area
    dc.SetPen(pinner);
    dc.SetBrush(binner);
    dc.DrawRectangle(rc.x + rc.width / 3, rc.y, rc.width / 3, rc.height);
        
    // Finished
    dc.SetPen(wxNullPen);
    dc.SetBrush(wxNullBrush);   
    
    // For visual purposes, it seems better not to draw a bevel around the item.
}

// Event table
BEGIN_EVENT_TABLE(GammaDialog, wxDialog)
    EVT_INIT_DIALOG(GammaDialog::OnInitDialog)
    EVT_BUTTON(wxID_OK, GammaDialog::OnOK)
    EVT_BUTTON(ID_CALIBRATEDISPLAY, GammaDialog::OnCalibrate)
    EVT_COMMAND_SCROLL(ID_GAMMASLIDER, GammaDialog::OnGammaChange)
    EVT_PAINT(GammaDialog::OnPaint)
END_EVENT_TABLE()

