/*
  Project:      POVColor
  File:         input.cpp
  Purpose:      Load POV input
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/string.h>
#include <wx/dynarray.h>
#include <wx/intl.h>
#include <wx/log.h>

#include "input.h"
#include "options.h"
#include "color.h"
#include "map.h"
#include "file.h"

#include <wx/arrimpl.cpp>


// PovInput
//-----------------------------------------------------------------------------

// Finish array
WX_DEFINE_OBJARRAY(TokenArray);

// Constructor
PovInput::PovInput(wxInputStream& stream) :
    // Remember stream and create text input
    m_stream(stream),
    m_text(stream),
    
    // Current read state from text
    m_didGet(false),
    m_didUnget(false),
    m_char(0),
    m_line(1),
    
    // Current token state
    m_next(0),
    
    // Items enabled by default
    m_enabled(true)
{
}
  
// Destructor
PovInput::~PovInput()
{
}

// Load from the input stream into file object
bool PovInput::Load(File& file)
{
    // Tokenize items first
    if(Tokenize() == false)
    {
        return false;
    }
        
    // Initial item state is enabled;
    m_enabled = true;

    // Scan through tokens
    Token t;
  
    t = GetNextToken();
    while(t.GetType() != Token::TOKEN_EOF)
    {
        switch(t.GetType())
        {
            // Process a #declare directive
            case Token::TOKEN_DIRECTIVE_DECLARE:
            {
                ParseDeclareDirective(file);
                break;
            }
        
            // Process a POVColor comment
            case Token::TOKEN_POVCOLOR:
            {
                wxString comment = t.GetString();
                
                // Handle POVColor comment
                if(comment.Find(wxT("ENABLED")) != wxNOT_FOUND)
                {
                    m_enabled = true;
                }
                else if(comment.Find(wxT("DISABLED")) != wxNOT_FOUND)
                {
                    m_enabled = false;
                }
                
                break;
            }
      
            case Token::TOKEN_KEYWORD_COLOR:
            case Token::TOKEN_KEYWORD_RGB:
            case Token::TOKEN_KEYWORD_RGBF:
            case Token::TOKEN_KEYWORD_RGBT:
            case Token::TOKEN_KEYWORD_RGBFT:
            {
                // Unget token for ColorItem::Load
                UngetToken();
            
                //Unnamed Color
                ColorItem* item = new ColorItem();
                if(item->Load(*this))
                {
                    // Set name and state
                    item->SetName(wxT("noname_color"));
                    item->Enable(false);
            
                    // Add item
                    file.AddItem(item);
                }
                else
                {
                    ::wxLogError(_("Unable to load unnamed color on line %d"), t.GetLine());
                    delete item;
                }
          
                break;
            } // case color
        
            case Token::TOKEN_KEYWORD_COLORMAP:
            {
                // Unget token for MapItem::Load
                UngetToken();

                // Unnamed Colormap
                MapItem* item = new MapItem();
                if(item->Load(*this))
                {
                    // Name and state
                    item->SetName(wxT("noname_map"));
                    item->Enable(false);
            
                    // Add
                    file.AddItem(item);
                }
                else
                {
                    ::wxLogError(_("Unable to load unnamed map on line %d"), t.GetLine());
                    delete item;
                }
         
                break;
            } // case colormap      
        
            default:
            {
                break; // To get rid of some warnings
            }
        } // switch
      
        t = GetNextToken();
    } // while
    
    // Finished
    return true;
}
  
// Get next token
const Token& PovInput::GetNextToken()
{
    static Token s_eof(0, Token::TOKEN_EOF);
  
    if(m_next < m_tokens.GetCount())
    {
        return m_tokens[m_next++];
    }
    else
    {
        // Even when reading past end of file, you must call
        // UngetToken as many times as GetNextToken before getting
        // back to the last item of the file
        m_next++;
        return s_eof;
    }
}
  
// Unget a token
void PovInput::UngetToken()
{
    if(m_next > 0)
    {
        m_next--; 
    }
}
  
// Read a vector
bool PovInput::ReadVector(size_t count, double* items, bool allowNumber /* = true */)
{
    Token t;
    size_t pos;
  
    // Open vector
    t = GetNextToken();
    if(t.GetType() != Token::TOKEN_OPEN_VECTOR)
    {
        // Wait before returning false, it may be just a number in case
        // we can treat it as a vector.  Like color rgb 1 is the same 
        // as color rgb <1, 1, 1>, so ReadVector should work for both
        if(allowNumber && t.GetType() == Token::TOKEN_NUMBER)
        {
            for(pos = 0; pos < count; pos++)
            {
                items[pos] = t.GetNumber();
            }
            return true;
        }
        else
        {
            UngetToken();
            return false;
        }
    }
      
    // Items
    for(pos = 0; pos < count; pos++)
    {
        // Comma in front of all but first
        if(pos != 0)
        {
            t = GetNextToken();
            if(t.GetType() != Token::TOKEN_COMMA)
            {
                UngetToken();
                return false;
            }
        }
      
        // Get number
        t = GetNextToken();
        if(t.GetType() != Token::TOKEN_NUMBER)
        {
            UngetToken();
            return false;
        }
      
        items[pos] = t.GetNumber();
    }
    
    // Closing vector
    t = GetNextToken();
    if(t.GetType() != Token::TOKEN_CLOSE_VECTOR)
    {
        UngetToken();
        return false;
    }
    
    return true;
}

// Build token list
bool PovInput::Tokenize()
{
    // Clear token list
    m_tokens.Clear();
  
    // Get a token
    // c1 is current character, c2 is next character
    wxChar c1, c2;
  
    // Get first character
    c1 = GetChar();
    c2 = 0;
  
    // Keep track of line numbers for reporting errors
    while(c1 != 0) // While not at the end
    {
        // Make sure there is no read error
        if(m_stream.GetLastError() == wxSTREAM_READ_ERROR)
        {
            ::wxLogError(_("An error occured while reading the file."));
            return false;
        }
    
        switch(c1)
        {
            // Open brace
            case wxT('{'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_OPEN_BLOCK));
                break;
            }
        
            // Close brace
            case wxT('}'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_CLOSE_BLOCK));
                break;
            }
        
            // Open vector
            case wxT('<'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_OPEN_VECTOR));
                break;
            }
        
            // Close vector
            case wxT('>'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_CLOSE_VECTOR));
                break;
            }
        
            // Open colormap
            case wxT('['):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_OPEN_MAP));
                break;
            }
        
            // Close colormap
            case wxT(']'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_CLOSE_MAP));
                break;
            }
        
            // Open parenthesis
            case wxT('('):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_OPEN_PAREN));
                break;
            }
        
            // Close parenthesis
            case wxT(')'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_CLOSE_PAREN));
                break;
            }
        
            // Assignment
            case wxT('='):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_ASSIGN));
                break;
            }
        
            // Comma
            case wxT(','):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_COMMA));
                break;
            }
        
            // Semicolon
            case wxT(';'):
            {
                m_tokens.Add(Token(m_line, Token::TOKEN_SEMICOLON));
                break;
            }
        
            // Directive
            case wxT('#'):
            {
                TokenizeDirective();
                break;
            }
        
            // String
            case wxT('\"'):
            {
                TokenizeString();
                break;
            }
        
            // Comment maybe
            case wxT('/'):
            {
                c2 = GetChar();
        
                if(c2 == wxT('*'))
                {
                    // C comment
                    TokenizeCComment();
                }
                else if(c2 == wxT('/'))
                {
                    // C++ comment
                    TokenizeCPPComment();
                }
                else
                {
                    // No comment
                    UngetChar();
                }
        
                break;
            }
        
            // None of the above
            default:
            {
                // Check for number
                if(wxIsdigit(c1) || c1 == wxT('.'))
                {
                    UngetChar();
                    TokenizeNumber();
                }
                else if(wxIsalpha(c1) || c1 == wxT('_'))
                {
                    // Symbol (idenitifer or keyword)
                    UngetChar();
                    TokenizeSymbol();
                }
                else if(wxIsspace(c1))
                {
                    // Whitespace, do nothing
                }
                else 
                {
                    // Unknown token
                    m_tokens.Add(Token(m_line, Token::TOKEN_UNKNOWN));
                }
        
                break;
            }
        }
      
        // Get next character
        c1 = GetChar();    
    }
  
    // Add one last token, the EOF token
    // This fixes some inf. loop problems if a recognized item is at file end
    m_tokens.Add(Token(m_line, Token::TOKEN_EOF));
    
    // Set position
    m_next = 0;
    
    return true;
}

// Tokenize a number
void PovInput::TokenizeNumber()
{
    wxString result;
    wxChar c;

    // Read input string  
    c = GetChar();
    while(wxIsdigit(c) || c == wxT('.'))
    {
        result += c;
        c = GetChar();
    }
    
    // Unget last invalid character (not part of number)
    UngetChar();
  
    // Convert to a double
    double d;
  
    if(result.ToDouble(&d))
    {
        // Add token
        m_tokens.Add(Token(m_line, d));
    }
    else
    {
        m_tokens.Add(Token(m_line, Token::TOKEN_UNKNOWN));
    }
}
  
// Tokenize symbol
void PovInput::TokenizeSymbol()
{
    wxString symbol;
    wxChar c;
  
    // Read symbol
    // Number are needed if it is an identifier
    c = GetChar();
  
    while(wxIsalnum(c) || c == wxT('_'))
    {
        symbol += c;
        c = GetChar();
    }
  
     UngetChar();
        
    // Check for valid keywords
    if(symbol == wxT("version"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_VERSION));
    else if(symbol == wxT("color") || symbol == wxT("colour"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_COLOR));
    else if(symbol == wxT("color_map") || symbol == wxT("colour_map"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_COLORMAP));
    else if(symbol == wxT("rgb"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_RGB));
    else if(symbol == wxT("rgbt"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_RGBT));
    else if(symbol == wxT("rgbf"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_RGBF));
    else if(symbol == wxT("rgbft"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_RGBFT));
    else if(symbol == wxT("red"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_RED));
    else if(symbol == wxT("green"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_GREEN));
    else if(symbol == wxT("blue"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_BLUE));
    else if(symbol == wxT("filter"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_FILTER));
    else if(symbol == wxT("transmit"))
        m_tokens.Add(Token(m_line, Token::TOKEN_KEYWORD_TRANSMIT));
    else
        m_tokens.Add(Token(m_line, Token::TOKEN_IDENTIFIER, symbol));
}
  
  
// Tokenize string
void PovInput::TokenizeString()
{
    // Already have opening quote
    
    // Not used by POVColor, just skip it
    wxChar c;
    
    while(true)
    {
        c = GetChar();
        
        if(c == wxT('\"'))
        {
            // Quote first, we are done
            break;
        }
        else if(c == wxT('\\'))
        {
            // Escape sequence, skip next character
            c = GetChar();
        }
        
        // End of file (first or second)
        if(c == 0)
        {
            UngetChar();
            break;
        }
    }
        
    m_tokens.Add(Token(m_line, Token::TOKEN_STRING));
}
  
// Tokenize a C comment
void PovInput::TokenizeCComment()
{
    // Already have first part
    wxString total;
    wxChar c1, c2;
    
    // Read
    while(true)
    {
        c1 = GetChar();
        
        if(c1 == wxT('*'))
        {
            // Maybe the end
            c2 = GetChar();
            
            if(c2 == wxT('/'))
            {
                // The end
                total = total.Strip(wxString::both);
                
                // Look for special POVColor comment
                if(total.Find(wxT("POVCOLOR:")) == 0)
                {
                    m_tokens.Add(Token(m_line, Token::TOKEN_POVCOLOR, total.Mid(9)));
                }
            
                break;
            }
            else
            {
                // Not the end
                UngetChar();
                total += c1;
            }
        }
        else if(c1 == 0)
        {
            UngetChar();
            break;
        }
        else
        {
            total += c1;
        }
    } // while
}
  
// Tokenize a C++ comment
void PovInput::TokenizeCPPComment()
{
    // Alread have the opening
    wxChar c;
    
    while(true)
    {
        c = GetChar();
    
        if(c == wxT('\n'))
        {
            break;
        }
        else if(c == 0)
        {
            UngetChar();
            break;
        }
    }
} 

// Tokenize a directive
void PovInput::TokenizeDirective()
{
    // Already have the #
    wxString directive;
    wxChar c;
    
    // Read
    c = GetChar();
  
    while(wxIsalpha(c) || c == wxT('_'))
    {
        // Don't need numbers for what we are processing
        directive += c;
        c = GetChar();
    }
  
    UngetChar();
  
    
    // What directive is it
    // Check for valid directives
    if(directive == wxT("ifdef"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_IFDEF));
    else if(directive == wxT("ifndef"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_IFNDEF));
    else if(directive == wxT("undef"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_UNDEF));
    else if(directive == wxT("end"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_END));
    else if(directive == wxT("declare"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_DECLARE));
    else if(directive == wxT("debug"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_DEBUG));
    else if(directive == wxT("version"))
        m_tokens.Add(Token(m_line, Token::TOKEN_DIRECTIVE_VERSION));
    else
        m_tokens.Add(Token(m_line, Token::TOKEN_UNKNOWN)); 
}
  
  
// Parse a declare directive
void PovInput::ParseDeclareDirective(File& file)
{
    Token t;
    wxString ident;
  
    // Get identifier
    t = GetNextToken();
    if(t.GetType() != Token::TOKEN_IDENTIFIER)
    {
        return;
    }
    
    ident = t.GetString();
  
    // Get assignment
    t = GetNextToken();
    if(t.GetType() != Token::TOKEN_ASSIGN)
    {
        return;
    }
    
    // Get type
    t = GetNextToken();

    switch(t.GetType())
    {
        case Token::TOKEN_KEYWORD_COLOR:
        case Token::TOKEN_KEYWORD_RGB:
        case Token::TOKEN_KEYWORD_RGBF:
        case Token::TOKEN_KEYWORD_RGBT:
        case Token::TOKEN_KEYWORD_RGBFT:
        {
            // Unget token for ColorItem::Load
            UngetToken();
    
            // Color
            ColorItem* item = new ColorItem();
            if(item->Load(*this))
            {
                // Set name and state
                item->SetName(ident);
        
                if(file.IsEnabled(ident))
                {
                    item->Enable(false);
                }
                else
                {
                    item->Enable(m_enabled);
                }
        
                // Add item
                file.AddItem(item);
            }
            else
            {
                ::wxLogError(_("Unable to load color '%s' on line %d"), ident.c_str(),
                    t.GetLine());
                delete item;
            }
        
             break;
        } // case color
      
        case Token::TOKEN_KEYWORD_COLORMAP:
        {
            // Unget token for MapItem::Load
            UngetToken();        

            // Colormap
            MapItem* item = new MapItem();
            if(item->Load(*this))
            {
                // Name and state
                item->SetName(ident);
        
                if(file.IsEnabled(ident))
                {
                    item->Enable(false);
                }
                else
                {
                    item->Enable(m_enabled);
                }
        
                // Add
                file.AddItem(item);
            }
            else
            {
                ::wxLogError(_("Unable to load map '%s' on line %d"), ident.c_str(),
                    t.GetLine());
                delete item;
            }
        
            break;
        } // case colormap
      
        case Token::TOKEN_UNKNOWN:
        {
            ::wxLogError(_("Unable to load item '%s' on line %d"), ident.c_str(),
                t.GetLine());
            break;
        }
      
        default:
        {
            break; // To get rid of some warnings
        }
      
    } // switch t.GetType()
} 

wxChar PovInput::GetChar()
{
    if(m_didUnget)
    {
        m_didUnget = false;
        return m_char;
    }
    else
    {
        m_char = m_text.GetChar();
    
        // Remember line number
        if(m_char == wxT('\n'))
            m_line++;
    }
    
    m_didGet = true;
  
    return m_char;
}
  
// Unget a character
void PovInput::UngetChar()
{
    if(m_didGet)
    {
        m_didUnget = true;
    }
}

