/*
  Project:      POVColor
  File:         input.h
  Purpose:      Load/tokenize/parse POV input from a stream
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __POVCOLOR_INPUT_H__
#define __POVCOLOR_INPUT_H__


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/string.h>
#include <wx/txtstrm.h>
#include <wx/dynarray.h>

// Forward declarations
class File;


// Token - A token object used by the input
//-----------------------------------------------------------------------------
class Token
{
public:
    // Token type
    enum Type
    {
        TOKEN_UNKNOWN = 0,
        TOKEN_INIT,
        TOKEN_EOF,
        TOKEN_DIRECTIVE_IFDEF, // #ifdef(id)
        TOKEN_DIRECTIVE_IFNDEF, // #ifndef(id)
        TOKEN_DIRECTIVE_UNDEF, // #undef
        TOKEN_DIRECTIVE_END, // #end
        TOKEN_DIRECTIVE_DECLARE, // #declare
        TOKEN_DIRECTIVE_DEBUG, // #debug
        TOKEN_DIRECTIVE_VERSION, // #version
        TOKEN_KEYWORD_VERSION, // version
        TOKEN_KEYWORD_COLOR, // color, colour
        TOKEN_KEYWORD_COLORMAP, // color_map, colour_map
        TOKEN_KEYWORD_RGB, // rgb
        TOKEN_KEYWORD_RGBT, // rgbt
        TOKEN_KEYWORD_RGBF, // rgbf
        TOKEN_KEYWORD_RGBFT, // rgbft
        TOKEN_KEYWORD_RED, // red
        TOKEN_KEYWORD_GREEN, // green
        TOKEN_KEYWORD_BLUE, // blue
        TOKEN_KEYWORD_FILTER, // filter
        TOKEN_KEYWORD_TRANSMIT, // transmit
        TOKEN_OPEN_BLOCK, // {
        TOKEN_CLOSE_BLOCK, // }
        TOKEN_OPEN_VECTOR, // <
        TOKEN_CLOSE_VECTOR, // >
        TOKEN_OPEN_MAP, // [
        TOKEN_CLOSE_MAP, // ]
        TOKEN_OPEN_PAREN, // (
        TOKEN_CLOSE_PAREN, // )
        TOKEN_ASSIGN, // =
        TOKEN_COMMA, // ,
        TOKEN_SEMICOLON, // ;
        TOKEN_IDENTIFIER,
        TOKEN_STRING,
        TOKEN_NUMBER,
        TOKEN_POVCOLOR, // POVColor comment data (stored as a string)
    };
    
    // Constructor
    Token();
    Token(const Token& copy);
    
    // Destructor
    ~Token();
    
    // Assignment
    Token& operator=(const Token& rhs);   
    
    // Type
    Type GetType() const;
    
    // Get information
    int GetLine() const; // Line number of token
    const wxString& GetString() const; // String, ID, or comment
    double GetNumber() const; // Number

  
private:
    // Constructors
    Token(int line, Type type, const wxString& str = wxEmptyString);
    Token(int line, double number);
    
    // Data
    int m_line;
    Type m_type;
    wxString m_string;
    double m_number;
  

friend class PovInput;
};

// Array of tokens
WX_DECLARE_OBJARRAY(Token, TokenArray); 

// Inlines

// Constructors
inline Token::Token() :
    m_type(Token::TOKEN_UNKNOWN),
    m_number(0.0)
{
}

inline Token::Token(int line, Token::Type type, const wxString& str) :
    m_line(line),
    m_type(type),
    m_string(str),
    m_number(0.0)
{
}
  
inline Token::Token(int line, double number) :
    m_line(line),
    m_type(TOKEN_NUMBER),
    m_number(number)
{
}
  
inline Token::Token(const Token& copy) :
    m_line(copy.m_line),
    m_type(copy.m_type),
    m_string(copy.m_string),
    m_number(copy.m_number)
{
}
  
// Destructor
inline Token::~Token()
{
}
  
// Assignment
inline Token& Token::operator=(const Token& rhs)
{
    if(this != &rhs)
    {
        m_line = rhs.m_line;
        m_type = rhs.m_type;
        m_string = rhs.m_string;
        m_number = rhs.m_number;
    }
        
    return *this;
}
  
// Get information
inline int Token::GetLine() const
{
    return m_line;
}

inline Token::Type Token::GetType() const
{
    return m_type;
}
  
inline const wxString& Token::GetString() const
{
    return m_string;
}
  
inline double Token::GetNumber() const
{
    return m_number;
}


// PovInput - Read input to a File object
//-----------------------------------------------------------------------------
class PovInput
{
public:

    // Constructor and destructor (no default ctor) 
    PovInput(wxInputStream& stream);
    ~PovInput();
    
    // Load the input
    bool Load(File& file);
    
    const Token& GetNextToken();
    void UngetToken();
    
    // Read a vector
    bool ReadVector(size_t count, double* items, bool allowNumber = true);
  
private:
    // No copy/assign
    PovInput(const PovInput& copy);
    PovInput& operator=(const PovInput& copy);

    bool Tokenize();
    
    void TokenizeNumber();
    void TokenizeSymbol();
    void TokenizeString();
    void TokenizeCComment();
    void TokenizeCPPComment();
    void TokenizeDirective();
    
    void ParseDeclareDirective(File& file);
    
    wxChar GetChar();
    void UngetChar(); // Can only unget the last read character
    
    // Input data
    wxInputStream& m_stream;
    wxTextInputStream m_text;

    bool m_didGet;
    bool m_didUnget;
    wxChar m_char;
    int m_line;
    
    // Tokens
    TokenArray m_tokens;
    size_t m_next;
    
    // Used for the state of the item
    bool m_enabled;
};
  

#endif // Include once wrapper

