/*
  Project:      POVColor
  File:         mainwnd.cpp
  Purpose:      Main window implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/frame.h>
#include <wx/intl.h>
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/sizer.h>
#include <wx/menu.h>
#include <wx/statusbr.h>
#include <wx/toolbar.h>
#include <wx/icon.h>
#include <wx/msgdlg.h>
#include <wx/filedlg.h>
#include <wx/notebook.h>
#include <wx/docview.h>
#include <wx/dnd.h>
#include <wx/config.h>
#include <wx/arrstr.h>

#include "app.h"
#include "art.h"
#include "mainwnd.h"
#include "options.h"
#include "view.h"
#include "coldlg.h"
#include "mapdlg.h"
#include "aboutdlg.h"


// MainWindow
//-----------------------------------------------------------------------------
enum
{
    ID_FILENEW = wxID_HIGHEST + 1,
    ID_FILEOPEN,
    ID_FILEOPENCOPY,
    ID_FILESAVE,
    ID_FILESAVEAS,
    ID_FILESAVECOPY,
    ID_FILECLOSE,
    ID_FILEEXIT,

    ID_EDITCOPY,
    ID_EDITPASTE,
    ID_EDITEDIT,
    ID_EDITRENAME,
    ID_EDITDELETE,
    ID_EDITCOPYCODE,
    ID_EDITENABLE,
    ID_EDITDISABLE,
    ID_EDITSORTA,
    ID_EDITSORTD,
    ID_EDITMOVEU,
    ID_EDITMOVED,

    ID_ITEMNEWCOLOR,
    ID_ITEMNEWMAP,

    ID_VIEWOPTIONS,
    ID_VIEWGAMMA,

    ID_HELPCONTENTS,
    ID_HELPABOUT
};

// Drop target
class MainWindowFileTarget : public wxFileDropTarget
{
public:
    MainWindowFileTarget(MainWindow* wnd) : m_wnd(wnd)
    {
    }

    ~MainWindowFileTarget()
    {
    }

    bool OnDropFiles(wxCoord WXUNUSED(x), wxCoord WXUNUSED(y), const wxArrayString& files)
    {
        // Open the files
        for(size_t pos = 0, count = files.GetCount(); pos < count; pos++)
        {
            m_wnd->OpenFile(files[pos]);
        }

        return true;
    }

private:
    MainWindow* m_wnd;
};

// Constructor
MainWindow::MainWindow()
{
    // Create frame
    wxFrame::Create(NULL, -1, wxT("POVColor " APP_VERSION), wxDefaultPosition, wxSize(900, 900));

    // Set Icon
    SetIcons(Art::GetIconBundle(wxT("mainicon")));

    // Create child items
    CreateWidgets();
    CenterOnParent(wxBOTH);

    // Set drop target
    SetDropTarget(new MainWindowFileTarget(this));

    // Load file history
    wxArrayString files = wxGetApp().GetOptions().GetFileHistory();
    size_t pos, count;
    
    count = files.GetCount();
    for(pos = count; pos > 0; pos--)
    {
        m_history->AddFileToHistory(files[pos - 1]);
    }
}

// Destructor
MainWindow::~MainWindow()
{
    // Save file history
    wxArrayString files;
    size_t pos, count;

    count = m_history->GetCount();
    for(pos = 0; pos < count; pos++)
    {
        files.Add(m_history->GetHistoryFile(pos));
    }

    wxGetApp().GetOptions().SetFileHistory(files);

    delete m_history;
}

// Create a new file
void MainWindow::OnFileNew(wxCommandEvent& WXUNUSED(evt))
{
    NewFile();
}

// Open a file
void MainWindow::OnFileOpen(wxCommandEvent& WXUNUSED(evt))
{
    // Ask for file to load
    wxString fname = ::wxFileSelector(_("Open file"), wxEmptyString,
        wxEmptyString, wxEmptyString, wxT("*.*"), wxFD_OPEN | wxFD_FILE_MUST_EXIST,
        this);

    if(fname.empty())
        return;

    OpenFile(fname);
}

// Open a copy of a file
void MainWindow::OnFileOpenCopy(wxCommandEvent& WXUNUSED(evt))
{
    // Ask for file to load
    wxString fname = ::wxFileSelector(_("Open file"), wxEmptyString,
        wxEmptyString, wxEmptyString, wxT("*.*"), wxFD_OPEN | wxFD_FILE_MUST_EXIST,
        this);

    if(fname.empty())
        return;

    OpenFile(fname, true);
}

// Open a recent file
void MainWindow::OnFileOpenRecent(wxCommandEvent& evt)
{
    // Get history filename
    wxString fname = m_history->GetHistoryFile(evt.GetId() - wxID_FILE1);

    if(fname.empty())
    {
        return;
    }

    // Open file, remove from history if fail
    if(!OpenFile(fname))
    {
        // Remove item from history
        m_history->RemoveFileFromHistory(evt.GetId() - wxID_FILE1);
    }
}

// Save the file
void MainWindow::OnFileSave(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    int sel = m_notebook->GetSelection();
    if(sel < 0)
    {
        return;
    }

    View* v = (View*)m_notebook->GetPage(sel);

    if(v == NULL)
    {
        return;
    }

    if(SaveFile(v))
    {
        // Update title
        m_notebook->SetPageText(sel, v->GetDisplayName());
    }
}

// Save As the file
void MainWindow::OnFileSaveAs(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    int sel = m_notebook->GetSelection();
    if(sel < 0)
    {
        return;
    }

    View* v = (View*)m_notebook->GetPage(sel);
    if(v == NULL)
    {
        return;
    }

    if(SaveFile(v, true))
    {
        // Update title
        m_notebook->SetPageText(sel, v->GetDisplayName());
    }
}

// Save a copy of the file
void MainWindow::OnFileSaveCopy(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    int sel = m_notebook->GetSelection();
    if(sel < 0)
    {
        return;
    }

    View* v = (View*)m_notebook->GetPage(sel);
    if(v == NULL)
    {
        return;
    }

    // Save the file
    SaveFile(v, true, true);
}

// Close the file
void MainWindow::OnFileClose(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    int sel = m_notebook->GetSelection();
    if(sel < 0)
    {
        return;
    }

    View* v = (View*)m_notebook->GetPage(sel);
    if(v == NULL)
    {
        return;
    }

    // Is the file modified, save
    if(QueryClose(v) == false)
    {
        return;
    }

    m_notebook->DeletePage(sel);
}

// Exit
void MainWindow::OnFileExit(wxCommandEvent& WXUNUSED(evt))
{
    if(CloseAllFiles())
    {
        Close();
    }
}

// Copy items
void MainWindow::OnEditCopy(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->CopyItems();
    }
}

// Paste items
void MainWindow::OnEditPaste(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->PasteItems();
    }
}

// Edit items
void MainWindow::OnEditEdit(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->EditItem();
    }
}

// Rename items
void MainWindow::OnEditRename(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->RenameItem();
    }
}

// Delete items
void MainWindow::OnEditDelete(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->DeleteItem();
    }
}

// Copy code
void MainWindow::OnEditCopyCode(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->CopyCode();
    }
}

// Enable items
void MainWindow::OnEditEnable(wxCommandEvent& evt)
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->EnableItem(evt.GetId() == ID_EDITENABLE);
    }
}

// Sort
void MainWindow::OnEditSort(wxCommandEvent& evt)
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->Sort(evt.GetId() == ID_EDITSORTA);
    }
}

// Move
void MainWindow::OnEditMove(wxCommandEvent& evt)
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        if(evt.GetId() == ID_EDITMOVEU)
        {
            v->MoveUp();
        }
        else
        {
            v->MoveDown();
        }
    }
}

// Create a new color
void MainWindow::OnItemNewColor(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->NewColor();
    }
    else
    {
        ColorDialog dlg(this);
        if(dlg.ShowModal() == wxID_OK)
        {
            // Copy code if user pressed ok
            // even if auto-copy code is off
            dlg.GetColorItem().CopyCode();
        }
    }
}

// Create a new map
void MainWindow::OnItemNewMap(wxCommandEvent& WXUNUSED(evt))
{
    // Get current view
    View* v = (View*)m_notebook->GetCurrentPage();

    if(v)
    {
        v->NewMap();
    }
    else
    {
        MapDialog dlg(this);
        if(dlg.ShowModal() == wxID_OK)
        {
            // Copy code if user pressed ok
            // even if auto-copy code is off
            dlg.GetMapItem().CopyCode();
        }
    }
}

// Adjust options
void MainWindow::OnViewOptions(wxCommandEvent& WXUNUSED(evt))
{
    Options::PromptOptions(this);
}

// Adjust gamma
void MainWindow::OnViewGamma(wxCommandEvent& WXUNUSED(evt))
{
    if(Options::PromptGamma(this))
    {
        // Get current view
        int sel = m_notebook->GetSelection();
        if(sel < 0)
        {
            return;
        }

        View* v = (View*)m_notebook->GetPage(sel);
        if(v == NULL)
        {
            return;
        }

        // Update
        v->RedrawList();
    }
}

// Show help
void MainWindow::OnHelpContents(wxCommandEvent& WXUNUSED(evt))
{
    ::wxLaunchDefaultBrowser(wxGetApp().GetHelpFile());
}

// Show about
void MainWindow::OnHelpAbout(wxCommandEvent& WXUNUSED(evt))
{
    AboutDialog dlg(this);

    dlg.ShowModal();
}

// Exit
void MainWindow::OnClose(wxCloseEvent& evt)
{
    if(evt.CanVeto())
    {
        if(CloseAllFiles())
            Destroy();
    }
    else
    {
        // TODO: What to do here?
        Destroy(); // Changes will be lost
    }
}

// Create items
void MainWindow::CreateWidgets()
{
    wxBusyCursor busy;

    // Create notebook
    wxBoxSizer* topSizer = new wxBoxSizer(wxVERTICAL);

    m_notebook = new wxNotebook(this, -1, wxDefaultPosition, wxDefaultSize,
        wxNB_TOP);
    m_notebook->SetMinSize(wxSize(500, 400));

    topSizer->Add(m_notebook, wxSizerFlags(1).Expand());

    // Menu bar

    // File
    wxMenu* fileMenu = new wxMenu();
    wxMenu* fileRecentMenu = new wxMenu();

    fileMenu->Append(ID_FILENEW, _("&New\tCTRL+N"), _("Create a new file."));
    fileMenu->AppendSeparator();
    fileMenu->Append(ID_FILEOPEN, _("&Open...\tCTRL+O"), _("Open a file."));
    fileMenu->Append(ID_FILEOPENCOPY, _("O&pen Copy..."), _("Open a copy of a file."));
    fileMenu->Append(-1, _("Open &Recent"), fileRecentMenu);
    fileMenu->AppendSeparator();
    fileMenu->Append(ID_FILESAVE, _("&Save\tCTRL+S"), _("Save the current file."));
    fileMenu->Append(ID_FILESAVEAS, _("Save &As..."), _("Save the current file with a new name."));
    fileMenu->Append(ID_FILESAVECOPY, _("Sa&ve Copy..."), _("Save a copy of the current file."));
    fileMenu->AppendSeparator();
    fileMenu->Append(ID_FILECLOSE, _("&Close\tCTRL+W"), _("Close the current file."));
    fileMenu->AppendSeparator();
    fileMenu->Append(ID_FILEEXIT, _("E&xit\tALT+F4"), _("Close all files and exit the program."));

    // Edit
    wxMenu* editMenu = new wxMenu();

    editMenu->Append(ID_EDITCOPY, _("&Copy\tCTRL+C"), _("Copy selected items."));
    editMenu->Append(ID_EDITPASTE, _("&Paste\tCTRL+V"), _("Paste copied items."));
    editMenu->AppendSeparator();
    editMenu->Append(ID_EDITEDIT, _("&Edit...\tENTER"), _("Edit the selected item."));
    editMenu->Append(ID_EDITRENAME, _("&Rename...\tCTRL+R"), _("Rename the selected item."));
    editMenu->Append(ID_EDITDELETE, _("&Delete\tDEL"), _("Delete the seletect items."));
    editMenu->Append(ID_EDITCOPYCODE, _("C&opy Code\tCTRL+G"), _("Copy code of the selected item."));
    editMenu->AppendSeparator();
    editMenu->Append(ID_EDITENABLE, _("E&nable\tCTRL+E"), _("Enable selected items."));
    editMenu->Append(ID_EDITDISABLE, _("D&isable\tCTRL+D"), _("Disable selected items."));
    editMenu->AppendSeparator();
    editMenu->Append(ID_EDITSORTA, _("&Sort Ascending"), _("Sort items."));
    editMenu->Append(ID_EDITSORTD, _("Sor&t Decending"), _("Sort items."));
    editMenu->AppendSeparator();
    editMenu->Append(ID_EDITMOVEU, _("Move &Up"), _("Move items."));
    editMenu->Append(ID_EDITMOVED, _("Move Do&wn"), _("Move items."));

    // Item
    wxMenu* itemMenu = new wxMenu();

    itemMenu->Append(ID_ITEMNEWCOLOR, _("New &Color...\tCTRL+1"), _("Create a new color."));
    itemMenu->Append(ID_ITEMNEWMAP, _("New &Map...\tCTRL+2"), _("Create a new map."));

    // View
    wxMenu* viewMenu = new wxMenu();

    viewMenu->Append(ID_VIEWOPTIONS, _("&Options..."), _("Adjust options."));
    viewMenu->Append(ID_VIEWGAMMA, _("&Gamma..."), _("Adjust gamma."));

    // Help
    wxMenu* helpMenu = new wxMenu();

    helpMenu->Append(ID_HELPCONTENTS, _("&Contents..."), _("View program help."));
    helpMenu->Append(ID_HELPABOUT, _("&About"), _("View information about the program."));

    // Main menu
    wxMenuBar* menu = new wxMenuBar();

    menu->Append(fileMenu, _("&File"));
    menu->Append(editMenu, _("&Edit"));
    menu->Append(itemMenu, _("&Item"));
    menu->Append(viewMenu, _("&View"));
    menu->Append(helpMenu, _("&Help"));

    SetMenuBar(menu);

    // Tool bar
    wxToolBar* tools = CreateToolBar();

    tools->AddTool(ID_FILENEW, wxEmptyString, Art::GetBitmap(wxT("new")), wxNullBitmap,
        wxITEM_NORMAL, _("New"), _("Create a new file."));
    tools->AddTool(ID_FILEOPEN, wxEmptyString, Art::GetBitmap(wxT("open")), wxNullBitmap,
        wxITEM_NORMAL, _("Open"), _("Open a file."));
    tools->AddTool(ID_FILESAVE, wxEmptyString, Art::GetBitmap(wxT("save")), wxNullBitmap,
        wxITEM_NORMAL, _("Save"), _("Save the current file."));
    tools->AddTool(ID_FILESAVEAS, wxEmptyString, Art::GetBitmap(wxT("saveas")), wxNullBitmap,
        wxITEM_NORMAL, _("Save As"), _("Save the current file with a new name."));
    tools->AddTool(ID_FILECLOSE, wxEmptyString, Art::GetBitmap(wxT("close")), wxNullBitmap,
        wxITEM_NORMAL, _("Close"), _("Close the current file."));

    tools->AddSeparator();

    tools->AddTool(ID_EDITCOPY, wxEmptyString, Art::GetBitmap(wxT("copy")), wxNullBitmap,
        wxITEM_NORMAL, _("Copy"), _("Copy selected items."));
    tools->AddTool(ID_EDITPASTE, wxEmptyString, Art::GetBitmap(wxT("paste")), wxNullBitmap,
        wxITEM_NORMAL, _("Paste"), _("Paste copied items."));

    tools->AddSeparator();

    tools->AddTool(ID_EDITENABLE, wxEmptyString, Art::GetBitmap(wxT("enable")), wxNullBitmap,
        wxITEM_NORMAL, _("Enable"), _("Enable selected items."));

    tools->AddTool(ID_EDITDISABLE, wxEmptyString, Art::GetBitmap(wxT("disable")), wxNullBitmap,
        wxITEM_NORMAL, _("Disable"), _("Disable selected items."));

    tools->AddSeparator();

    tools->AddTool(ID_EDITSORTA, wxEmptyString, Art::GetBitmap(wxT("sorta")), wxNullBitmap,
        wxITEM_NORMAL, _("Sort Ascending"), _("Sort items."));
    tools->AddTool(ID_EDITSORTD, wxEmptyString, Art::GetBitmap(wxT("sortd")), wxNullBitmap,
        wxITEM_NORMAL, _("Sort Decending"), _("Sort items."));

    tools->AddSeparator();

    tools->AddTool(ID_EDITMOVEU, wxEmptyString, Art::GetBitmap(wxT("moveu")), wxNullBitmap,
        wxITEM_NORMAL, _("Move Up"), _("Move items."));
    tools->AddTool(ID_EDITMOVED, wxEmptyString, Art::GetBitmap(wxT("moved")), wxNullBitmap,
        wxITEM_NORMAL, _("Move Down"), _("Move items."));

    tools->AddSeparator();

    tools->AddTool(ID_ITEMNEWCOLOR, wxEmptyString, Art::GetBitmap(wxT("color")), wxNullBitmap,
        wxITEM_NORMAL, _("New Color"), _("Create a new color."));
    tools->AddTool(ID_ITEMNEWMAP, wxEmptyString, Art::GetBitmap(wxT("map")), wxNullBitmap,
        wxITEM_NORMAL, _("New Map"), _("Create a new map."));

    tools->AddSeparator();

    tools->AddTool(ID_VIEWOPTIONS, wxEmptyString, Art::GetBitmap(wxT("options")), wxNullBitmap,
        wxITEM_NORMAL, _("Options"), _("Adjust options."));
    tools->AddTool(ID_VIEWGAMMA, wxEmptyString, Art::GetBitmap(wxT("gamma")), wxNullBitmap,
        wxITEM_NORMAL, _("Gamma"), _("Adjust gamma."));

    tools->AddSeparator();

    tools->AddTool(ID_HELPCONTENTS, wxEmptyString, Art::GetBitmap(wxT("help")), wxNullBitmap,
        wxITEM_NORMAL, _("Help"), _("View program help."));
    tools->AddTool(ID_HELPABOUT, wxEmptyString, Art::GetBitmap(wxT("about")), wxNullBitmap,
        wxITEM_NORMAL, _("About"), _("View information about the program."));

    tools->Realize();
    SetToolBar(tools);


    // Status bar
    CreateStatusBar(1, wxST_SIZEGRIP);

    // File history
    m_history = new wxFileHistory(9, wxID_FILE1);
    m_history->UseMenu(fileRecentMenu);

    // Set sizer (no need for main sizer)
    SetSizerAndFit(topSizer);
}

// Create a new file
void MainWindow::NewFile()
{
    // Create a new view
    View* v = new View(m_notebook);

    // Add file
    m_notebook->AddPage(v, v->GetDisplayName(), true);
}

// Open a file
bool MainWindow::OpenFile(const wxString& file, bool copy)
{
    // New view
    View* v = new View(m_notebook);

    // Open file
    bool result;

    if(copy)
    {
        result = v->LoadCopy(file);
    }
    else
    {
        result = v->Load(file);
    }

    if(result == true)
    {
        // Add file
        m_notebook->AddPage(v, v->GetDisplayName(), true);

        // Update file history
        m_history->AddFileToHistory(file);
        return true;
    }
    else
    {
        delete v;
        return false;
    }
}

// Save file
bool MainWindow::SaveFile(View* v, bool ask, bool copy)
{
    wxString fname = v->GetFileName();
    bool add = false;

    if(fname.empty() || ask == true || copy == true)
    {
        // Get file name
        fname = ::wxFileSelector(_("Save file"), wxEmptyString,
            wxEmptyString, wxEmptyString, wxT("*.*"),
            wxFD_SAVE | wxFD_OVERWRITE_PROMPT, this);

        // If still no file name, return
        if(fname.empty())
            return false;

        // Add to history only when asked for file name.  This way, normal save
        // does not move history item to top all the time
        add = true;
    }

    // Save the file
    bool result;

    if(copy == true)
        result = v->SaveCopy(fname);
    else
        result = v->Save(fname);

    if(result == true)
    {
        if(add)
            m_history->AddFileToHistory(fname);

        return true;
    }
    else
    {
        return false;
    }
}

// Query to close (ask to save or not)
bool MainWindow::QueryClose(View* v)
{
    if(v->IsModified())
    {
        int result;

        // Ask if user wants to save
        result = wxMessageBox(_("Save changes to file?"), wxGetApp().GetAppName(),
            wxYES_NO | wxCANCEL);

        if(result == wxYES)
        {
            // Continue if save is successful
            return SaveFile(v);
        }
        else if(result == wxNO)
        {
            // No pressed, continue without save
            return true;
        }
        else
        {
            // Cancel pressed, do not save and do not continue
            return false;
        }
    }
    else
    {
        // Not modified, okay to continue
        return true;
    }
}

// Close all files
bool MainWindow::CloseAllFiles()
{
    size_t count;

    count = m_notebook->GetPageCount();

    // Close all items
    while(count > 0)
    {
        count--; // Subtract one to get to the last item

        // Get view of the item
        View* v = (View*)m_notebook->GetPage(count);

        if(v)
        {
            // Select the page first
            m_notebook->SetSelection(count);

            // Close the item, if not, quit the close process
            if(QueryClose(v) == false)
                return false;
        }

        m_notebook->DeletePage(count);
    }

    return true;
}

// Event table
BEGIN_EVENT_TABLE(MainWindow, wxFrame)
    EVT_MENU(ID_FILENEW, MainWindow::OnFileNew)
    EVT_MENU(ID_FILEOPEN, MainWindow::OnFileOpen)
    EVT_MENU(ID_FILEOPENCOPY, MainWindow::OnFileOpenCopy)
    EVT_MENU_RANGE(wxID_FILE1, wxID_FILE9, MainWindow::OnFileOpenRecent)
    EVT_MENU(ID_FILESAVE, MainWindow::OnFileSave)
    EVT_MENU(ID_FILESAVEAS, MainWindow::OnFileSaveAs)
    EVT_MENU(ID_FILESAVECOPY, MainWindow::OnFileSaveCopy)
    EVT_MENU(ID_FILECLOSE, MainWindow::OnFileClose)
    EVT_MENU(ID_FILEEXIT, MainWindow::OnFileExit)

    EVT_MENU(ID_EDITCOPY, MainWindow::OnEditCopy)
    EVT_MENU(ID_EDITPASTE, MainWindow::OnEditPaste)
    EVT_MENU(ID_EDITEDIT, MainWindow::OnEditEdit)
    EVT_MENU(ID_EDITRENAME, MainWindow::OnEditRename)
    EVT_MENU(ID_EDITDELETE, MainWindow::OnEditDelete)
    EVT_MENU(ID_EDITCOPYCODE, MainWindow::OnEditCopyCode)
    EVT_MENU(ID_EDITENABLE, MainWindow::OnEditEnable)
    EVT_MENU(ID_EDITDISABLE, MainWindow::OnEditEnable)
    EVT_MENU(ID_EDITSORTA, MainWindow::OnEditSort)
    EVT_MENU(ID_EDITSORTD, MainWindow::OnEditSort)
    EVT_MENU(ID_EDITMOVEU, MainWindow::OnEditMove)
    EVT_MENU(ID_EDITMOVED, MainWindow::OnEditMove)

    EVT_MENU(ID_ITEMNEWCOLOR, MainWindow::OnItemNewColor)
    EVT_MENU(ID_ITEMNEWMAP, MainWindow::OnItemNewMap)

    EVT_MENU(ID_VIEWOPTIONS, MainWindow::OnViewOptions)
    EVT_MENU(ID_VIEWGAMMA, MainWindow::OnViewGamma)

    EVT_MENU(ID_HELPCONTENTS, MainWindow::OnHelpContents)
    EVT_MENU(ID_HELPABOUT, MainWindow::OnHelpAbout)

    EVT_CLOSE(MainWindow::OnClose)
END_EVENT_TABLE()


