/*
  Project:      POVColor
  File:         map.cpp
  Purpose:      Map item implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/dc.h>
#include <wx/icon.h>
#include <wx/dataobj.h>
#include <wx/module.h>
#include <wx/clipbrd.h>
#include <wx/txtstrm.h>
#include <wx/sstream.h>

#include "app.h"
#include "art.h"
#include "map.h"
#include "mapdlg.h"
#include "input.h"
#include "output.h"
#include "options.h"

#include <wx/arrimpl.cpp>


// MapItem
//-----------------------------------------------------------------------------

// Finish list
WX_DEFINE_OBJARRAY(MapColorArray);

// Destructor
MapItem::~MapItem()
{
    m_items.Clear();
}

// Add an item to the list
size_t MapItem::AddItem(const MapColor& color)
{
    // Stop optimization of ColorAtPos
    m_optimizeColorAtPos = false;

    // Create and insert item
    m_items.Add(color);
    
    m_lastAddedIndex = m_items.GetCount() - 1;
    m_lastAddedPos = color.GetPosition();

    return m_lastAddedIndex;
}

// Add an item to the list
size_t MapItem::AddItem(double pos, const Color& color)
{
    MapColor tmp(pos, color);
    
    return AddItem(tmp);
}
  
// Add an item to the list
size_t MapItem::AddItem(double pos, const Color& lower, const Color& upper)
{
    MapColor tmp(pos, lower, upper);
    
    return AddItem(tmp);
}
  
// Add an item to the list
// First, check the position of the last item added
size_t MapItem::AddItemCheckLast(double pos, const Color& color)
{
    // Stop optimizing ColorAtPos
    m_optimizeColorAtPos = false;

    pos = clip(pos, 0.0, 1.0);

    if(m_lastAddedIndex != (size_t)-1 && (pos == m_lastAddedPos))
    {
        if(m_lastAddedIndex < m_items.GetCount())
        {
            m_items[m_lastAddedIndex].UpperColor() = color;
            return m_lastAddedIndex;
        }
        else
        {
            return AddItem(pos, color);
        }
    }
    else
    {
        return AddItem(pos, color);
    }
}
  
// Delete an item 
void MapItem::Delete(size_t i)
{
    // Stop optimizing color lookup
    m_optimizeColorAtPos = false;
    m_lastAddedIndex = (size_t)-1;  

    // Find item and delete it
    m_items.RemoveAt(i, 1);
}

// Get an item
MapColor& MapItem::operator [](size_t k)
{
    // Item can be changed
    m_optimizeColorAtPos = false;
    
    return m_items[k];
}

// Get an item
const MapColor& MapItem::operator [](size_t k) const
{
    return m_items[k];
}
  
// Sort the items by position
static int MapItemSortFunc(MapColor** i1, MapColor** i2)
{
    double delta;

    delta =  (*i1)->GetPosition() - (*i2)->GetPosition();

    if(delta < 0.0)
        return -1;
    else if(delta > 0.0)
        return 1;
    else
        return 0;
}

void MapItem::Sort()
{
    m_optimizeColorAtPos = false;
    m_lastAddedIndex = (size_t)-1;

    m_items.Sort(MapItemSortFunc);
} 
  
// Determine the color at a given position
Color MapItem::ColorAtPos(double pos) const
{
    bool foundLower = false;
    bool foundUpper = false;
    MapColor lowerBound;
    MapColor upperBound;
    size_t i, count;

    // Can we optimize this method
    if(m_optimizeColorAtPos == false || pos <= m_lastLP || pos > m_lastUP)
    {
        // We can not optimize it
        count = GetCount();

        if(count == 0)
        {
            // No items, return black
            return Color();
        }

        // Find bounding items
        for(i = 0; i < count; i++)
        {
            MapColor curItem;
            double curPos;

            // Get item and position
            curItem = m_items[i];
            curPos = curItem.GetPosition();

            if(curPos < pos)
            {
                // Found lower color
                if(foundLower == false)
                {
                    foundLower = true;
                    lowerBound = curItem;
                }
                else
                {
                    // Is it closer than previous lower color
                    if(curPos > lowerBound.GetPosition())
                        lowerBound = curItem;
                }
            }
            else
            {
                // Found upper color
                if(foundUpper == false)
                {
                    foundUpper = true;
                    upperBound = curItem;
                }
                else
                {
                    // Is it closer than previous upper color
                    if(curPos < upperBound.GetPosition())
                        upperBound = curItem;
                }
            }
        }

        // Only one color may have be found
        if(foundLower == false)
        {
            lowerBound.SetColor(upperBound.LowerColor());
            lowerBound.SetPosition(0.0);
        }

        if(foundUpper == false)
        {
            upperBound.SetColor(lowerBound.UpperColor());
            upperBound.SetPosition(1.0);
        }


        // Remember closest lower and upper position and colors
        m_lastLP = lowerBound.GetPosition();
        m_lastLR = lowerBound.UpperColor().GetRed();
        m_lastLG = lowerBound.UpperColor().GetGreen();
        m_lastLB = lowerBound.UpperColor().GetBlue();
        m_lastLF = lowerBound.UpperColor().GetFilter();
        m_lastLT = lowerBound.UpperColor().GetTransmit();
        
        m_lastUP = upperBound.GetPosition();
        m_lastUR = upperBound.LowerColor().GetRed();
        m_lastUG = upperBound.LowerColor().GetGreen();
        m_lastUB = upperBound.LowerColor().GetBlue();
        m_lastUF = upperBound.LowerColor().GetFilter();
        m_lastUT = upperBound.LowerColor().GetTransmit();

        m_optimizeColorAtPos = true;
    }


    if(m_lastLP == m_lastUP)
    {
        // Can't divide by zero
        return Color(m_lastLR, m_lastLG, m_lastLB, m_lastLF, m_lastLT);
    }

    double ratio;
    double fr, fg, fb, ff, ft;

    ratio = (pos - m_lastLP) / (m_lastUP - m_lastLP);

    fr = m_lastLR + (ratio * (m_lastUR - m_lastLR));
    fg = m_lastLG + (ratio * (m_lastUG - m_lastLG));
    fb = m_lastLB + (ratio * (m_lastUB - m_lastLB));
    ff = m_lastLF + (ratio * (m_lastUF - m_lastLF));
    ft = m_lastLT + (ratio * (m_lastUT - m_lastLT));

    return Color(fr, fg, fb, ff, ft);
}
  
// Copy map data
void MapItem::Set(const MapItem& copy)
{
    // Clear any existing data
    m_items.Clear();
    
    // Copy any new items
    size_t count, pos;
    count = copy.GetCount();
    
    for(pos = 0; pos < count; pos++)
    {
        m_items.Add(copy[pos]);
    }

    // Clear optimization and last item (after adding items)
    m_lastAddedIndex = (size_t)-1;
    m_optimizeColorAtPos = false;   
} 
  
// Draw a range of the map
void MapItem::DrawRange(wxDC& dc, const wxRect& rect, double start, double end) const
{
    double delta;
    int block, swapPos, pos;
    bool swap;
    
    // Determine delta position
    delta = (end - start) / (double)(rect.width - 1);
    
    // Block size
    block = rect.height / 2;
    
    // Horizontal pass
    swapPos = 0;
    swap = false;
    for(pos = rect.x; pos < rect.x + rect.width; pos++, start += delta, swapPos++)
    {
        // Time to swap?
        if(swapPos == block)
        {
            swap = !swap;
            swapPos = 0;
        }
        
        // Pens for drawing
        Color c = ColorAtPos(start);
        
        wxPen pn1(c.Blend(Color(0.0, 0.0, 0.0)));
        wxPen pn2(c.Blend(Color(1.0, 1.0, 1.0)));
        
        // Draw
        if(swap)
        {
            dc.SetPen(pn1);
            dc.DrawLine(pos, rect.y, pos, rect.y + block);
            dc.SetPen(pn2);
            dc.DrawLine(pos, rect.y + block, pos, rect.y + rect.height);
        }
        else
        {
            dc.SetPen(pn2);
            dc.DrawLine(pos, rect.y, pos, rect.y + block);
            dc.SetPen(pn1);
            dc.DrawLine(pos, rect.y + block, pos, rect.y + rect.height);
        }   
    }
        
    // Done with map
    dc.SetPen(wxNullPen);
} 
  
// Get code
void MapItem::CopyCode() const
{
    size_t count, pos;
    wxString cur;
    
    // Since copying multiple lines, make sure they are correct line endings
    wxStringOutputStream result;
    wxTextOutputStream text(result);
        
    // Make copy of ourself
    // Do this because Save is const, but the item needs to be sorted before
    // writing it out to the file
    MapItem item(*this);
        
    // Sort items first
    item.Sort();

    // Items
    count = item.GetCount();
    for(pos = 0; pos < count; pos++)
    {
        // Position, upper, and lower items
        double p;
        Color lower, upper;
        
        p = item[pos].GetPosition();
        lower = item[pos].LowerColor();
        upper = item[pos].UpperColor();
        
        // Write item
        cur.Printf(wxT("[%.4f %s]"), p, lower.PrepareColorString().c_str());
        text.WriteString(cur);
        
        // Newline if more items
        if(pos != count - 1 || !(lower == upper))
            text.WriteString(wxT("\n"));
            
        // If lower and upper are not the same, write upper color
        if(!(lower == upper))
        {
            cur.Printf(wxT("[%.4f %s]"), p, upper.PrepareColorString().c_str());
            text.WriteString(cur);
            
            // Newline if more items
            if(pos != count - 1)
                text.WriteString(wxT("\n"));
        }
    }
    
    // Copy code
    if(wxTheClipboard->Open())
    {
        wxTextDataObject *data = new wxTextDataObject(result.GetString());
        
        wxTheClipboard->SetData(data);
        wxTheClipboard->Close();
    }
}

// Clone
PovItem* MapItem::Clone() const
{
    return new MapItem(*this);
}

// Load map item
bool MapItem::Load(PovInput& in)
{
    Token tok;

    // Skip colormap if there
    tok = in.GetNextToken();
    if(tok.GetType() == Token::TOKEN_KEYWORD_COLORMAP)
    {
        tok = in.GetNextToken();
    }
  
    // Open brace
    if(tok.GetType() != Token::TOKEN_OPEN_BLOCK)
    {
        in.UngetToken();
        return false;
    }
    
    // Get all items until closing brace
    while(true)
    {
        double p1, p2;
        ColorItem c1, c2;
        bool twoItems;
    
        p2 = 0.0; // Get rid of warning
    
        // Get token
        tok = in.GetNextToken();
    
        if(tok.GetType() == Token::TOKEN_CLOSE_BLOCK)
        {
            // Finished
            break;
        }
        else if(tok.GetType() == Token::TOKEN_OPEN_MAP)
        {
            // Beginning of map item
            twoItems = false;
      
            // Position 1
            tok = in.GetNextToken();
            if(tok.GetType() != Token::TOKEN_NUMBER)
            {
                in.UngetToken();
                return false;
            }
        
            p1 = tok.GetNumber();
      
            // Is there a comma
            tok = in.GetNextToken();
      
            if(tok.GetType() == Token::TOKEN_COMMA)
            {
                // Two items
                twoItems = true;
        
                // Second position
                tok = in.GetNextToken();
                if(tok.GetType() != Token::TOKEN_NUMBER)
                {
                    in.UngetToken();
                    return false;
                }
          
                p2 = tok.GetNumber();
            }
            else
            {
                // No comma
                in.UngetToken();
            }
        
            // Get first color
            if(!c1.Load(in))
            {
                return false;
            }
        
            // Get second color if any
            if(twoItems)
            {
                if(!c2.Load(in))
                {
                    return false;
                }
            }
        
            // Closing bracket
            tok = in.GetNextToken();
            if(tok.GetType() != Token::TOKEN_CLOSE_MAP)
            {
                in.UngetToken();
                return false;
            }
        
            // Fix values
            p1 = clip(p1, 0.0, 1.0);
            AddItemCheckLast(p1, c1);
      
            if(twoItems)
            {
                p2 = clip(p2, 0.0, 1.0);
                AddItemCheckLast(p2, c2);
            }
        }
        else
        {
            // Unknown token
            in.UngetToken();
            return false;
        }
    } // while
    
    return true;
}
  
// Save map item
bool MapItem::Save(PovOutput& out, long indent) const
{
    size_t count, pos;
    wxString name, result;
    
    // Get options
    Options& opt = wxGetApp().GetOptions();
    
    const wxString mapSpell = opt.GetColorMapKeyword();
    const bool openNL = opt.GetOpenBraceNewLine();
    const long openIndent = opt.GetOpenBraceIndent();
    const long bodyIndent = opt.GetBodyIndent();
    const long closeIndent = opt.GetCloseBraceIndent();

        
    // Make copy of ourself
    // Do this because Save is const, but the item needs to be sorted before
    // writing it out to the file
    MapItem item(*this);
        
    // Get name
    name = item.GetName();
    
    // Line 1: declare
    result.Printf(wxT("#declare %s = %s"), name.c_str(), mapSpell.c_str());
    out.WriteIndent(indent);
    out.WriteString(result);
  
    // Line 2 (maybe still 1): open brace
    if(!openNL)
    {
        out.WriteString(wxT(" {\n"));
    }
    else
    {
        out.WriteString(wxT("\n"));
        out.WriteIndent(indent + openIndent);
        out.WriteString(wxT("{\n"));
    }

    // Sort items first
    item.Sort();

    // Items
    count = item.GetCount();
    for(pos = 0; pos < count; pos++)
    {
        // Position, upper, and lower items
        double p;
        Color lower, upper;
        
        p = item[pos].GetPosition();
        lower = item[pos].LowerColor();
        upper = item[pos].UpperColor();
        
        // Write item
        result.Printf(wxT("[%.4f %s]\n"), p, lower.PrepareColorString().c_str());
        out.WriteIndent(indent + bodyIndent);
        out.WriteString(result);
        
        // If lower and upper are not the same, write upper color
        if(!(lower == upper))
        {
            result.Printf(wxT("[%.4f %s]\n"), p, upper.PrepareColorString().c_str());
            out.WriteIndent(indent + bodyIndent);
            out.WriteString(result);
        }
    }
  
    // Closing brace
    out.WriteIndent(indent + closeIndent);
    out.WriteString(wxT("}\n"));
    
    return (out.GetLastError() != wxSTREAM_WRITE_ERROR);
}
  
// Edit map item
bool MapItem::Edit(wxWindow* parent)
{
    MapDialog dlg(parent, this);
    
    if(dlg.ShowModal() == wxID_OK)
    {
        // Get items and sort them
        *this = dlg.GetMapItem();
        return true;
    }
    else
        return false;
}
  
// Draw map item
void MapItem::Draw(wxDC& dc, const wxRect& rect) const
{
    DrawRange(dc, rect, 0.0, 1.0);
}

// Draw the icon
void MapItem::DrawIcon(wxDC& dc, const wxRect& rect) const
{
    wxBitmap bmp;

    if(IsEnabled())
    {
        bmp = Art::GetBitmap(wxT("mape"));
    }
    else
    {
        bmp = Art::GetBitmap(wxT("mapd"));
    }

    if(bmp.IsOk())
    {
        dc.DrawBitmap(bmp, rect.x, rect.y, true);
    }
}
  
  
