/*
  Project:      POVColor
  File:         map.h
  Purpose:      Map item
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __POVCOLOR_MAP_H__
#define __POVCOLOR_MAP_H__


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/dynarray.h>

#include "item.h"
#include "color.h"
#include "util.h"


// MapColor - A color entry in a colormap
//-----------------------------------------------------------------------------
class MapColor
{
public:
    // Constructors
    MapColor();
    MapColor(const MapColor& copy);
    MapColor(double pos, const Color& c);
    MapColor(double pos, const Color& lower, const Color& upper);

    // Destructor
    ~MapColor();

    // Assignment
    MapColor& operator =(const MapColor& rhs);
        
    // Position
    void SetPosition(double pos);
    double GetPosition() const;
    
    // Color
    void SetColor(const Color& c);
    void SetColor(double r, double g, double b, double f, double t);
    
    // Red
    void SetRed(double r);
    
    // Green
    void SetGreen(double r);
        
    // Blue
    void SetBlue(double r);
        
    // Filter
    void SetFilter(double r);
        
    // Transmit
    void SetTransmit(double r);
        
    // Lower bound color
    Color& LowerColor();
    const Color& LowerColor() const;
        
    // Upper bound color
    Color& UpperColor();
    const Color& UpperColor() const;
        
private:
    Color m_lower, m_upper; 
    double m_pos;
};
  
// Array of map colors for map item
WX_DECLARE_OBJARRAY(MapColor, MapColorArray); 
  
// Inlines

// Constructors
inline MapColor::MapColor() : m_pos(0.0)
{
}
  
inline MapColor::MapColor(const MapColor& copy) :
    m_lower(copy.m_lower),
    m_upper(copy.m_upper),
    m_pos(clip(copy.m_pos, 0.0, 1.0))
{
}
  
inline MapColor::MapColor(double pos, const Color& c) :
    m_lower(c),
    m_upper(c),
    m_pos(clip(pos, 0.0, 1.0))
{
}
  
inline MapColor::MapColor(double pos, const Color& lower, const Color& upper) :
    m_lower(lower),
    m_upper(upper),
    m_pos(clip(pos, 0.0, 1.0))
{
}

// Destructors
inline MapColor::~MapColor()
{
}

// Assignment
inline MapColor& MapColor::operator =(const MapColor& rhs)
{
    if(this != &rhs)
    {
        m_pos = clip(rhs.m_pos, 0.0, 1.0);
        m_lower = rhs.m_lower;
        m_upper = rhs.m_upper;
    }
        
    return *this;
} 
  
// Position
inline void MapColor::SetPosition(double pos)
{
    m_pos = clip(pos, 0.0, 1.0);
}
  
inline double MapColor::GetPosition() const
{
    return m_pos;
}
  
// Color
inline void MapColor::SetColor(const Color& c)
{
    m_lower = m_upper = c;
}
  
inline void MapColor::SetColor(double r, double g, double b, double f, double t)
{
    m_lower.SetColor(r, g, b, f, t);
    m_upper.SetColor(r, g, b, f, t);
}

// Red
inline void MapColor::SetRed(double r)
{
    m_lower.SetRed(r);
    m_upper.SetRed(r);
}
  
// Green
inline void MapColor::SetGreen(double r)
{
    m_lower.SetGreen(r);
    m_upper.SetGreen(r);
}
  
// Blue
inline void MapColor::SetBlue(double r)
{
    m_lower.SetBlue(r);
    m_upper.SetBlue(r);
}
  
// Filter
inline void MapColor::SetFilter(double r)
{
    m_lower.SetFilter(r);
    m_upper.SetFilter(r);
}
  
// Transmit
inline void MapColor::SetTransmit(double r)
{
    m_lower.SetTransmit(r);
    m_upper.SetTransmit(r);
}     
  
// Lower bound color
inline Color& MapColor::LowerColor()
{
    return m_lower;
}   
  
inline const Color& MapColor::LowerColor() const
{
    return m_lower;
}
  
// Upper bound color
inline Color& MapColor::UpperColor()
{
    return m_upper;
}
        
inline const Color& MapColor::UpperColor() const
{
    return m_upper;
}
  



// MapItem - An item that is a colormap
//-----------------------------------------------------------------------------
class MapItem : public PovItem
{
public:
    // Constructors
    MapItem();      
    MapItem(const MapItem& copy);
    
    // Destructor
    ~MapItem();

    // Assignment
    MapItem& operator =(const MapItem& rhs);  

    // Add items to the list
    size_t AddItem(const MapColor& color);
    size_t AddItem(double pos, const Color& color);
    size_t AddItem(double pos, const Color& lower, const Color& upper);
    size_t AddItemCheckLast(double pos, const Color& color);    
    
    // Delete an item from the list
    void Delete(size_t i);
    
    // Clear all items from color map
    void Clear();
    
    // How many items are there
    size_t GetCount() const;
    
    // Access an item
    MapColor& operator [](size_t k);
    const MapColor& operator [](size_t k) const;
    
    // Sort items by position
    void Sort();
    
    // Calculate the color at a position
    Color ColorAtPos(double pos) const;
    
    // Draw part of a map
    void DrawRange(wxDC& dc, const wxRect& rect, double start, double end) const;
    
    void CopyCode() const;
    PovItem* Clone() const;
    bool Load(PovInput& in);
    bool Save(PovOutput& out, long indent) const;
    bool Edit(wxWindow* parent);
    void Draw(wxDC& dc, const wxRect& rect) const;
    void DrawIcon(wxDC& dc, const wxRect& rect) const;

private:
    void Set(const MapItem& copy);

    MapColorArray m_items;
    
    // Used by AddItemCheckLast
    double m_lastAddedPos;
    size_t m_lastAddedIndex;

    // Used to optimize color lookup in ColorAtPos
    mutable bool m_optimizeColorAtPos;

    mutable double m_lastLP;
    mutable double m_lastLR;
    mutable double m_lastLG;
    mutable double m_lastLB;
    mutable double m_lastLF;
    mutable double m_lastLT;

    mutable double m_lastUP;
    mutable double m_lastUR;
    mutable double m_lastUG;
    mutable double m_lastUB;
    mutable double m_lastUF;
    mutable double m_lastUT;
};
  
// Inlines

// Constructors
inline MapItem::MapItem() :
    m_lastAddedIndex((size_t)-1),
    m_optimizeColorAtPos(false)
{
}
  
inline MapItem::MapItem(const MapItem& copy) : PovItem(copy)
{
    Set(copy);
}

// Assignment
inline MapItem& MapItem::operator =(const MapItem& rhs)
{
    if(this != &rhs)
    {
        PovItem::operator =(rhs);
        Set(rhs);
    }
        
    return *this;
}

// Clear all items from color map
inline void MapItem::Clear()
{
    m_optimizeColorAtPos = false;
    m_lastAddedIndex = (size_t)-1;
    m_items.Clear();
}

// How many items are there
inline size_t MapItem::GetCount() const
{
    return m_items.GetCount();
}




#endif // Include once wrapper
