/*
  Project:      POVColor
  File:         mapdlg.cpp
  Purpose:      Map dialog implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/sizer.h>
#include <wx/dcclient.h>
#include <wx/dcbuffer.h>
#include <wx/button.h>
#include <wx/textctrl.h>
#include <wx/scrolbar.h>
#include <wx/stattext.h>
#include <wx/settings.h>
#include <wx/intl.h>
#include <wx/utils.h>
#include <wx/msgdlg.h>
#include <wx/spinctrl.h>

#include "app.h"
#include "mapdlg.h"
#include "coldlg.h"
#include "options.h"


// Defines
#define BAR_X 500 // Size of the bar
#define BAR_Y 80
#define BARSPACE_Y 2 // Scrollbar to bar

#define PREVIEW_X 16 // Size of a single preview area (there are 4 of them)
#define PREVIEW_Y 64
#define PREVIEWSPACE_X 10 // Preview area to texts
#define PREVIEWSPACE_Y 10 // Bar to preview area

#define SLIDER_X 14 // Size of the slider
#define SLIDER_Y 12

// MapBarWidget - The main editor widget
//-----------------------------------------------------------------------------

// Events for a map bar
const wxEventType EVENT_MAPBAR_CHANGED = wxNewEventType();
const wxEventType EVENT_MAPBAR_SELCHANGED = wxNewEventType();


// Constructor
MapBarEvent::MapBarEvent(wxEventType type, wxWindowID id, size_t sel)
        : wxCommandEvent(type, id), m_sel(sel)
{
}

MapBarEvent::MapBarEvent(const MapBarEvent& copy)
        : wxCommandEvent(copy), m_sel(copy.m_sel)
{
}

// Destructor
MapBarEvent::~MapBarEvent()
{
}


// Map bar widget
    
// Constructor
MapBarWidget::MapBarWidget(wxWindow* parent,
                           wxWindowID id,
                           const wxPoint& pos,
                           const wxSize& size,
                           const MapItem* item)
{
    m_selectedItem = (size_t)-1;

    wxControl::Create(parent, id, pos, size, wxBORDER_SUNKEN);
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    SetMinSize(size);

    if(item)
        m_map = *item;
}

// Destructor
MapBarWidget::~MapBarWidget()
{
}

// Set color
void MapBarWidget::SetSelectedColor(const Color& color1, const Color& color2)
{
    if(!HasSelection())
        return;

    m_map[m_selectedItem].LowerColor() = color1;
    m_map[m_selectedItem].UpperColor() = color2;

    Refresh();
    Update();
}

// Get color
void MapBarWidget::GetSelectedColor(Color& color1, Color& color2) const
{
    if(!HasSelection())
        return;

    color1 = m_map[m_selectedItem].LowerColor();
    color2 = m_map[m_selectedItem].UpperColor();
}

// Events
void MapBarWidget::OnLeftDown(wxMouseEvent& evt)
{
    // Capture
    CaptureMouse();
    int x = (int)evt.GetX();

    // Find the closest item
    int closestDistance;
    size_t closestItem = FindClosest(x, closestDistance);

    // Close enough?
    if(closestDistance <= SLIDER_X / 2)
    {
        if(m_selectedItem != closestItem)
        {
            m_selectedItem = closestItem;

            MapBarEvent cevt(EVENT_MAPBAR_SELCHANGED, GetId(), m_selectedItem);
            cevt.SetEventObject(this);
            GetParent()->ProcessEvent(cevt);
        }

        return;
    }
    else
    {
        // No item close enough, create a new item
        wxRect rect(GetClientRect());
        double pos = clip((double)(x - rect.x) / (double)(rect.width - 1), 0.0, 1.0);

        m_selectedItem = m_map.AddItem(pos, m_map.ColorAtPos(pos));

        Refresh();
        Update();

        // Fire event
        MapBarEvent cevt(EVENT_MAPBAR_SELCHANGED, GetId(), m_selectedItem);
        cevt.SetEventObject(this);
        GetParent()->ProcessEvent(cevt);
    }
}

void MapBarWidget::OnLeftUp(wxMouseEvent& WXUNUSED(evt))
{
    // Release
    if(HasCapture())
        ReleaseMouse();

    MapBarEvent cevt(EVENT_MAPBAR_CHANGED, GetId(), m_selectedItem);
    cevt.SetEventObject(this);
    GetParent()->ProcessEvent(cevt);
}

void MapBarWidget::OnMotion(wxMouseEvent& evt)
{
    if(!HasCapture())
        return;

    if(HasSelection())
    {
        // Position
        int x = (int)evt.GetX();
        wxRect rect(GetClientRect());

        double pos = clip((double)(x - rect.x) / (double)(rect.width - 1), 0.0, 1.0);

        m_map[m_selectedItem].SetPosition(pos);

        // Draw
        Refresh();
        Update();
    }
}

void MapBarWidget::OnRightDown(wxMouseEvent& evt)
{
    if(HasCapture())
        return;

    // Find closest item
    size_t closest;
    int dist;

    closest = FindClosest((int)evt.GetX(), dist);
    if(dist > SLIDER_X / 2)
        return;

    // Delete the closest item, and select no item
    m_map.Delete(closest);
    m_selectedItem = (size_t)-1;

    // Draw
    Refresh();
    Update();

    // Event (selection changed)
    MapBarEvent cevt(EVENT_MAPBAR_SELCHANGED, GetId(), m_selectedItem);
    cevt.SetEventObject(this);
    GetParent()->ProcessEvent(cevt);
}

void MapBarWidget::OnCaptureLost(wxMouseCaptureLostEvent& WXUNUSED(evt))
{
    MapBarEvent cevt(EVENT_MAPBAR_CHANGED, GetId(), m_selectedItem);
    cevt.SetEventObject(this);
    GetParent()->ProcessEvent(cevt);
}
    
void MapBarWidget::OnPaint(wxPaintEvent& WXUNUSED(evt))
{
    wxAutoBufferedPaintDC dc(this);
    wxRect rect(GetClientRect());

    dc.SetClippingRegion(rect);
    
    // Bar and sliders
    m_map.DrawRange(dc, rect, 0.0, 1.0);
    PaintSliders(dc, rect);

    dc.DestroyClippingRegion();
}

// Find closest entry
size_t MapBarWidget::FindClosest(int x, int& dist) const
{
    // Find closest item
    size_t pos, count;
    int closestDistance = 150000;
    size_t closestItem = (size_t)-1;

    wxRect rect(GetClientRect());

    for(pos = 0, count = m_map.GetCount(); pos < count; pos++)
    {
        // Screen position
        int curX = rect.x + (int)(m_map[pos].GetPosition() * (double)(rect.width - 1));

        // Distance from mouse point
        int curDist = x - curX;
        if(curDist < 0)
            curDist = -curDist;

        // Closer?
        if(closestItem == (size_t)-1 || curDist < closestDistance)
        {
            closestDistance = curDist;
            closestItem = pos;
        }
    }

    dist = closestDistance;
    return closestItem;
}

// Paint the sliders
void MapBarWidget::PaintSliders(wxDC& dc, const wxRect& rect)
{
    size_t pos, count;

    for(pos = 0, count = m_map.GetCount(); pos < count; pos++)
    {
        const MapColor& c = m_map[pos];
        int x = rect.x + (int)(c.GetPosition() * (double)(rect.width - 1));

        // Prepare drawing points
        wxPoint slider1[4];   
        wxPoint slider2[4];
        
        slider1[0].x = x - (SLIDER_X / 2);
        slider1[0].y = rect.y;

        slider1[1].x = x + (SLIDER_X / 2);
        slider1[1].y = rect.y;

        slider1[2].x = x;
        slider1[2].y = rect.y + SLIDER_Y;
        
        slider1[3].x = slider1[0].x;
        slider1[3].y = slider1[0].y;
        
        slider2[0].x = x - (SLIDER_X / 2);
        slider2[0].y = rect.y + rect.height - 1;

        slider2[1].x = x + (SLIDER_X / 2);
        slider2[1].y = rect.y + rect.height - 1;

        slider2[2].x = x;
        slider2[2].y = rect.y + rect.height - SLIDER_Y;
        
        slider2[3].x = slider2[0].x;
        slider2[3].y = slider2[0].y;  
            
        // Draw both
        dc.SetPen(*wxWHITE_PEN);
        dc.SetBrush(*wxBLACK_BRUSH);
        dc.DrawPolygon(4, slider1);
        
        dc.SetPen(*wxBLACK_PEN);
        dc.SetBrush(*wxWHITE_BRUSH);
        dc.DrawPolygon(4, slider2);
    }

    dc.SetPen(wxNullPen);
    dc.SetBrush(wxNullBrush);
}

// Event table
BEGIN_EVENT_TABLE(MapBarWidget, wxControl)
    EVT_LEFT_DOWN(MapBarWidget::OnLeftDown)
    EVT_LEFT_UP(MapBarWidget::OnLeftUp)
    EVT_MOTION(MapBarWidget::OnMotion)
    EVT_RIGHT_DOWN(MapBarWidget::OnRightDown)
    EVT_MOUSE_CAPTURE_LOST(MapBarWidget::OnCaptureLost)
    EVT_PAINT(MapBarWidget::OnPaint)
END_EVENT_TABLE()


// MapDialog
//-----------------------------------------------------------------------------

enum
{
    ID_BAR = wxID_HIGHEST + 1,
    ID_COLOR1,
    ID_COLOR2,
    ID_COLOR3,
    ID_COLOR4
};

// Constructors
MapDialog::MapDialog(wxWindow* parent, const MapItem* item)
{
    // Create dialog
    if(!wxDialog::Create(parent, -1, _("Map Editor"), wxDefaultPosition,
        wxSize(900, 900), wxDEFAULT_DIALOG_STYLE))
    {
        return;
    }
        
    CreateWidgets(item);
    CopyCode(); // Copy code of item
    CenterOnParent(wxBOTH);
}
  
// Destructor
MapDialog::~MapDialog()
{
}
  
// Get map item
const MapItem& MapDialog::GetMapItem() const
{
    return m_bar->GetMapItem();;
}

// Color selected
void MapDialog::OnColorSelected(ColorCellEvent& evt)
{
    // Make sure item is selected
    if(m_bar->HasSelection() == false)
        return;

    // Show dialog
    ColorItem item(evt.GetColor());
    ColorDialog dlg(this, &item);

    if(dlg.ShowModal() != wxID_OK)
        return;

    // Set the cell color
    item = dlg.GetColorItem();
    wxWindowID id = evt.GetId();

    if(id == ID_COLOR1 || id == ID_COLOR2 || id == ID_COLOR3)
    {
        m_color1->SetColor(item);
        m_color2->SetColor(item);
    }

    if(id == ID_COLOR2 || id == ID_COLOR3 || id == ID_COLOR4)
    {
        m_color3->SetColor(item);
        m_color4->SetColor(item);
    }

    // Set the bar item's color
    m_bar->SetSelectedColor(m_color1->GetColor(), m_color4->GetColor());

    CopyCode();
}

// Bar item's position has changed (finished dragging)
void MapDialog::OnBarChanged(MapBarEvent& WXUNUSED(evt))
{
    CopyCode();
}

// Another bar item has been selected, created, or deleted
void MapDialog::OnBarSelChanged(MapBarEvent& evt)
{
    if(evt.HasSelection())
    {
        Color c1, c2;

        m_bar->GetSelectedColor(c1, c2);
        
        m_color1->SetColor(c1);
        m_color2->SetColor(c1);
        m_color3->SetColor(c2);
        m_color4->SetColor(c2);
    }
    else
    {
        Color c1;

        m_color1->SetColor(c1);
        m_color2->SetColor(c1);
        m_color3->SetColor(c1);
        m_color4->SetColor(c1);
    }

    CopyCode();
}

// Create dialog elements
void MapDialog::CreateWidgets(const MapItem* item)
{
    wxBusyCursor busy;
    
    // Top sizer
    wxBoxSizer* topSizer = new wxBoxSizer(wxVERTICAL);
    
    // Bar area
    topSizer->Add(new wxStaticText(this, -1, _("Preview")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));

    m_bar = new MapBarWidget(this, ID_BAR, wxDefaultPosition, wxSize(BAR_X, BAR_Y), item);
    topSizer->Add(m_bar, wxSizerFlags(0).Expand().Border(wxBOTTOM));

    // Selected item color
    topSizer->Add(new wxStaticText(this, -1, _("Selected Color")),
        wxSizerFlags(0).Expand().Border(wxBOTTOM));

    // Color cells
    wxBoxSizer* colorSizer = new wxBoxSizer(wxHORIZONTAL);

    m_color1 = new ColorCellWidget(this, ID_COLOR1, wxDefaultPosition, wxSize(-1, BAR_Y / 2));
    colorSizer->Add(m_color1, wxSizerFlags(1).Expand().Border(wxRIGHT));

    m_color2 = new ColorCellWidget(this, ID_COLOR2, wxDefaultPosition, wxSize(-1, BAR_Y / 2));
    colorSizer->Add(m_color2, wxSizerFlags(1).Expand().Border(wxRIGHT));

    m_color3 = new ColorCellWidget(this, ID_COLOR3, wxDefaultPosition, wxSize(-1, BAR_Y / 2));
    colorSizer->Add(m_color3, wxSizerFlags(1).Expand().Border(wxRIGHT));

    m_color4 = new ColorCellWidget(this, ID_COLOR4, wxDefaultPosition, wxSize(-1, BAR_Y / 2));
    colorSizer->Add(m_color4, wxSizerFlags(1).Expand());

    topSizer->Add(colorSizer, wxSizerFlags(0).Expand());

    
    // Main sizer
    wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
    
    mainSizer->Add(topSizer, wxSizerFlags(1).Expand().DoubleBorder());
    
    // Buttons
    wxSizer* buttons = CreateSeparatedButtonSizer(wxOK | wxCANCEL);
    if(buttons)
    {
        mainSizer->Add(buttons, wxSizerFlags(0).Expand().DoubleBorder(wxLEFT | wxRIGHT | wxBOTTOM));
    }
    
    // Apply sizers and fit size
    SetSizerAndFit(mainSizer);
}
  
// Copy code
void MapDialog::CopyCode()
{
    Options &opt = wxGetApp().GetOptions();
    
    if(opt.GetAutoCopyCode())
    {
        m_bar->GetMapItem().CopyCode();
    }
} 
  
// Event table
BEGIN_EVENT_TABLE(MapDialog, wxDialog)
    EVT_COLORCELL_SELECTED(ID_COLOR1, MapDialog::OnColorSelected)
    EVT_COLORCELL_SELECTED(ID_COLOR2, MapDialog::OnColorSelected)
    EVT_COLORCELL_SELECTED(ID_COLOR3, MapDialog::OnColorSelected)
    EVT_COLORCELL_SELECTED(ID_COLOR4, MapDialog::OnColorSelected)
    EVT_MAPBAR_CHANGED(ID_BAR, MapDialog::OnBarChanged)
    EVT_MAPBAR_SELCHANGED(ID_BAR, MapDialog::OnBarSelChanged)
END_EVENT_TABLE()

