/*
  Project:      POVColor
  File:         mapdlg.h
  Purpose:      Map dialog
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __POVCOLOR_MAPDLG_H__
#define __POVCOLOR_MAPDLG_H__


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/dialog.h>
#include <wx/control.h>

#include "map.h"
#include "colorcell.h"


// Forward declarations


// MapBarWidget - The main editor widget
//-----------------------------------------------------------------------------

// Events for a map bar
extern const wxEventType EVENT_MAPBAR_CHANGED;
extern const wxEventType EVENT_MAPBAR_SELCHANGED;

class MapBarEvent : public wxCommandEvent
{
public:
    MapBarEvent(wxEventType type, wxWindowID id, size_t sel);
    MapBarEvent(const MapBarEvent& copy);

    ~MapBarEvent();

    size_t GetSelection() const { return m_sel; }
    bool HasSelection() const { return m_sel != (size_t)-1; }

    wxEvent* Clone() const { return new MapBarEvent(*this); }

private:
    // No assign
    MapBarEvent& operator=(const MapBarEvent& rhs);

    size_t m_sel;
};

typedef void (wxEvtHandler::*MapBarEventFunction)(MapBarEvent&);

#define MapBarEventHandler(func) \
    (wxObjectEventFunction)(wxEventFunction)(wxCommandEventFunction) \
    wxStaticCastEvent(MapBarEventFunction, &func)

#define EVT_MAPBAR_CHANGED(id, fn) \
    DECLARE_EVENT_TABLE_ENTRY(EVENT_MAPBAR_CHANGED, id, wxID_ANY, \
    MapBarEventHandler(fn), (wxObject*)NULL),

#define EVT_MAPBAR_SELCHANGED(id, fn) \
    DECLARE_EVENT_TABLE_ENTRY(EVENT_MAPBAR_SELCHANGED, id, wxID_ANY, \
    MapBarEventHandler(fn), (wxObject*)NULL),

// Map bar widget
class MapBarWidget : public wxControl
{
public:
    // Constructor
    MapBarWidget(wxWindow* parent,
                 wxWindowID id,
                 const wxPoint& pos,
                 const wxSize& size,
                 const MapItem* item = NULL);
    // Destructor
    ~MapBarWidget();

    // Get the map
    const MapItem& GetMapItem() const { return m_map; } 
    
    // Get the selected item
    size_t GetSelection() const { return m_selectedItem; }
    bool HasSelection() const { return m_selectedItem != (size_t)-1; }

    // Set/get the selected color
    void SetSelectedColor(const Color& color1, const Color& color2);
    void GetSelectedColor(Color& color1, Color& color2) const;

    // Events
    void OnLeftDown(wxMouseEvent& evt);
    void OnLeftUp(wxMouseEvent& evt);
    void OnMotion(wxMouseEvent& evt);
    void OnRightDown(wxMouseEvent& evt);
    void OnCaptureLost(wxMouseCaptureLostEvent& evt);
    void OnPaint(wxPaintEvent& evt);

private:
    // No copy/assign
    MapBarWidget(const MapBarWidget& copy);
    MapBarWidget& operator=(const MapBarWidget& rhs);

    size_t FindClosest(int x, int& dist) const;
    void PaintSliders(wxDC& dc, const wxRect& rect);

    MapItem m_map;
    size_t m_selectedItem;

DECLARE_EVENT_TABLE();
};


// MapDialog - Color map editor
//-----------------------------------------------------------------------------
class MapDialog : public wxDialog
{
public:
    // Constructor/Destructor
    MapDialog(wxWindow* parent, const MapItem* item = NULL);
    ~MapDialog();
    
    // Return map item
    const MapItem& GetMapItem() const;

    // Color changed
    void OnColorSelected(ColorCellEvent& evt);

    // Bar changing/changed
    void OnBarChanged(MapBarEvent& evt);
    void OnBarSelChanged(MapBarEvent& evt);
    
private:
    // No copy/assign
    MapDialog(const MapDialog& copy);
    MapDialog& operator=(const MapDialog& copy);

    // Create item
    void CreateWidgets(const MapItem* item);
    
    // Copy code
    void CopyCode();

    MapBarWidget* m_bar;
    ColorCellWidget* m_color1;
    ColorCellWidget* m_color2;
    ColorCellWidget* m_color3;
    ColorCellWidget* m_color4;

    
DECLARE_EVENT_TABLE();
};
  


#endif // Include once wrapper
