/*
  Project:      POVColor
  File:         optdlg.cpp
  Purpose:      Option dialog implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/sizer.h>
#include <wx/intl.h>
#include <wx/textctrl.h>
#include <wx/radiobut.h>
#include <wx/checkbox.h>
#include <wx/stattext.h>
#include <wx/utils.h>
#include <wx/spinctrl.h>  

#include "app.h"
#include "options.h"
#include "optdlg.h"


// OutputDialog
//-----------------------------------------------------------------------------

// ctor
OptionsDialog::OptionsDialog(wxWindow* parent)
{
    // Create dialog
    if(!wxDialog::Create(parent, -1, _("Options"), wxDefaultPosition,
        wxSize(900, 900), wxDEFAULT_DIALOG_STYLE))
    {
        return;
    }
        
    // Create items
    CreateWidgets();
    CenterOnParent(wxBOTH);
}
  
// dtor
OptionsDialog::~OptionsDialog()
{
}
  
// Set up values
void OptionsDialog::OnInitDialog(wxInitDialogEvent& WXUNUSED(evt))
{
    // Set items
    Options& opt = wxGetApp().GetOptions();
    
    // Indent
    m_indentSize->SetValue(opt.GetIndentSize());
    if(opt.GetIndentUseTab())
    {
        m_indentTab->SetValue(true);
    }
    else
    {
        m_indentSpace->SetValue(true);
    }
        
    m_openBraceIndent->SetValue(opt.GetOpenBraceIndent());
    if(opt.GetOpenBraceNewLine())
    {
        m_openBraceNL->SetValue(true);
    }
    else
    {
        m_openBraceSL->SetValue(true);
    }
        
    m_bodyIndent->SetValue(opt.GetBodyIndent());
    m_closeBraceIndent->SetValue(opt.GetCloseBraceIndent());
    
    // Spelling
    if(opt.GetSpellingType() == Options::OPTION_SPELLING_COLOR)
    {
        m_spellColor->SetValue(true);
    }
    else
    {
        m_spellColour->SetValue(true);
    }
        
    // Create type
    if(opt.GetCreateType() == Options::OPTION_CREATE_ALWAYS)
    {
        m_createAlways->SetValue(true);
    }
    else if(opt.GetCreateType() == Options::OPTION_CREATE_IFNDEF)
    {
        m_createIfndef->SetValue(true);
    }
    else
    {
        m_createUndef->SetValue(true);
    }
        
    // Version
    m_version->SetValue(opt.GetOutputVersion());
    
    // Autocopy
    m_autoCopy->SetValue(opt.GetAutoCopyCode());
}

// User pressed ok
void OptionsDialog::OnOk(wxCommandEvent& WXUNUSED(evt))
{
    Options& opt = wxGetApp().GetOptions();
    
    // Indent
    opt.SetIndentSize(m_indentSize->GetValue());
    opt.SetIndentUseTab(m_indentTab->GetValue());
    
    opt.SetOpenBraceIndent(m_openBraceIndent->GetValue());
    opt.SetOpenBraceNewLine(m_openBraceNL->GetValue());
    
    opt.SetBodyIndent(m_bodyIndent->GetValue());
    opt.SetCloseBraceIndent(m_closeBraceIndent->GetValue());
    
    // Spelling
    opt.SetSpellingType(m_spellColor->GetValue() ? Options::OPTION_SPELLING_COLOR : Options::OPTION_SPELLING_COLOUR);
    
    // Create type
    if(m_createAlways->GetValue())
        opt.SetCreateType(Options::OPTION_CREATE_ALWAYS);
    else if(m_createIfndef->GetValue())
        opt.SetCreateType(Options::OPTION_CREATE_IFNDEF);
    else
        opt.SetCreateType(Options::OPTION_CREATE_UNDEF);
        
    // Version
    opt.SetOutputVersion(m_version->GetValue());
    
    // Autocopy
    opt.SetAutoCopyCode(m_autoCopy->GetValue());
    
    // Done
    EndModal(wxID_OK);
}
  
  
// Create dialog items
void OptionsDialog::CreateWidgets()
{
    wxBusyCursor busy;
    
    // Indent 
    wxStaticBoxSizer* indentSizer = new wxStaticBoxSizer(wxVERTICAL, this, _("Indent"));
    
    wxFlexGridSizer* indentItemSizer = new wxFlexGridSizer(4, 4, wxSizerFlags::GetDefaultBorder(), wxSizerFlags::GetDefaultBorder());

    // Part 1, indent size
    indentItemSizer->Add(new wxStaticText(this, -1, _("Size")), wxSizerFlags(0).Expand());
        
    m_indentSize = new wxSpinCtrl(this, -1, wxEmptyString, wxDefaultPosition, wxDefaultSize,
        wxSP_ARROW_KEYS, 0, 100);
    indentItemSizer->Add(m_indentSize, wxSizerFlags(0).Expand());
    
    m_indentTab = new wxRadioButton(this, -1, _("Tabs"), wxDefaultPosition,
        wxDefaultSize, wxRB_GROUP);
    indentItemSizer->Add(m_indentTab, wxSizerFlags(0).Expand());
    
    m_indentSpace = new wxRadioButton(this, -1, _("Spaces"));
    indentItemSizer->Add(m_indentSpace, wxSizerFlags(0).Expand());
    
    // Part 2, open brace
    indentItemSizer->Add(new wxStaticText(this, -1, _("Open Brace")), wxSizerFlags(0).Expand());
        
    m_openBraceIndent = new wxSpinCtrl(this, -1, wxEmptyString, wxDefaultPosition,
        wxDefaultSize, wxSP_ARROW_KEYS, 0, 100);
    indentItemSizer->Add(m_openBraceIndent, wxSizerFlags(0).Expand());
    
    m_openBraceNL = new wxRadioButton(this, -1, _("New Line"), wxDefaultPosition,
        wxDefaultSize, wxRB_GROUP);
    indentItemSizer->Add(m_openBraceNL, wxSizerFlags(0).Expand());
    
    m_openBraceSL = new wxRadioButton(this, -1, _("Same Line"));
    indentItemSizer->Add(m_openBraceSL, wxSizerFlags(0).Expand());
    
    // Part 3, body
    indentItemSizer->Add(new wxStaticText(this, -1, _("Body")), wxSizerFlags(0).Expand());
        
    m_bodyIndent = new wxSpinCtrl(this, -1, wxEmptyString, wxDefaultPosition,
        wxDefaultSize, wxSP_ARROW_KEYS, 0, 100);
    indentItemSizer->Add(m_bodyIndent, wxSizerFlags(0).Expand());

    indentItemSizer->AddSpacer(0);
    indentItemSizer->AddSpacer(0);
    
    // Part 4, close brace
    indentItemSizer->Add(new wxStaticText(this, -1, _("Close Brace")), wxSizerFlags(0).Expand());
        
    m_closeBraceIndent = new wxSpinCtrl(this, -1, wxEmptyString, wxDefaultPosition,
        wxDefaultSize, wxSP_ARROW_KEYS, 0, 100);
    indentItemSizer->Add(m_closeBraceIndent, wxSizerFlags(0).Expand());
    
    indentItemSizer->AddSpacer(0);
    indentItemSizer->AddSpacer(0);
    
    indentSizer->Add(indentItemSizer, wxSizerFlags(0).Expand().Border());
    
    // Spelling sizer
    wxStaticBoxSizer* spellingSizer = new wxStaticBoxSizer(wxHORIZONTAL, this, _("Spelling"));
    
    m_spellColor = new wxRadioButton(this, -1, _("color, color_map"), wxDefaultPosition,
        wxDefaultSize, wxRB_GROUP);
    spellingSizer->Add(m_spellColor, wxSizerFlags(1).Expand().Border());
    
    m_spellColour = new wxRadioButton(this, -1, _("colour, colour_map"));
    spellingSizer->Add(m_spellColour, wxSizerFlags(1).Expand().Border(wxTOP | wxBOTTOM | wxRIGHT));
    
    // Create type sizer
    wxStaticBoxSizer *createSizer = new wxStaticBoxSizer(wxHORIZONTAL, this, _("Create Type"));
    
    m_createAlways = new wxRadioButton(this, -1, _("Always"), wxDefaultPosition,
        wxDefaultSize, wxRB_GROUP);
    createSizer->Add(m_createAlways, wxSizerFlags(1).Expand().Border());
    
    m_createIfndef = new wxRadioButton(this, -1, _("#ifndef"));
    createSizer->Add(m_createIfndef, wxSizerFlags(1).Expand().Border(wxTOP | wxBOTTOM | wxRIGHT));
    
    m_createUndef = new wxRadioButton(this, -1, _("#undef"));
    createSizer->Add(m_createUndef, wxSizerFlags(1).Expand().Border(wxTOP | wxBOTTOM | wxRIGHT));  
    
    // Version number
    wxStaticBoxSizer* versionSizer = new wxStaticBoxSizer(wxHORIZONTAL, this, _("Version"));
    
    m_version = new wxTextCtrl(this, -1);
    versionSizer->Add(m_version, wxSizerFlags(1).Expand().Border());
    
    // Autocopy
    wxStaticBoxSizer* autoCopySizer = new wxStaticBoxSizer(wxHORIZONTAL, this, _("Clipboard"));
    
    m_autoCopy = new wxCheckBox(this, -1, _("Automatically copy code to clipboard."));
    autoCopySizer->Add(m_autoCopy, wxSizerFlags(1).Expand().Border());
    
    // Top sizer
    wxBoxSizer* topSizer = new wxBoxSizer(wxVERTICAL);
    
    topSizer->Add(indentSizer, wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(spellingSizer, wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(createSizer, wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(versionSizer, wxSizerFlags(0).Expand().Border(wxBOTTOM));
    topSizer->Add(autoCopySizer, wxSizerFlags(0).Expand());
    
    // Main sizer
    wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
    
    mainSizer->Add(topSizer, wxSizerFlags(0).Expand().DoubleBorder());
    
    mainSizer->Add(CreateButtonSizer(wxOK | wxCANCEL),
        wxSizerFlags(0).Expand().DoubleBorder(wxLEFT | wxRIGHT | wxBOTTOM));
    
    // Apply sizers and fit size
    SetSizerAndFit(mainSizer);
}  
  
// Events
BEGIN_EVENT_TABLE(OptionsDialog, wxDialog)
    EVT_INIT_DIALOG(OptionsDialog::OnInitDialog)
    EVT_BUTTON(wxID_OK, OptionsDialog::OnOk)
END_EVENT_TABLE() 
