/*
  Project:      POVColor
  File:         options.cpp
  Purpose:      Load/save program options
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/config.h>
#include <wx/filename.h>

#include "app.h"
#include "options.h"
#include "optdlg.h"
#include "gammadlg.h"


// Options
//-----------------------------------------------------------------------------

// ctor/dtor
Options::Options()
{
    Init();
}


Options::~Options()
{
}

// Load options
void Options::Load(wxConfigBase* cfg)
{
    Init();
    if(cfg == NULL)
    {
        return;
    }

    // Set path
    wxConfigPathChanger pathChanger(cfg, wxT("/Options/"));
  
    // Gamma correction settings
    cfg->Read(wxT("GammaCorrectionEnabled"), &m_gammaCorrectionEnabled);
    cfg->Read(wxT("DisplayGamma"), &m_displayGamma);
    
    // Indent
    cfg->Read(wxT("IndentSize"), &m_indentSize);
    cfg->Read(wxT("IndentTab"), &m_indentTab);
    
    // Brace/body indent
    cfg->Read(wxT("OpenBraceNewLine"), &m_openBraceNewLine);
    cfg->Read(wxT("OpenBraceIndent"), &m_openBraceIndent);
    cfg->Read(wxT("BodyIndent"), &m_bodyIndent);
    cfg->Read(wxT("CloseBraceIndent"), &m_closeBraceIndent);
    
    // Spelling type
    cfg->Read(wxT("SpellingType"), &m_spelling);
    
    // Create type
    cfg->Read(wxT("CreateType"), &m_create);
    
    // Version
    cfg->Read(wxT("Version"), &m_version);
    
    // Auto copy
    cfg->Read(wxT("AutoCopy"), &m_autoCopy);

    // File history
    long fileCount = 0;
    cfg->Read(wxT("FileCount"), &fileCount);

    for(long filePos = 0; filePos < fileCount; filePos++)
    {
        wxString entry, file;

        // Format entry
        entry.Printf(wxT("File%d"), filePos + 1);

        if(cfg->Read(entry, &file))
        {
            m_fileHistory.Add(file);
        }
    }
}
  
// Save options
void Options::Save(wxConfigBase* cfg) const
{
    if(cfg == NULL)
    {
        return;
    }

    // Set path
    wxConfigPathChanger pathChanger(cfg, wxT("/Options/"));
    
    // Gamma correction settings
    cfg->Write(wxT("GammaCorrectionEnabled"), m_gammaCorrectionEnabled);
    cfg->Write(wxT("DisplayGamma"), m_displayGamma);
    
    // Indent
    cfg->Write(wxT("IndentSize"), m_indentSize);
    cfg->Write(wxT("IndentTab"), m_indentTab);
    
    // Brace/body indent
    cfg->Write(wxT("OpenBraceNewLine"), m_openBraceNewLine);
    cfg->Write(wxT("OpenBraceIndent"), m_openBraceIndent);
    cfg->Write(wxT("BodyIndent"), m_bodyIndent);
    cfg->Write(wxT("CloseBraceIndent"), m_closeBraceIndent);
    
    // Spelling type
    cfg->Write(wxT("SpellingType"), m_spelling);
    
    // Create type
    cfg->Write(wxT("CreateType"), m_create);  
    
    // Version
    cfg->Write(wxT("Version"), m_version);
    
    // Auto copy
    cfg->Write(wxT("AutoCopy"), m_autoCopy);

    // File history
    long fileCount = (long)m_fileHistory.GetCount();
    cfg->Write(wxT("FileCount"), fileCount);

    for(long filePos = 0; filePos < fileCount; filePos++)
    {
        wxString entry, file;

        // Format entry
        entry.Printf(wxT("File%d"), filePos + 1);

        // Get file
        file = m_fileHistory[(size_t)filePos];
        cfg->Write(entry, file);
    }
}
  
// Ouptut options
bool Options::PromptOptions(wxWindow* parent)
{
    OptionsDialog dlg(parent);
    
    return (dlg.ShowModal() == wxID_OK);
}

// Gamma options  
bool Options::PromptGamma(wxWindow* parent)
{
    GammaDialog dlg(parent);
    
    return (dlg.ShowModal() == wxID_OK);
}

// Init defaults
void Options::Init()
{
    // Gamma
    m_gammaCorrectionEnabled = true;
    m_displayGamma = 2.2;
    
    // Indent
    m_indentSize = 2;
    m_indentTab = false;
    
    // Brace/body indent
    m_openBraceNewLine = true;
    m_openBraceIndent = 0;
    m_bodyIndent = 1;
    m_closeBraceIndent = 0;
    
    // Spelling
    m_spelling = Options::OPTION_SPELLING_COLOR;
    
    // Create
    m_create = Options::OPTION_CREATE_ALWAYS;
    
    // Version
    m_version = wxT("3.6");
    
    // Autocopy
    m_autoCopy = true;

    // File history
    m_fileHistory.Clear();
}


