/*
  Project:      POVColor
  File:         options.h
  Purpose:      Options object
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __POVCOLOR_OPTIONS_H__
#define __POVCOLOR_OPTIONS_H__


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/string.h>
#include <wx/arrstr.h>

#include "util.h"


// Forward declarations
class WXDLLIMPEXP_BASE wxConfigBase;
class WXDLLIMPEXP_CORE wxWindow;


// Options
//-----------------------------------------------------------------------------
class Options
{
public:
    // Enumerations
    enum
    {
        // Spelling
        OPTION_SPELLING_COLOR = 1,
        OPTION_SPELLING_COLOUR,

        // Create type
        OPTION_CREATE_ALWAYS = 1,
        OPTION_CREATE_IFNDEF,
        OPTION_CREATE_UNDEF
    };

    // Constructor/destructor
    Options();
    ~Options();

    // Load or save settings
    void Load(wxConfigBase* cfg);
    void Save(wxConfigBase* cfg) const;
    
    // Gamma corrections
    void SetGammaCorrectionEnabled(bool enabled)
    {
        m_gammaCorrectionEnabled = enabled;
    }
    bool GetGammaCorrectionEnabled() const
    {
        return m_gammaCorrectionEnabled;
    }
    
    void SetDisplayGamma(double gamma)
    {
        m_displayGamma = gamma;
    }
    double GetDisplayGamma() const
    {
        return clip(m_displayGamma, 0.1, 9.9);
    }
    
    // Output indention options
    void SetIndentSize(long size)
    {
        m_indentSize = size;
    }
    long GetIndentSize() const
    {
        return m_indentSize;
    }
    
    void SetIndentUseTab(bool tab)
    {
        m_indentTab = tab;
    }
    bool GetIndentUseTab() const
    {
        return m_indentTab;
    }
    
    // Brace/body options
    void SetOpenBraceNewLine(bool newLine)
    {
        m_openBraceNewLine = newLine;
    }
    bool GetOpenBraceNewLine() const
    {
        return m_openBraceNewLine;
    }
    
    void SetOpenBraceIndent(long count)
    {
        m_openBraceIndent = count;
    }
    long GetOpenBraceIndent() const
    {
        return m_openBraceIndent;
    }
    
    void SetBodyIndent(long count)
    {
        m_bodyIndent = count;
    }
    long GetBodyIndent() const
    {
        return m_bodyIndent;
    }
    
    void SetCloseBraceIndent(long count)
    {
        m_closeBraceIndent = count;
    }
    long GetCloseBraceIndent() const
    {
        return m_closeBraceIndent;
    }
    
    // Output spelling options
    void SetSpellingType(long type)
    {
        m_spelling = type;
    }
    long GetSpellingType() const
    {
        return m_spelling;
    }
    wxString GetColorKeyword() const
    {
        return (m_spelling == Options::OPTION_SPELLING_COLOR) ?
            wxT("color") : wxT("colour");
    }
    wxString GetColorMapKeyword() const
    {
        return (m_spelling == Options::OPTION_SPELLING_COLOR) ?
            wxT("color_map") : wxT("colour_map");
    }
    
    // Output create type
    void SetCreateType(long type)
    {
        m_create = type;
    }
    long GetCreateType() const
    {
        return m_create;
    }
    
    // Output version
    void SetOutputVersion(const wxString& version)
    {
        m_version = version;
    }
    const wxString& GetOutputVersion() const
    {
        return m_version;
    }
    
    // Autocopy to clipboard
    void SetAutoCopyCode(bool copy)
    {
        m_autoCopy = copy;
    }
    bool GetAutoCopyCode() const
    {
        return m_autoCopy;
    }

    // Recent file history
    void SetFileHistory(const wxArrayString& history)
    {
        m_fileHistory = history;
    }

    const wxArrayString& GetFileHistory() const
    {
        return m_fileHistory;
    }
    
    // Prompt user to adjust options
    static bool PromptOptions(wxWindow* parent);   
    static bool PromptGamma(wxWindow* parent);
    
private:
    // No copy/assignment
    Options(const Options& copy);
    Options& operator=(const Options& copy);

    // Init defaults
    void Init();

    // Gamma correction
    bool m_gammaCorrectionEnabled;
    double m_displayGamma;
    
    // Indent
    long m_indentSize;
    bool m_indentTab;
    
    bool m_openBraceNewLine;
    long m_openBraceIndent;
    long m_bodyIndent;
    long m_closeBraceIndent;
    
    
    // Other output options
    long m_spelling;
    long m_create;
    
    // Output version
    wxString m_version;
    
    // Copy code automatically
    bool m_autoCopy;

    // File history
    wxArrayString m_fileHistory;
};
  

#endif // Include once header

