/*
  Project:      POVColor
  File:         output.cpp
  Purpose:      POV output implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/utils.h>
#include <wx/log.h>
#include <wx/wfstream.h>
#include <wx/txtstrm.h>
#include <wx/filename.h>
#include <wx/datetime.h>

#include "app.h"
#include "output.h"
#include "options.h"
#include "file.h"


// PovOutput
//-----------------------------------------------------------------------------

// Constructor
PovOutput::PovOutput(wxOutputStream& stream, const wxString& fname, wxEOL mode) :
    m_fname(fname),
    m_stream(stream),
    m_text(stream, mode)
{
}
  
// Destructor
PovOutput::~PovOutput()
{
}
  
// Any errors
wxStreamError PovOutput::GetLastError() const
{
    return m_stream.GetLastError();
}
  
// Write an indent (or more)
void PovOutput::WriteIndent(long count)
{
    Options &opt = wxGetApp().GetOptions();
    wxChar ch;
    long size;
    
    // Get indent options

    // Indent size
    size = opt.GetIndentSize();
    
    if(opt.GetIndentUseTab())
    {
        ch = wxT('\t');
    }
    else
    {
        ch = wxT(' ');
    }
        
    // Write indent
    long pos = size * count;
    if(pos < 0)
        pos = 0;
    else if(pos > 1024)
        pos = 1024;
        
    while(pos--)
        m_text.PutChar(ch);
}


// Write a string
void PovOutput::WriteString(const wxString &string)
{
    m_text.WriteString(string);
}
  
// Save a file
bool PovOutput::SaveFile(File &file)
{
    wxString id;
    bool success = true;
  
    // ID for include once code
    id = GenerateIncludeOnceID();

    // Write header comments
    if(WriteHeader() == false)
    {
        success = false;
    }

    // Write include once opening
    if(success)
    {
        if(WriteIncludeOnceOpen(id) == false)
            success = false;
    }

    // View stack
    if(success)
    {
        if(WriteViewIncludeStack() == false)
            success = false;
    }

    // Items
    if(success)
    {
        if(WriteItems(file) == false)
            success = false;
    }

    // Write include once closing
    if(success)
    {
        if(WriteIncludeOnceClose(id) == false)
            success = false;
    }

    // Check for any write errors
    if(success)
    {
        return true;
    }
    else
    {
        // Unable to save
        ::wxLogError(wxT("An error occured saving the file."));
        return false;
    }
} 

// Save an individual item
bool PovOutput::SaveItem(const PovItem& item, bool extras)
{
    // Get options
    Options &opt = wxGetApp().GetOptions();
    
    long type;
    long bIndent; // body indent
    long eIndent; // enable/disable indent
    
    type = opt.GetCreateType();
    bIndent = opt.GetBodyIndent();

    // Get item name
    wxString name = item.GetName();
    
    // Part 1: Enabled or disabled
    if(extras)
    {
        if(item.IsEnabled())
        {
            // POVColor Comment
            WriteString(wxT("/*POVCOLOR:ENABLED*/\n"));
            
            // Indent amount
            eIndent = 0;
        }
        else
        {
            // POVColor Comment
            WriteString(wxT("/*POVCOLOR:DISABLED*/\n"));
            
            // Block next item
            WriteString(wxT("#if(0)\n"));
            
            // Indent amount
            eIndent = bIndent;
        }
        
        // Part 2: create type
        if(type == Options::OPTION_CREATE_UNDEF)
        {
            // Undefine the symbol first
            
            // Line 1: ifdef
            WriteIndent(eIndent);
            WriteString(wxT("#ifdef("));
            WriteString(name);
            WriteString(wxT(")\n"));
            
            // Line 2: undef
            WriteIndent(eIndent + bIndent);
            WriteString(wxT("#undef "));
            WriteString(name);
            WriteString(wxT("\n"));
            
            // Line 3: end
            WriteIndent(eIndent);
            WriteString(wxT("#end\n"));
            
            // Space
            WriteString(wxT("\n"));
        }
        else if(type == Options::OPTION_CREATE_IFNDEF)
        {
            // Only if the symbol is not defined
            WriteIndent(eIndent);
            WriteString(wxT("#ifndef("));
            WriteString(name);
            WriteString(wxT(")\n"));
        }
    }
    else // !extras
    {
        eIndent = 0;
    }
        
    // Part 3, write the item
    if(!item.Save(*this, (type == Options::OPTION_CREATE_IFNDEF && extras == true) ?
        eIndent + bIndent : eIndent))
    {
        return false;
    }
        
    // Part 4: close ifndef
    if(extras)
    {
        if(type == Options::OPTION_CREATE_IFNDEF)
        {
            WriteIndent(eIndent);
            WriteString(wxT("#end // #ifndef\n"));
        }
        
        // Part 5: Enabled or disabled ending
        if(item.IsEnabled() == false)
            WriteString(wxT("#end // #if(0)\n"));
        
        // Space
        WriteString(wxT("\n\n"));      
    }
    
    // Done
    return (GetLastError() != wxSTREAM_WRITE_ERROR);
}
  
// Write headers
bool PovOutput::WriteHeader()
{
    wxFileName fn(m_fname);
    
    // Line 1: file name
    WriteString(wxT("// "));
    WriteString(fn.GetFullName());
    WriteString(wxT("\n"));
    
    // Line 2: created by
    WriteString(wxT("// Created with POVColor " APP_VERSION "\n"));
    
    // Line 3/4 : where
    WriteString(wxT("// " APP_WEBSITE "\n"));
    
    // Line 5 : Space
    WriteString(wxT("\n\n"));
    
    return (GetLastError() != wxSTREAM_WRITE_ERROR);
}
  
  
// Generate include once id
wxString PovOutput::GenerateIncludeOnceID() const
{
    wxFileName fn(m_fname);
    wxString id, user;
    size_t ms, mn;
    size_t ulen, pos, len, tmp;
    wxChar c1;
    
    // Get filename
    id = fn.GetFullName();
    
    // Make uppercase
    id.MakeUpper();
    id += wxT("_");
    
    // Use some system specific information to create the rest
    ms = (size_t)(wxDateTime::Now().GetMillisecond());
    mn = (size_t)(wxDateTime::Now().GetMinute());
    
    user = ::wxGetUserId();
    ulen = user.Len();
    
    // Calculate rest
    len = id.Len();
    
    while(len < 40)
    {
        // Compute some stuff
        c1 = user[mn % ulen];
        ms -= (size_t)c1;
        mn += (size_t)c1 * ms - len;
        tmp = ms;
        ms = mn;
        mn = tmp;
        
        // Use it to append data
        c1 = wxT('A') + (wxChar)(ms % 26);
        id += c1;
        
        // Next character
        len++;
    }
        
    // Replace any invalid characters with an undescore
    len = id.Len();
    for(pos = 0; pos < len; pos++)
    {
        if(pos == 0)
        {
            if(!wxIsalpha(id[pos]))
                id[pos] = wxT('_');
        }
        else
        {
            if(!wxIsalnum(id[pos]))
                id[pos] = wxT('_');
        }
    }
    
    return id;
}
  
// Write include once opening
bool PovOutput::WriteIncludeOnceOpen(const wxString &id)  
{
    // Options
    Options &opt = wxGetApp().GetOptions();
    wxString version;
    
    version = opt.GetOutputVersion();
        
    // Line 1: #ifndef
    WriteString(wxT("#ifndef("));
    WriteString(id);
    WriteString(wxT(")\n"));
    
    // Line 2: #declare
    WriteString(wxT("#declare "));
    WriteString(id);
    WriteString(wxT(" = version;\n"));
    
    // Line 3 : version
    WriteString(wxT("#version "));
    WriteString(version);
    WriteString(wxT(";\n"));
    
    // Space
    WriteString(wxT("\n\n"));
    
    return (GetLastError() != wxSTREAM_WRITE_ERROR);
}
  
// Write closing part
bool PovOutput::WriteIncludeOnceClose(const wxString &id)
{
    // Line 1 : version
    WriteString(wxT("#version "));
    WriteString(id);
    WriteString(wxT(";\n"));
    
    // Line 2 : #end
    WriteString(wxT("#end\n"));
    
    // Space
    WriteString(wxT("\n\n"));
    
    return (GetLastError() != wxSTREAM_WRITE_ERROR);
} 
  
// Write view include stack code
bool PovOutput::WriteViewIncludeStack()
{
    wxFileName fn(m_fname);
    
    // Line 1 : #ifdef
    WriteString(wxT("#ifdef(View_POV_Include_Stack)\n"));
    
    // Line 2 : #debug
    WriteIndent(1);
    WriteString(wxT("#debug \"including "));
    WriteString(fn.GetFullName());
    WriteString(wxT("\\n\"\n"));
    
    // Line 3, #end
    WriteString(wxT("#end\n"));
    
    // Space
    WriteString(wxT("\n\n"));
    
    return (GetLastError() != wxSTREAM_WRITE_ERROR);
}
  
// Actually write out the items
bool PovOutput::WriteItems(File &file)
{
    size_t count, pos;
    
    // Write each item in correct placement order
    count = file.GetCount();
    
    for(pos = 0; pos < count; pos++)
    {
        if(SaveItem(file[pos], true) == false) // Save each items with the extras
            return false;
    }
        
    return true;
}

