/*
  Project:      POVColor
  File:         view.cpp
  Purpose:      View implementation
  Author:       Brian Vanderburg II
  Copyright:    (C) 2007-2008 Brian Allen Vanderburg II

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Requirements
//-----------------------------------------------------------------------------

// Includes
#include "config.h"

#include <wx/defs.h>
#include <wx/panel.h>
#include <wx/sizer.h>
#include <wx/dcclient.h>
#include <wx/msgdlg.h>
#include <wx/textdlg.h>
#include <wx/log.h>
#include <wx/settings.h>
#include <wx/dataobj.h>
#include <wx/vlbox.h>
#include <wx/dcbuffer.h>
#include <wx/clipbrd.h>
#include <wx/filename.h>
#include <wx/sstream.h>
#include <wx/dynarray.h>

#include "view.h"
#include "coldlg.h"
#include "mapdlg.h"
#include "app.h"
#include "input.h"
#include "output.h"
#include "file.h"


// List box to show items
//-----------------------------------------------------------------------------
static const wxCoord iconXY = 32;
static const wxCoord previewY = 32;
static const wxCoord paddingXY = 4;

class ItemListBox : public wxVListBox
{
public:
    ItemListBox(wxWindow* parent, wxWindowID id, const wxPoint& pos,
        const wxSize& size, long style, const File& file);
    ~ItemListBox();
  
    wxCoord OnMeasureItem(size_t n) const;
    void OnDrawItem(wxDC& dc, const wxRect& rect, size_t n) const;
    
private:
    const File& m_file;    
};

ItemListBox::ItemListBox(wxWindow* parent, wxWindowID id, const wxPoint& pos,
        const wxSize& size, long style, const File& file) :
        wxVListBox(parent, id, pos, size, style), m_file(file)
    {
    }
  
ItemListBox::~ItemListBox()
{
}

wxCoord ItemListBox::OnMeasureItem(size_t n) const
{
    // Get the item
    wxString text;
  
    if(n >= m_file.GetCount())
        text = wxT("ASDFqypg");
    else
        text = m_file[n].GetName();
    
    // Left side has icon, right top is name, right bottom is preview
    int textX = 0, textY = 0;
    GetTextExtent(text, &textX, &textY);
  
    const wxCoord leftY = iconXY + (2 * paddingXY);
    const wxCoord rightY = previewY + textY + (3 * paddingXY);
  
    return (leftY > rightY) ? leftY : rightY;
}

void ItemListBox::OnDrawItem(wxDC& dc, const wxRect& rect, size_t n) const
{
    // Get the item
    if(n >= m_file.GetCount())
        return;
    
    const PovItem& item = m_file[n];
  
    // Draw icon
    wxRect iconRect;
  
    iconRect.x = rect.x + paddingXY;
    iconRect.width = iconXY;
    iconRect.y = rect.y + (rect.height - iconXY) / 2;
    iconRect.height = iconXY;
  
    item.DrawIcon(dc, iconRect);

    // Draw the item name
    wxColour prevColor = dc.GetTextForeground();
  
    // The background is drawn correctly with the window/highlight
    // color.  The text needs to be drawn correctly.
    if(IsSelected(n))
    {
        dc.SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_HIGHLIGHTTEXT));
    }
    else
    {
        dc.SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT));
    }
  
    dc.DrawText(item.GetName(), rect.x + iconXY + (2 * paddingXY), rect.y + paddingXY);

    dc.SetTextForeground(prevColor);
  
    // Draw the item
    wxRect itemRect;
  
    itemRect.x = rect.x + (iconXY + (2 * paddingXY));
    itemRect.width = rect.width - (iconXY + (3 * paddingXY));
    itemRect.y = rect.y + (rect.height - previewY - paddingXY);
    itemRect.height = previewY;
  
    item.Draw(dc, itemRect); 
    PovItem::DrawBevel(dc, itemRect);
}


// View
//-----------------------------------------------------------------------------

enum
{
    ID_LIST = wxID_HIGHEST + 1
};
  
// Constructor
View::View(wxWindow* parent)
{
    // Save parent for later operations
    m_parent = parent;
  
    // Create panel
    if(!wxPanel::Create(parent, -1, wxDefaultPosition, wxSize(900, 900)))
    {
        return;
    }
    
    // Create child items
    CreateWidgets();
  
    return;
}

// Destructor
View::~View()
{
    // Clear file no matter what
    m_file.Clear();
}
  
// Load a file
bool View::Load(const wxString& fname)
{
    // Load
    if(!m_file.Load(fname))
    {
        return false;
    }
    
    // Udpate list
    UpdateList();
  
    return true;
}

// Load a copy of a file
bool View::LoadCopy(const wxString& fname)
{
    // Load
    if(!m_file.LoadCopy(fname))
    {
        return false;
    }
    
    // Udpate item list
    UpdateList();
  
    return true;
} 
  
// Save the file
bool View::Save(const wxString& fname)
{
    return m_file.Save(fname);
}

// Save a copy of the file
bool View::SaveCopy(const wxString& fname)
{
     return m_file.SaveCopy(fname);
}
  
// Filename
wxString View::GetFileName() const
{
    return m_file.GetFileName();
}

// Display name
wxString View::GetDisplayName() const
{
    wxString file = m_file.GetFileName();
  
    if(file.empty())
    {
        return _("Untitled");
    }
    else
    {
        wxFileName fn(file);
    
        return fn.GetFullName();
    }
} 

// Modified or not
bool View::IsModified() const
{
    return m_file.IsModified();
}
  
// Copy the selected items
void View::CopyItems()
{
    // Stream to store items
    wxStringOutputStream stream;
  
    // Output
    PovOutput output(stream); 
  
    // Get selected items, and copy them
    unsigned long cookie;
    int sel;
  
    sel = m_list->GetFirstSelected(cookie);
    while(sel != wxNOT_FOUND)
    {
        // Get item from our file
        if((size_t)sel < m_file.GetCount())
        {
            if(output.SaveItem(m_file[sel]) == false)
            {
                ::wxLogError(_("An error occured while copying selected items."));
                return;        
            }
        }
      
        // Next item
        sel = m_list->GetNextSelected(cookie);
    }
  
    // Open clipboard and copy text
    // PovOutput uses wxTextOutputStream so line endings are set
    // as they should be
    if(wxTheClipboard->Open())
    {
        wxTextDataObject* tdo = new wxTextDataObject(stream.GetString());
        wxTheClipboard->SetData(tdo);
        wxTheClipboard->Close();
    }
}
  
// Paste items
void View::PasteItems()
{
    // Paste from the clipboard
    wxString text;
  
    if(wxTheClipboard->Open())
    {
        if(wxTheClipboard->IsSupported(wxDF_TEXT))
        {
            wxTextDataObject data;
            wxTheClipboard->GetData(data);
            text = data.GetText();
        }  
        wxTheClipboard->Close();
    }
  
    if(text.IsEmpty())
    {
        return;
    }
    
    // Parse items into a new file
    wxStringInputStream stream(text);
    PovInput input(stream);
    File clipbrd;
  
    if(input.Load(clipbrd) == false)
        return;

    // clipbrd now has items
    size_t count, pos;
  
    // Add items
    count = clipbrd.GetCount();
    for(pos = 0; pos < count; pos++)
    {
        // Copy item in clipboard
        PovItem *item = clipbrd[pos].Clone();
    
        // Add to our file
        m_file.AddItem(item);
    }
  
    // Update list
    UpdateList((int)(m_file.GetCount() - 1));
} 
  
// Just redraw the selected item
void View::RedrawList(int sel)
{
    // Redraw the list
    if(sel != REDRAW_REFRESHONLY)
    {
        m_list->DeselectAll();
    
        if(sel >= 0)
            m_list->Select((size_t)sel);
    }
    m_list->Refresh();
} 

// Update the list contents
void View::UpdateList(int sel)
{
    // If clearing the selection
    m_list->SetItemCount(m_file.GetCount());
    RedrawList(sel);
}
  
// Create a new color
void View::NewColor()
{
    ColorDialog dlg(m_parent);
  
    if(dlg.ShowModal() == wxID_OK)
    {
        // Get name
        wxString name;
        if(!GetValidName(name))
        {
            return;
        }
      
        // Get color
        PovItem* item = dlg.GetColorItem().Clone();
    
        // Name
        item->SetName(name);
    
        // Add it to the file
        m_file.AddItem(item);
    
        // Add it to list
        UpdateList((int)(m_file.GetCount() - 1));
    } 
} 
  
// Create a new map
void View::NewMap()
{
    MapDialog dlg(m_parent);
  
    if(dlg.ShowModal() == wxID_OK)
    {
        // Get name
        wxString name;
        if(!GetValidName(name))
        {
            return;
        }
      
        // Get map
        PovItem* item = dlg.GetMapItem().Clone();
    
        // Name
        item->SetName(name);
    
        // Add it to the file
        m_file.AddItem(item);
    
        // Add it to list
        UpdateList((int)(m_file.GetCount() - 1));
    } 
} 
        
    
// Delete selected items
void View::DeleteItem()
{
    // Confirm first
    if(::wxMessageBox(_("Realy delete selected items?"), wxGetApp().GetAppName(), wxYES_NO) == wxNO)
    {
        return;
    }
    
    // Delete items
    int sel;
    unsigned long cookie;
  
    int offset = 0;
    sel = m_list->GetFirstSelected(cookie);
    while(sel != wxNOT_FOUND)
    {
        // Delete from file
        // Remember how many items are removed, so we remove the correct items.
        // If item 2 and 4 is selected.  Removing item 2 from the file will make
        // item 4 become item 3.  So, the next selected item (item 4) - the
        // number of items removed (1) is 3, the item we want to remove
        m_file.Remove((size_t)(sel - offset));
        offset++;
    
        // Next item
        sel = m_list->GetNextSelected(cookie);
    }
    
    // Redraw
    // Because of a small bug in wxVListBox::SetItemCount, the list should be
    // deselected before it is redrawn, else GetItemCount may under certain
    // circumstances report the wrong number of items (>1) when it appears only
    // one item is selected.  Remove this when it is fixed in wxWidgets
    m_list->DeselectAll();
    UpdateList(REDRAW_DESELECT);
} 
  
// Rename an item
void View::RenameItem()
{
    if(m_list->GetSelectedCount() != 1)
    {
        return;
    }

    int sel = m_list->GetSelection();
    if(sel == wxNOT_FOUND)
    {
        return;
    }
    
    if((size_t)sel >= m_file.GetCount())
    {
        return;
    }
  
    // Get the item
    PovItem& item = m_file[(size_t)sel];
    
    // Get name
    wxString oldName, newName;
  
    oldName = newName = item.GetName();
  
    if(!GetValidName(newName, true))
    {
        return;
    }
    
    if(oldName == newName)
    {
        return;
    }
    
    // Disable if needed, before renaming
    if(m_file.IsEnabled(newName))
    {
        item.Enable(false);
    }
      
    item.SetName(newName);

    m_file.SetModified();
  
    // Just refresh
    RedrawList();
} 
  
// Copy code of the selected item
void View::CopyCode()
{
    if(m_list->GetSelectedCount() != 1)
    {
        return;
    }

    int sel = m_list->GetSelection();
    if(sel == wxNOT_FOUND)
    {
        return;
    }
    
    if((size_t)sel >= m_file.GetCount())
    {
        return;
    }
  
    // Copy code
    m_file[(size_t)sel].CopyCode();
} 
  
// Edit the selected item
void View::EditItem()
{
    if(m_list->GetSelectedCount() != 1)
    {
        return;
    }

    int sel = m_list->GetSelection();
    if(sel == wxNOT_FOUND)
    {
        return;
    }
  
    if((size_t)sel >= m_file.GetCount())
    {
        return;
    }
    
    // Edit item
    if(m_file[(size_t)sel].Edit(m_parent))
    {
        m_file.SetModified();
        RedrawList();
    }
} 
  
// Enable or disable items
void View::EnableItem(bool enabled)
{
    int sel;
    unsigned long cookie;
  
    sel = m_list->GetFirstSelected(cookie);
    while(sel != wxNOT_FOUND)
    {
        m_file.Enable(sel, enabled);

        sel = m_list->GetNextSelected(cookie); 
    }
  
    RedrawList();
} 

 
// Sort the items
void View::Sort(bool ascending /* = true */)
{
    m_file.Sort(ascending);
    RedrawList(REDRAW_DESELECT);
}

// Move selected items up
void View::MoveUp()
{
    int dest, sel;
    int pos, count;
    unsigned long cookie;
    wxArrayInt selections; // For the selections
    
    // Get the selections
    sel = m_list->GetFirstSelected(cookie);
    while(sel != wxNOT_FOUND)
    {
        selections.Add(sel);
        sel = m_list->GetNextSelected(cookie);
    }

    // Move items up
    count = selections.GetCount();
    dest = -1;

    for(pos = 0; pos < count; pos++)
    {
        sel = selections[pos];

        // Determine new position
        if(sel > 0)
        {
            if(dest < 0)
            {
                dest = sel - 1;
            }
            else
            {
                dest++;
            }
        }
        else
        {
            dest = 0;
        }

        // Move the item
        if(dest != sel)
        {
            m_file.Move(dest, sel);

            // Remember the new destination
            selections[pos] = dest;            
        }
    }

    // Update selections
    m_list->DeselectAll();
    for(pos = 0; pos < count; pos++)
    {
        m_list->Select(selections[pos]);
    }

    RedrawList();
}

// Move selected items down
void View::MoveDown()
{
    int dest, sel;
    int pos, count, fileCount;
    unsigned long cookie;
    wxArrayInt selections; // For the selections
    
    // Get the selections
    sel = m_list->GetFirstSelected(cookie);
    while(sel != wxNOT_FOUND)
    {
        selections.Add(sel);
        sel = m_list->GetNextSelected(cookie);
    }

    // Move items down
    count = selections.GetCount();
    fileCount = m_file.GetCount();
    dest = -1;

    for(pos = count - 1; pos >= 0; pos--)
    {
        sel = selections[pos];

        // Determine new position
        if(sel < fileCount - 1)
        {
            if(dest < 0)
            {
                dest = sel + 1;
            }
            else
            {
                dest--;
            }
        }
        else
        {
            dest = fileCount - 1;
        }

        // Move the item
        if(dest != sel)
        {
            m_file.Move(dest, sel);

            // Remember the new destination
            selections[pos] = dest;            
        }
    }

    // Update selections
    m_list->DeselectAll();
    for(pos = 0; pos < count; pos++)
    {
        m_list->Select(selections[pos]);
    }

    RedrawList();
}

  
// List item activated
void View::OnListItemActivate(wxCommandEvent& WXUNUSED(evt))
{
    EditItem();
}
  
// Create items
void View::CreateWidgets()
{
    // List
    m_list = new ItemListBox(this, ID_LIST, wxDefaultPosition, wxDefaultSize,
        wxLB_MULTIPLE | wxSUNKEN_BORDER, m_file);
      
    // Sizer
    wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  
    mainSizer->Add(m_list, wxSizerFlags(1).Expand().DoubleBorder());
  
    // Apply sizers and fit size
    SetSizer(mainSizer);
} 
  
// Get a valid name
bool View::GetValidName(wxString& name, bool force) const
{
    // If already valid and not forced, return true
    if(PovItem::NameIsValid(name) && force == false)
    {
        return true;     
    }
  
    // Text dialog
    wxString newName;
    wxTextEntryDialog dlg(m_parent, _("Enter a name for the item"),
            wxGetApp().GetAppName());
    
    // Set default name
    newName = name;        
  
    while(true)
    {
        dlg.SetValue(newName);
    
        if(dlg.ShowModal() == wxID_OK)
        {
            newName = dlg.GetValue();
      
            if(PovItem::NameIsValid(newName) == false)
            {
                // Invalid name
                ::wxMessageBox(_("The name is not valid"), wxGetApp().GetAppName());
            }
            else
            {
                // Valid name
                name = newName;
                return true;
            }
        }
        else
        {
            return false;
        }
    }
    
    // Should never get here
    return false;
}   
  

// Event table
BEGIN_EVENT_TABLE(View, wxPanel)
    EVT_LISTBOX_DCLICK(ID_LIST, View::OnListItemActivate)
END_EVENT_TABLE()

